#!/bin/bash

FILE_NAME=$(basename -- "$0")
FILE_DIR=$(dirname $(readlink -f $0))

TARGETS=(mc12705 nm_card nm_mezzo nm_quad NeuroMezzanine)

if [[ -z "${VERSION_MAJOR}" ]]; then
	VERSION_MAJOR=2
fi

if [[ -z "${VERSION_MINOR}" ]]; then
	VERSION_MINOR=1
fi

if [[ -z "${CUR_DATE}" ]]; then
	CUR_DATE=`date '+%y%m%d'`
fi

if [[ -z "${GIT_REVD}" ]]; then
	GIT_REVD=`git rev-parse --short=4 HEAD`
fi

PACKAGE_POSTFIX="-pci"
PACKAGE_VERSION="${VERSION_MAJOR}.${VERSION_MINOR}.${CUR_DATE}"

TEMP_DIR="$FILE_DIR/temp"

case "$1" in
	install)
		ACTION='install'
		shift

		;;
	remove)
		ACTION='remove'
		shift

		;;
	mkdeb)
		ACTION='mkdeb'
		shift

		;;
	mkrpm)
		ACTION='mkrpm'
		shift

		;;
	*)
		echo "Usage: $FILE_NAME [action] [target] [options]"
		echo "  [action]   = { install | remove | mkdeb | mkrpm}"
		echo "  [target]   = [ mc12705 | nm_card | nm_mezzo | nm_quad | all ]"
		echo "  [options]  = [--noclean] [--legacy-postinst] [--verbose]"
		echo ""
		exit -1
esac

case "$1" in
	-*)

		;;
	mc12705)
		TARGETS=(mc12705)
		shift
	
		;;
	nm_card)
		TARGETS=(nm_card)
		shift
	
		;;

	nm_mezzo)
		TARGETS=(nm_mezzo)
		shift

		;;
	nm_quad)
		TARGETS=(nm_quad)
		shift

		;;
	NeuroMezzanine)
		TARGETS=(NeuroMezzanine)
		shift

		;;
	all)
		TARGETS=(mc12705 nm_card nm_mezzo nm_quad NeuroMezzanine)
		shift
	
		;;
	?*)
		echo "Unknown target: $1"
		exit -1

		;;
esac

while [ -n "$1" ]
do
	case "$1" in
	--noclean)	
		NO_CLEAN=yes

		;;
	--legacy-postinst)	
		LEGACY_POSTINST=yes

		;;
	--verbose)	
		VERBOSE=yes

		;;
	*)
		echo "Unknown option: $1"
		exit -1

		;;
	esac

	shift
done

# All functions are executed in the "environment" with three global variables: 
# TARGET_NAME, PACKAGE_NAME, PACKAGE_VERSION. These variables must be set 
# before calling the function.

function copy_src()
{
	local SRC_DIR=$1
	local TARGET_DIR=$2
	
	sed -e "s/P_NAME/$PACKAGE_NAME/g" \
	    -e "s/P_VERSION/$PACKAGE_VERSION/g" \
	    -e "s/TARGET/$TARGET_NAME/g" \
	    "$SRC_DIR/dkms/dkms.conf" > "$TEMP_DIR/dkms.conf.pp"

	sed -e "s/V_MAJOR/$VERSION_MAJOR/g" \
	    -e "s/V_MINOR/$VERSION_MINOR/g" \
	    -e "s/V_DATE/$CUR_DATE/g" \
	    -e "s/V_REVD/$GIT_REVD/g" \
	    "$SRC_DIR/dkms/Makefile" > "$TEMP_DIR/Makefile.pp"
	
	install -d $TARGET_DIR
	cp -ar "$SRC_DIR/$TARGET_NAME/" "$TARGET_DIR"
	install "$SRC_DIR/nm6408.c" "$SRC_DIR/nm6408.h" \
		"$SRC_DIR/nm6408_ioctl.h" "$SRC_DIR/nm6408_map.h" "$TARGET_DIR"
	install "$TEMP_DIR/Makefile.pp" "$TARGET_DIR/Makefile"
	install "$TEMP_DIR/dkms.conf.pp" "$TARGET_DIR/dkms.conf"
}

function install_module()
{
	local DKMS_SRC="/usr/src/${PACKAGE_NAME}-${PACKAGE_VERSION}"
	
	copy_src $FILE_DIR $DKMS_SRC
	
	/usr/sbin/dkms install "${PACKAGE_NAME}/${PACKAGE_VERSION}" --force
	[ $? -eq 0 ] && cp $FILE_DIR/$TARGET_NAME/99-$TARGET_NAME.rules /etc/udev/rules.d
	return $?
}

function remove_module()
{
	local DKMS_SRC="/usr/src/${PACKAGE_NAME}-${PACKAGE_VERSION}"
	
	/usr/sbin/dkms remove "${PACKAGE_NAME}/${PACKAGE_VERSION}" --all
	[ $? -eq 0 ] && rm -rf $DKMS_SRC
	[ $? -eq 0 ] && rm -f /etc/udev/rules.d/99-$TARGET_NAME.rules
	return $?
}

function preproc_file()
{
	local BUILD_ARCH='all'
	local DATE_STR="$(date -R)"
	echo "modifying $1..."
	sed -e "s/DEBIAN_PACKAGE/${PACKAGE_NAME//_/-}/g" \
	    -e "s/DEBIAN_BUILD_ARCH/$BUILD_ARCH/g" \
	    -e "s/KERNEL_VERSION/$()/g" \
	    -e "s/MODULE_NAME/$PACKAGE_NAME/g" \
	    -e "s/MODULE_VERSION/$PACKAGE_VERSION/g" \
	    -e "s/TARGET/$TARGET_NAME/g" \
	    -e "s/DATE_STAMP/$DATE_STR/" "$1" > "$1.dkms-pp"
	chmod --reference="$1" "$1.dkms-pp"
	mv "$1.dkms-pp" "$1"
}

function make_deb()
{
	local BUILD_DIR="$TEMP_DIR/deb_build"
	
	# Creating build directory and copying src tree
	copy_src $FILE_DIR "$BUILD_DIR/${PACKAGE_NAME}-${PACKAGE_VERSION}"
	
	# Copying template
	cp -ar "$FILE_DIR/dkms/template-dkms-mkdeb/." "$BUILD_DIR"
	for file in $BUILD_DIR/debian/*
	do
		preproc_file "$file"
	done
	
	# If using legacy mode, install common postinst
	if [[ -n $LEGACY_POSTINST ]] 
	then
		echo "copying legacy postinstall template..."
		cp "$FILE_DIR/dkms/common.postinst" "$BUILD_DIR"
	fi
	
	# Creating deb package
	pushd "$BUILD_DIR" > /dev/null 2>&1
	echo $""
	if [[ -n $VERBOSE ]]; then
		export DH_VERBOSE=1
		dpkg-buildpackage -rfakeroot -d -b -us -uc
	else
		dpkg-buildpackage -rfakeroot -d -b -us -uc 1>/dev/null
	fi
	[ $? -eq 0 ] || return 1
	popd > /dev/null 2>&1
	cp "$TEMP_DIR"/*.deb "$FILE_DIR"
	[ $? -eq 0 ] || return 1
}

function make_rpm()
{
	local BUILD_DIR="$TEMP_DIR/rpm_build"
	
	# Creating build directory and copying template
	install -d "$BUILD_DIR"
	mkdir -p $BUILD_DIR/{BUILD,RPMS,SRPMS,SPECS,SOURCES}
	cp "$FILE_DIR/dkms/template-dkms-mkrpm.spec" "$BUILD_DIR/SPECS/dkms_mkrpm.spec"
	
	# If using legacy mode, install common postinst
	if [[ -n $LEGACY_POSTINST ]] 
	then
		echo "copying legacy postinstall template..."
		cp "$FILE_DIR/dkms/common.postinst" "$BUILD_DIR/SOURCES/"
	fi
	
	# Copying src tree
	copy_src $FILE_DIR "$BUILD_DIR/SOURCES/${PACKAGE_NAME}-${PACKAGE_VERSION}"
	
	LC_ALL=C rpmbuild --define "_topdir $BUILD_DIR" \
	                  --define "version ${PACKAGE_VERSION}" \
	                  --define "module_name ${PACKAGE_NAME}" \
	                  --define "target_name ${TARGET_NAME}" \
	                  --define "kernel_versions $(uname -r)" \
	                  --define "_use_internal_dependency_generator 0" \
	                  -ba $BUILD_DIR/SPECS/dkms_mkrpm.spec \
	                  > $BUILD_DIR/rpmbuild.log 2>&1
	[ $? -eq 0 ] || ( cat $BUILD_DIR/rpmbuild.log >&2 && return 1 )
	cp -a "$BUILD_DIR"/RPMS/noarch/* "$FILE_DIR"
	[ $? -eq 0 ] || return 1
	if [[ -n $VERBOSE ]]; then
		cat $BUILD_DIR/rpmbuild.log
	else
		cat $BUILD_DIR/rpmbuild.log | grep ^Wrote
	fi
}

function exec_act()
{
	local FUNC=$2
	local ACTION=$1
	local EXIT_CODE=$3
	
	$FUNC
	
	if [ $? -ne 0 ]
	then
		echo $""; echo "$ACTION failed!"
		exit $EXIT_CODE
	else
		echo $""; echo "$ACTION completed"
	fi
}

mkdir -p $TEMP_DIR

for TRG in ${TARGETS[*]}
do
	TARGET_NAME=$TRG
	PACKAGE_NAME=$TRG$PACKAGE_POSTFIX
	
	case "$ACTION" in
		install) exec_act install install_module 1
			;;
		remove)  exec_act remove remove_module 2
			;;
		mkdeb)	 exec_act mkdeb make_deb 3
			;;
		mkrpm)	 exec_act mkrpm make_rpm 4
			;;
		*)
			echo "It shouldn't have happened. Error."
			exit -1
	esac
done

if [[ $NO_CLEAN != "yes" ]]; then rm -rf "$TEMP_DIR"; fi

exit 0

