#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>

#include <linux/pci.h>
#include <linux/interrupt.h>

#include <linux/atomic.h>
#include <linux/scatterlist.h>

#include <linux/cdev.h>
#include <linux/fs.h>
#include <linux/ioctl.h>

#include <linux/uaccess.h>
#include <linux/sched.h>
#include <linux/dma-mapping.h>
#include <linux/pagemap.h>
#include <linux/mm.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/wait.h>
#include <linux/workqueue.h>
#include <linux/spinlock.h>
#include <linux/highmem.h>

#include "nm6408.h"

static struct class	*nm6408_class;
static dev_t		nm6408_dev;
static int		nm6408_major;
static int		nm6408_minor;

static unsigned long	nm6408_brds_indx;

#define CCPU_CCMB_FW_MAGIC		0x00000800

#define CCPU_MEMS_DMA_STATUS		(NM6408_INTERNEL_ADDR + 0x54)
#define CCPU_MEMS_SWI_COMMAND		(NM6408_INTERNEL_ADDR + 0x58)

#define CCPU_MEMS_DESC_OFFSET		(NM6408_INTERNEL_ADDR + 0x8000)
#define CCPU_MEMS_DESC_ADDR		(CCPU_INBOUND_AXI_BASE + CCPU_MEMS_DESC_OFFSET)

#define CCPU_MEMS_BOARD_READY		(NM6408_INTERNEL_ADDR + 0xF000)
#define CCPU_MEMS_BOARD_CHECK		(NM6408_INTERNEL_ADDR + 0xFFFC)

#define CCPU_REGS_GIC_SGIR		0x000B0F00
#define GIC_SGIR_SWI0			0x00010000

#define CCPU_REGS_CRG_CRGCR		0x000CA014
#define CRG_CRGCR_SWRSTREQ		(1 << 1)
#define CRG_CRGCR_REFLEN		(1 << 2)

#define CCPU_REGS_CRG_CRGSR		0x000CA01C
#define CRG_CRGSR_SWRSTACK		(1 << 1)

#define CCPU_REGS_CSC_PCIE_REG1		0x000CB018
#define PCIE_REG1_INT_PEND_STAUS	(1 << 8)

#define CCPU_REGS_CSC_PCIE_REG2		0x000CB01C
#define PCIE_REG2_PM_AUTOACK		(1 << 2)

#define CCPU_REGS_CSC_PCIE_REG3		0x000CB020
#define PCIE_REG3_INT_ASSERT		(1 << 0)
#define PCIE_REG3_INT_DEASSERT		(1 << 1)
#define PCIE_REG3_INT_READY		(1 << 2)

#define CCPU_REGS_WDT_LOAD		0x000CE000
#define WDT_LOAD_1_SEC			0x0BEBC200

#define CCPU_REGS_WDT_CTRL		0x000CE008
#define WDT_CTRL_INTEN			(1 << 0)
#define WDT_CTRL_RSTEN			(1 << 1)

#define CCPU_REGS_WDT_CLR		0x000CE00C
#define CCPU_REGS_WDT_RIS		0x000CE010
#define CCPU_REGS_WDT_MIS		0x000CE014
#define WDT_INT_STATUS			(1 << 0)

#define	CCPU_REGS_WDT_LOCK		0x000CEC00
#define WDT_UNLOCK_TEMPLATE		0x1ACCE551

#define CCPU_REGS_EP_INREGIONS		0x000DC840
#define CCPU_REGS_FW_CORESTART		0x000DC870

#define CCPU_REGS_DMA_CN0_CTRL		0x000DE000
#define DMA_CN_CTRL_GO			(1 << 0)
#define DMA_CN_CTRL_OUTBOUND		(1 << 1)

#define CCPU_REGS_DMA_CN0_SP_L		0x000DE004
#define CCPU_REGS_DMA_CN0_ATTR		0x000DE00C

#define CCPU_REGS_DMA_INT		0x000DE0A0
#define DMA_INT_CN0_DONE		(1 << 0)
#define DMA_INT_CN0_ERROR		(1 << 8)

#define CCPU_REG_DMA_INT_ENA		0x000DE0A4
#define CCPU_REG_DMA_INT_DIS		0x000DE0A8

#undef NM6408_DEBUG

#ifdef NM6408_DEBUG
#define nm6408_print_dbg(format, ...)	printk(KERN_INFO "%s#%d: " format ".\n", \
					__FUNCTION__, __LINE__, ##__VA_ARGS__)
#else
#define nm6408_print_dbg(format, ...)
#endif

#if LINUX_VERSION_CODE == KERNEL_VERSION(4, 4, 189)
#define LEGACY 1
#else
#define LEGACY 0
#endif

static int irq_type = 1;	// 1 - PCI_IRQ_MSI, 2 - PCI_IRQ_LEGACY;
module_param(irq_type, int, 0);

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 15, 1)
static void _check_timer(unsigned long argp)
#else
static void _check_timer(struct timer_list *timer)
#endif
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 15, 1)
	struct nm6408_brd	*brd = (struct nm6408_brd *)argp;
#else
	struct nm6408_brd	*brd = container_of(timer, struct nm6408_brd, timer);
#endif
	unsigned long		flags;
	void __iomem		*iomem;
	u32			tmp32;
	u16			tmp16;
	int			err = 0;

	nm6408_print_dbg("%s: begin", brd->name);

	spin_lock_irqsave(&brd->lock, flags);

	if (brd->irq <= 0) {
		spin_unlock_irqrestore(&brd->lock, flags);
		nm6408_print_dbg("%s: exit - warn", brd->name);
		return;
	}

	err = pci_read_config_word(brd->pdev, PCI_VENDOR_ID, &tmp16);
	if (err || (NM6408_VENDOR_ID != tmp16))
		err = 1;
	else {
		iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK;
		iowrite32(WDT_UNLOCK_TEMPLATE, iomem);

#if defined(CONFIG_E2K)
		ioread32(iomem); // barrier to ensure write
#endif

		if (ioread32(iomem) == 0) {
			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOAD;
			iowrite32(WDT_LOAD_1_SEC, iomem);

			mb();

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_RIS;
			tmp32 = ioread32(iomem);

			if (tmp32 & WDT_INT_STATUS) {
				iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CLR;
				iowrite32(tmp32, iomem);
			}
		}

		iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK;
		iowrite32(~WDT_UNLOCK_TEMPLATE, iomem);

		iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_BOARD_CHECK;

		tmp32 = ioread32(iomem);
		tmp32 = ~tmp32;
		iowrite32(tmp32, iomem);

#if defined(CONFIG_E2K)
		ioread32(iomem); // barrier to ensure write
#endif

		if (tmp32 != ioread32(iomem))
			err = 1;
	}

	if (err) {
		brd->irq = -1;

		if (NULL == brd->xfer_cur) {
			spin_unlock_irqrestore(&brd->lock, flags);
			printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 01).\n", brd->name);

			nm6408_print_dbg("%s: exit - err", brd->name);
			return;
		}

		brd->xfer_cur->stat = XFER_COMPLETE_ERROR;
		brd->xfer_cur = NULL;

		spin_unlock_irqrestore(&brd->lock, flags);
		printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 02).\n", brd->name);

		nm6408_print_dbg("%s: exit - err", brd->name);
		wake_up(&brd->wait_queue);

		return;
	}

	nm6408_print_dbg("%s: fire - check", brd->name);

	if (!brd->need_wait) {
		mod_timer(&brd->timer, jiffies + (NM6408_XFER_TIMEOUT * HZ) / 1000);
		spin_unlock_irqrestore(&brd->lock, flags);

		nm6408_print_dbg("%s: exit - ok", brd->name);
		return;
	}

	brd->need_wait = 0;

	if (NULL != brd->xfer_cur) {
		brd->xfer_cur->prev_dma_desc = brd->xfer_cur->cur_dma_desc;

		tmp32 = 0;

		while (brd->xfer_cur->cur_dma_desc < brd->xfer_cur->cnt_dma_desc) {
			if (-1 == brd->xfer_cur->cur_dma_desc)
				iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_DESC_OFFSET +
					brd->desc_num * sizeof(struct dma_desc);
			else
				iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_DESC_OFFSET +
					brd->xfer_cur->cur_dma_desc * sizeof(struct dma_desc);


			tmp32 = ioread32(iomem + offsetof(struct dma_desc, xfer_stat));
			if ((0x00000000 == tmp32) || (0xFFFFFFFF == tmp32))
				break;

			brd->xfer_cur->cur_dma_desc += 1;
		}

		if ((brd->xfer_cur->prev_dma_desc >= brd->xfer_cur->cnt_dma_desc) || (0xFFFFFFFF == tmp32))
			brd->xfer_cur->wait_dma_done = 0;
		else if (brd->xfer_cur->prev_dma_desc == brd->xfer_cur->cur_dma_desc)
			brd->xfer_cur->wait_dma_done -= 1;
		else
			brd->xfer_cur->wait_dma_done = 100;

		// Be careful! If interrupt relay occurs then we can break system. <--- !!!
		if (!brd->xfer_cur->wait_dma_done) {
			brd->irq = 0;

			brd->xfer_cur->stat = XFER_COMPLETE_TIMEOUT;
			brd->xfer_cur = NULL;

			spin_unlock_irqrestore(&brd->lock, flags);
			printk(KERN_WARNING "%s: Timeout of waiting operation completion with board.\n", brd->name);

			nm6408_print_dbg("%s: exit - timeout", brd->name);
			wake_up(&brd->wait_queue);

			return;
		}
	}

	mod_timer(&brd->timer, jiffies + (NM6408_XFER_TIMEOUT * HZ) / 1000);
	brd->need_wait = 1;

	spin_unlock_irqrestore(&brd->lock, flags);

	nm6408_print_dbg("%s: end", brd->name);
	return;
}

static void _pump_work(struct work_struct *work)
{
	struct nm6408_xfer	*xfer = container_of(work, struct nm6408_xfer, work);
	struct nm6408_brd	*brd = xfer->brd;
	unsigned long		flags;
	struct dma_desc		*desc;
	struct scatterlist	*sg;
	void __iomem		*iomem;
	size_t			axi_addr;
	u64			tmp64;
	u32			tmp32;
	int			i, n = 0;

	nm6408_print_dbg("%s: begin", brd->name);

	spin_lock_irqsave(&brd->lock, flags);
	if (brd->irq <= 0) {
		spin_unlock_irqrestore(&brd->lock, flags);
		nm6408_print_dbg("%s: exit - warn", brd->name);
		return;
	}

	brd->need_wait = 0;

	iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REG_DMA_INT_DIS;
	iowrite32(DMA_INT_CN0_DONE | DMA_INT_CN0_ERROR, iomem);
	tmp32 = ioread32(iomem);

	spin_unlock_irqrestore(&brd->lock, flags);

	desc = NULL;
	axi_addr = xfer->cur_pos;

	for_each_sg(xfer->sc_entry, sg, xfer->sc_nents, i) {
		dma_addr_t	dma_addr;
		size_t		xfer_len;

		dma_addr = sg_dma_address(sg) + xfer->dma_offset;
		xfer_len = sg_dma_len(sg) - xfer->dma_offset;

		while (xfer_len) {
			size_t	cur_size;

			if (n >= brd->desc_num) {
				xfer->dma_offset = dma_addr - sg_dma_address(sg);
				goto pump_work_end;
			}

			if (xfer_len > 0x1000000)
				cur_size = 0x1000000;
			else
				cur_size = xfer_len;

			desc = brd->desc_cpu_addr + n;
			memset(desc, 0, sizeof(struct dma_desc));

			tmp32 = CCPU_MEMS_DESC_ADDR + (n + 1) * sizeof(struct dma_desc);
			desc->next_desc_addr = cpu_to_le32(tmp32);

			tmp32 = axi_addr;
			desc->axi_data_addr = cpu_to_le32(tmp32);

			tmp64 = dma_addr;
			desc->pcie_data_addr = cpu_to_le64(tmp64);

			tmp32 = XFER_PRM_SIZE_BYTE(cur_size);
			tmp32 |= XFER_PRM_ALLOW_CHAIN;
			desc->xfer_prm = cpu_to_le32(tmp32);

			nm6408_print_dbg("%s: desc[%d] - global axi address = 0x%lx,"
					" host dma address = 0x%lx, size = 0x%lx",
					brd->name, n, axi_addr, dma_addr, cur_size);

			axi_addr += cur_size;
			dma_addr += cur_size;
			xfer_len -= cur_size;

			n++;
		}

		xfer->dma_offset = 0;
	}

pump_work_end:

	xfer->cur_pos = axi_addr;

	tmp32 = le32_to_cpu(desc->xfer_prm);
	tmp32 &= ~XFER_PRM_ALLOW_CHAIN;
	tmp32 |= XFER_PRM_ALLOW_INTR;
	desc->xfer_prm = cpu_to_le32(tmp32);

	if (i == xfer->sc_nents) {
		xfer->sc_entry = NULL;
		xfer->sc_nents = 0;
	}
	else {
		xfer->sc_entry = sg;
		xfer->sc_nents -= i;
	}

	iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_DESC_OFFSET +
		brd->desc_num * sizeof(struct dma_desc);

	iowrite32(CCPU_MEMS_DESC_ADDR, iomem + offsetof(struct dma_desc, axi_data_addr));
	iowrite32(0x0, iomem + offsetof(struct dma_desc, axi_data_attr));

	tmp64 = brd->desc_bus_addr;

	tmp32 = (u32)(tmp64 & 0xFFFFFFFF);
	iowrite32(tmp32, iomem + offsetof(struct dma_desc, pcie_data_addr));

	tmp32 = (u32)((tmp64 >> 32) & 0xFFFFFFFF);
	iowrite32(tmp32, iomem + offsetof(struct dma_desc, pcie_data_addr) + 0x4);

	iowrite32(0x0, iomem + offsetof(struct dma_desc, pcie_data_attr));

	tmp32 = XFER_PRM_SIZE_BYTE(n * sizeof(struct dma_desc));
	tmp32 |= XFER_PRM_ALLOW_INTR;
	iowrite32(tmp32, iomem + offsetof(struct dma_desc, xfer_prm));

	iowrite32(0x0, iomem + offsetof(struct dma_desc, xfer_stat));
	iowrite32(CCPU_MEMS_DESC_ADDR, iomem + offsetof(struct dma_desc, next_desc_addr));

	iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_DMA_INT;
	iowrite32(DMA_INT_CN0_DONE | DMA_INT_CN0_ERROR, iomem);

	iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_DMA_CN0_SP_L;
	tmp32 = CCPU_MEMS_DESC_ADDR + brd->desc_num * sizeof(struct dma_desc);
	iowrite32(tmp32, iomem);

	iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_DMA_CN0_ATTR;
	iowrite32(0x0, iomem);

	spin_lock_irqsave(&brd->lock, flags);

	if (brd->irq <= 0) {
		spin_unlock_irqrestore(&brd->lock, flags);

		nm6408_print_dbg("%s: exit - warn", brd->name);
		return;
	}

	iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REG_DMA_INT_ENA;
	iowrite32(DMA_INT_CN0_DONE | DMA_INT_CN0_ERROR, iomem);
	tmp32 = ioread32(iomem);

	xfer->prev_dma_desc = -1;
	xfer->cur_dma_desc = -1;
	xfer->cnt_dma_desc = n;
	xfer->wait_dma_done = 100;

	brd->need_wait = 1;
	spin_unlock_irqrestore(&brd->lock, flags);

	iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_DMA_CN0_CTRL;
	iowrite32(DMA_CN_CTRL_GO, iomem);

	nm6408_print_dbg("%s: end", brd->name);
	return;
}

static irqreturn_t nm6408_irq_handler(int irq, void *dev_id)
{
	struct nm6408_brd	*brd = (struct nm6408_brd *)dev_id;
	struct nm6408_xfer	*xfer;
	void __iomem		*iomem;
	u32			tmp32;
	u32			dma_status;
	u16			pci_status;
	u16			msi_flags;
	int			ret, err = 0;

	nm6408_print_dbg("%s: begin", brd->name);

	spin_lock(&brd->lock);

	if (brd->irq < 0) {
		spin_unlock(&brd->lock);

		nm6408_print_dbg("%s: exit - warn", brd->name);
		return IRQ_NONE;
	}

	ret = pci_read_config_word(brd->pdev, brd->pdev->msi_cap + PCI_MSI_FLAGS, &msi_flags);
	if (ret || (0xFFFF == msi_flags))
		err = 1;

	ret = pci_read_config_word(brd->pdev, PCI_STATUS, &pci_status);
	if (ret || (0xFFFF == pci_status))
		err = 1;

	if (err || (!(pci_status & PCI_STATUS_INTERRUPT) && !(msi_flags & PCI_MSI_FLAGS_ENABLE))) {
		if (err) {
			brd->irq = -1;

			if (NULL == brd->xfer_cur) {
				spin_unlock(&brd->lock);
				printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 03).\n",
				brd->name);

				nm6408_print_dbg("%s: exit - err", brd->name);
				return IRQ_NONE;
			}

			brd->xfer_cur->stat = XFER_COMPLETE_ERROR;
			brd->xfer_cur = NULL;

			spin_unlock(&brd->lock);
			printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 04).\n", brd->name);

			nm6408_print_dbg("%s: exit - err", brd->name);
			wake_up(&brd->wait_queue);

			return  IRQ_HANDLED;
		}

		spin_unlock(&brd->lock);

		nm6408_print_dbg("%s: exit - ok", brd->name);
		return IRQ_NONE;
	}

	nm6408_print_dbg("%s: fire - handle", brd->name);

	xfer = brd->xfer_cur;
	brd->need_wait = 0;

	iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REG_DMA_INT_DIS;
	iowrite32(DMA_INT_CN0_DONE | DMA_INT_CN0_ERROR, iomem);
	tmp32 = ioread32(iomem);

	spin_unlock(&brd->lock);

	iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_DMA_STATUS;
	do {
		dma_status = ioread32(iomem);
	} while (!dma_status);

	if (!(msi_flags & PCI_MSI_FLAGS_ENABLE)) {
		iomem = brd->region_virt[NM6408_BAR_CTRL]  + CCPU_REGS_CSC_PCIE_REG3;
		iowrite32(PCIE_REG3_INT_DEASSERT, iomem);

		while (!(ioread32(iomem) & PCIE_REG3_INT_READY)) {};

		iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_CSC_PCIE_REG1;

		tmp32 = ioread32(iomem);
		tmp32 &= ~PCIE_REG1_INT_PEND_STAUS;
		iowrite32(tmp32, iomem);
	}

	iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_DMA_STATUS;
	iowrite32(0, iomem);
	tmp32 = ioread32(iomem);

	// TODO: Work what is below may be move in the tasklet!!!

	spin_lock(&brd->lock);

	if (NULL == xfer) {
		brd->irq = -1;

		spin_unlock(&brd->lock);
		printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 05).\n", brd->name);

		nm6408_print_dbg("%s: exit - timeout", brd->name);
		return  IRQ_HANDLED;
	}

	if (dma_status & DMA_INT_CN0_ERROR) {
		xfer->stat = XFER_COMPLETE_ERROR;
		brd->xfer_cur = NULL;

		spin_unlock(&brd->lock);

		nm6408_print_dbg("%s: dma err", brd->name);
		wake_up(&brd->wait_queue);
	}
	else if (dma_status & DMA_INT_CN0_DONE) {
		iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_DMA_CN0_SP_L;

		if (CCPU_MEMS_DESC_ADDR != ioread32(iomem)) {
			nm6408_print_dbg("%s: desc xfer done", brd->name);

			iowrite32(CCPU_MEMS_DESC_ADDR, iomem);
			tmp32 = ioread32(iomem);

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REG_DMA_INT_ENA;
			iowrite32(DMA_INT_CN0_DONE | DMA_INT_CN0_ERROR, iomem);

			nm6408_print_dbg("%s: data xfer run", brd->name);

			brd->need_wait = 1;
			spin_unlock(&brd->lock);

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_DMA_CN0_CTRL;
			if (NM6408_READ == xfer->rw)
				iowrite32(DMA_CN_CTRL_GO | DMA_CN_CTRL_OUTBOUND, iomem);
			else
				iowrite32(DMA_CN_CTRL_GO, iomem);
		}
		else if (!xfer->sc_nents) {
			xfer->stat = XFER_COMPLETE_SUCCESS;
			brd->xfer_cur = NULL;

			spin_unlock(&brd->lock);

			nm6408_print_dbg("%s: dma ok", brd->name);
			wake_up(&brd->wait_queue);
		}
		else {
			spin_unlock(&brd->lock);

			nm6408_print_dbg("%s: data xfer done", brd->name);
			queue_work(brd->work_queue, &xfer->work);
		}
	}
	else {
		brd->irq = -1;

		xfer->stat = XFER_COMPLETE_ERROR;
		brd->xfer_cur = NULL;

		spin_unlock(&brd->lock);

		printk(KERN_ERR "%s: unknown interrupt from board.\n", brd->name);
		wake_up(&brd->wait_queue);
	}

	nm6408_print_dbg("%s: end", brd->name);
	return IRQ_HANDLED;
}

static int nm6408_open(struct inode *inode, struct file *filp)
{
	struct nm6408_brd	*brd = container_of(inode->i_cdev, struct nm6408_brd, cdev);
	struct nm6408_xfer	*xfer;
	int			i;

	xfer = kmalloc(sizeof(struct nm6408_xfer), GFP_KERNEL);
	if (!xfer)
		return -ENOMEM;

	xfer->brd = brd;

	for (i = 0; i < NM6408_CCPU; i++) {
		xfer->cluster_base[i] = PC0_INBOUND_AXI_BASE + i * NM6408_INBOUND_AXI_SIZE;
		xfer->cluster_limit[i] = PC0_INBOUND_AXI_LIMIT + i * NM6408_INBOUND_AXI_SIZE;
		xfer->cluster_offset[i] = 0x00000000;
	}

	xfer->cluster_base[NM6408_CCPU] = CCPU_INBOUND_AXI_BASE;
	xfer->cluster_limit[NM6408_CCPU] = CCPU_INBOUND_AXI_LIMIT;
	xfer->cluster_offset[NM6408_CCPU] = 0x00000000;

	xfer->cluster_indx = NM6408_CCPU;

	xfer->str_pos = CCPU_INBOUND_AXI_BASE;
	xfer->cur_pos = CCPU_INBOUND_AXI_BASE;

	INIT_WORK(&xfer->work, _pump_work);
	filp->private_data = xfer;

	nm6408_print_dbg("%s: pid = %d, tgid = 0x%x, inode = %lu, filp = %lx, board = %d, cluster = %d,"
			" local axi address = 0x%lx, global axi address = 0x%lx", brd->name, current->pid,
			current->tgid, inode->i_ino, filp, brd->indx, xfer->cluster_indx,
			xfer->cluster_offset[xfer->cluster_indx], xfer->cur_pos);

	return 0;
}

static int nm6408_release(struct inode *inode, struct file *filp)
{
	struct nm6408_brd	*brd = container_of(inode->i_cdev, struct nm6408_brd, cdev);
	struct nm6408_xfer	*xfer = filp->private_data;

	nm6408_print_dbg("%s: pid = %d, tgid = 0x%x, inode = %lu, filp = %lx, board = %d, cluster = %d,"
			" local axi address = 0x%lx, global axi address = 0x%lx", brd->name, current->pid,
			current->tgid, inode->i_ino, filp, brd->indx, xfer->cluster_indx,
			xfer->cluster_offset[xfer->cluster_indx], xfer->cur_pos);

	if (brd->xfer_cur == xfer)
		return -EBUSY;

	filp->private_data = NULL;
	kfree(xfer);

	return 0;
}

static int nm6408_check_axi_addr(size_t addr, size_t base, size_t cnt, int rw)
{
	size_t pos_start = addr - base;
	size_t pos_end = pos_start + cnt;
	int rv = 0;

	if ((pos_start & 0x3) || (pos_end & 0x3))
		return -1;

	if (base !=  CCPU_INBOUND_AXI_BASE) {
		if ((pos_start < (PC_RESERVE_AXI_ADDR + PC_RESERVE_AXI_SIZE)) &&
		    (pos_end > PC_RESERVE_AXI_ADDR)) {
			rv = -2;
			goto axi_check_end;
		}

		if ((pos_start < (PC_REG_AXI_ADDR + PC_REG_AXI_SIZE)) &&
		    (pos_end > PC_REG_AXI_ADDR)) {
			rv = -3;
			goto axi_check_end;
		}
	}
	else {
		if ((NM6408_WRITE == rw) &&
		    (pos_start < (NM6408_INTERNEL_ADDR + NM6408_INTERNEL_SIZE)) &&
		    (pos_end > NM6408_INTERNEL_ADDR)) {
			rv = -4;
			goto axi_check_end;
		}

		if ((pos_start < (CCPU_REG_AXI_ADDR + CCPU_REG_AXI_SIZE)) &&
		    (pos_end > CCPU_REG_AXI_ADDR)) {
			rv = -5;
			goto axi_check_end;
		}
	}

axi_check_end:

	return rv;
}

static int nm6408_is_ddr_addr(int indx, size_t addr, size_t offset, size_t cnt)
{
	size_t pos_start = addr + offset;
	size_t pos_end = pos_start + cnt;

	if ((indx != NM6408_CCPU) &&
	    (pos_start < (PC_EMI_AXI_ADDR + PC_EMI_AXI_SIZE)) &&
	    (pos_end > PC_EMI_AXI_ADDR)) {
		return 1;
	}

	if ((indx == NM6408_CCPU) &&
	    (pos_start < (CCPU_EMI_AXI_ADDR + CCPU_EMI_AXI_SIZE)) &&
	    (pos_end > CCPU_EMI_AXI_ADDR)) {
		return 1;
	}

	return 0;
}

static long nm6408_ioctl(struct file *filp, unsigned int cmd, unsigned long argp)
{
	struct nm6408_xfer		*xfer = filp->private_data;
	struct nm6408_brd		*brd = xfer->brd;

	union {
		struct nm6408_access	access;
		struct nm6408_wind	wind;
		struct nm6408_init_code	init_code;
	} tmp;

	unsigned long			flags;
	void __iomem			*iomem;
	void				*buffer;
	size_t				axi_addr;
	size_t				axi_offset;
	size_t				axi_size;
	u32				bar_ctrl;
	u16				vendor_id;
	long				i, rv = 0;

	nm6408_print_dbg("%s: begin - pid = %d, tgid = 0x%x, filp = %lx, board = %d, command = 0x%x",
			brd->name, current->pid, current->tgid, filp, brd->indx, cmd);

	if (down_interruptible(&brd->semap)) {
		printk(KERN_INFO "%s: interrupt - pid = %d, tgid = 0x%x, board = %d.\n",
			brd->name, current->pid, current->tgid, brd->indx);

		return -ERESTARTSYS;
	}

	switch (cmd) {
	case NM6408_DO_ACCESS:
		spin_lock_irqsave(&brd->lock, flags);
		if (brd->irq <= 0) {
			spin_unlock_irqrestore(&brd->lock, flags);

			rv = -EIO;
			break;
		}
		spin_unlock_irqrestore(&brd->lock, flags);

		rv = copy_from_user(&tmp.access, (void __user *)argp, sizeof(tmp.access));
		if (rv) {
			rv = -EFAULT;
			break;
		}

		iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_EP_INREGIONS +
			8 * brd->region_indx[NM6408_BAR_MEM];

		iowrite32(xfer->cluster_base[xfer->cluster_indx] +
			  xfer->cluster_offset[xfer->cluster_indx], iomem);

#if defined(CONFIG_E2K)
		ioread32(iomem); // barrier to ensure write
#endif

		axi_addr = ioread32(iomem);
		if (axi_addr != (xfer->cluster_base[xfer->cluster_indx] +
		    xfer->cluster_offset[xfer->cluster_indx])) {
			spin_lock_irqsave(&brd->lock, flags);
			brd->irq = -1;
			spin_unlock_irqrestore(&brd->lock, flags);

			printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 06).\n", brd->name);
			rv = -EIO;
			break;
		}

		axi_size = brd->region_len[NM6408_BAR_MEM];
		axi_offset = tmp.access.offset & ~(tmp.access.width - 1);
		axi_addr = xfer->cluster_offset[xfer->cluster_indx];

		if (axi_offset > (axi_size - tmp.access.width))
			axi_offset = axi_size - tmp.access.width;

		if (!brd->is_alive) {
			rv = nm6408_is_ddr_addr(xfer->cluster_indx, axi_addr, axi_offset, tmp.access.width);
			if (rv) {
				rv = -EINVAL;
				break;
			}
		}

		iomem = brd->region_virt[NM6408_BAR_MEM] + axi_offset;

		if (NM6408_WIDTH32 == tmp.access.width) {
			if (NM6408_READ == tmp.access.cmd)
				tmp.access.data.b32 = ioread32(iomem);
			else if (NM6408_WRITE == tmp.access.cmd)
				iowrite32(tmp.access.data.b32, iomem);
			else {
				rv = -EINVAL;
				break;
			}
		}
		else if (NM6408_WIDTH16 == tmp.access.width) {
			if (NM6408_READ == tmp.access.cmd)
				tmp.access.data.b16 = ioread16(iomem);
			else if (NM6408_WRITE == tmp.access.cmd)
				iowrite16(tmp.access.data.b16, iomem);
			else {
				rv = -EINVAL;
				break;
			}
		}
		else if (NM6408_WIDTH8 == tmp.access.width) {
			if (NM6408_READ == tmp.access.cmd)
				tmp.access.data.b8 = ioread8(iomem);
			else if (NM6408_WRITE == tmp.access.cmd)
				iowrite8(tmp.access.data.b8, iomem);
			else {
				rv = -EINVAL;
				break;
			}
		}
		else {
			rv = -EINVAL;
			break;
		}

		nm6408_print_dbg("%s: %s%d - filp = %lx, board = %d, cluster = %d,"
				" host iomem = %lx, local axi address = 0x%lx, offset = 0x%lx",
				brd->name, (tmp.access.cmd == NM6408_READ) ? "read" : "write",
				8 * tmp.access.width, filp, brd->indx, xfer->cluster_indx, iomem,
				axi_addr, axi_offset);

		tmp.access.offset = axi_offset;

		spin_lock_irqsave(&brd->lock, flags);
		if (brd->irq <= 0) {
			spin_unlock_irqrestore(&brd->lock, flags);

			rv = -EIO;
			break;
		}
		spin_unlock_irqrestore(&brd->lock, flags);

		rv = copy_to_user((void __user *)argp, &tmp.access, sizeof(tmp.access));
		if (rv) {
			rv = -EFAULT;
			break;
		}

		break;
	case NM6408_CFG_WIND:
		rv = copy_from_user(&tmp.wind, (void __user *)argp, sizeof(tmp.wind));
		if (rv) {
			rv = -EFAULT;
			break;
		}

		if (NM6408_WRITE == tmp.wind.cmd) {
			if ((tmp.wind.cluster < NM6408_PC0) || (tmp.wind.cluster > NM6408_CCPU)) {
				rv = -EINVAL;
				break;
			}

			xfer->cluster_indx = tmp.wind.cluster;
			axi_addr = tmp.wind.addr;

			if (axi_addr > (xfer->cluster_limit[xfer->cluster_indx] -
			    xfer->cluster_base[xfer->cluster_indx]))
				axi_addr = xfer->cluster_limit[xfer->cluster_indx] -
					   xfer->cluster_base[xfer->cluster_indx];

			axi_size = brd->region_len[NM6408_BAR_MEM];
			axi_addr &= ~(axi_size - 1);

			xfer->cluster_offset[xfer->cluster_indx] = axi_addr;
		}
		else if (NM6408_READ == tmp.wind.cmd) {
			axi_addr = xfer->cluster_offset[xfer->cluster_indx];
			axi_size = brd->region_len[NM6408_BAR_MEM];

			tmp.wind.cluster = xfer->cluster_indx;
		}
		else {
			rv = -EINVAL;
			break;
		}

		nm6408_print_dbg("%s: wind %s - filp = %lx, board = %d, cluster = %d,"
				" local axi address = 0x%lx, region size = 0x%lx",
				brd->name, (tmp.wind.cmd == NM6408_READ)?"get":"set", filp,
				brd->indx, xfer->cluster_indx, axi_addr, axi_size);

		tmp.wind.addr = axi_addr;
		tmp.wind.size = axi_size;

		rv = copy_to_user((void __user *)argp, &tmp.wind, sizeof(tmp.wind));
		if (rv) {
			rv = -EFAULT;
			break;
		}

		break;
	case NM6408_DO_INIT:
		if (brd->is_alive) {
			rv = 0;
			break;
		}

		spin_lock_irqsave(&brd->lock, flags);
		if (brd->irq <= 0) {
			spin_unlock_irqrestore(&brd->lock, flags);

			rv = -EIO;
			break;
		}
		brd->irq = -1;
		spin_unlock_irqrestore(&brd->lock, flags);

		rv = pci_read_config_dword(
				brd->pdev,
				PCI_BASE_ADDRESS_0 + 4*brd->region_indx[NM6408_BAR_CTRL],
				&bar_ctrl
		);

		if (rv || ((bar_ctrl & PCI_BASE_ADDRESS_MEM_MASK) !=
		    (brd->bar_ctrl & PCI_BASE_ADDRESS_MEM_MASK))) {
			printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 07).\n", brd->name);
			rv = -EIO;
			break;
		}
		else {
			u32 tmp32;

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_BOARD_READY;
			tmp32 = ioread32(iomem);

			if (tmp32 == 0x5A5A5A5A)
				goto nm6408_init_exit;
#if 0
			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_FW_CORESTART;
			tmp32 = ioread32(iomem);
#endif
			if (tmp32 != 0) {
				spin_lock_irqsave(&brd->lock, flags);
#if LEGACY
					brd->irq = brd->pdev->irq;
#else
					brd->irq = pci_irq_vector(brd->pdev, 0);
#endif
				spin_unlock_irqrestore(&brd->lock, flags);

				printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 08).\n", brd->name);

				rv = -EBUSY;
				break;
			}
		}

		rv = copy_from_user(&tmp.init_code, (void __user *)argp, sizeof(tmp.init_code));
		if (rv) {
			spin_lock_irqsave(&brd->lock, flags);
#if LEGACY
				brd->irq = brd->pdev->irq;
#else
				brd->irq = pci_irq_vector(brd->pdev, 0);
#endif
			spin_unlock_irqrestore(&brd->lock, flags);

			rv = -EFAULT;
			break;
		}

		rv = 0;

		do {
			void __iomem	*iomagic, *iodesc;
			u32		tmp32;

			iomagic = brd->region_virt[NM6408_BAR_CTRL] + CCPU_CCMB_FW_MAGIC;
			axi_addr = ioread32(iomagic + offsetof(struct nm6408_magic, magic_addr));
#if 0
			if (axi_addr != CCPU_REGS_FW_CORESTART) {
				spin_lock_irqsave(&brd->lock, flags);
#if LEGACY
					brd->irq = brd->pdev->irq;
#else
					brd->irq = pci_irq_vector(brd->pdev, 0);
#endif
				spin_unlock_irqrestore(&brd->lock, flags);

				rv = -EIO;
				break;
			}
#endif
			tmp32 = ioread32(iomagic + offsetof(struct nm6408_magic, data_point));
			iodesc = brd->region_virt[NM6408_BAR_CTRL] + tmp32;

			axi_size = ioread32(iomagic + offsetof(struct nm6408_magic, load_limit));
			axi_offset = ioread32(iomagic + offsetof(struct nm6408_magic, load_begin));

			tmp32 = axi_size - axi_offset;
			axi_size = ((tmp.init_code.size + 3) >> 2) << 2;

			if ((NULL == tmp.init_code.data) || (0 == tmp.init_code.size) || (axi_size > tmp32)) {
				spin_lock_irqsave(&brd->lock, flags);
#if LEGACY
					brd->irq = brd->pdev->irq;
#else
					brd->irq = pci_irq_vector(brd->pdev, 0);
#endif
				spin_unlock_irqrestore(&brd->lock, flags);

				rv = -EINVAL;
				break;
			}

			iowrite32(axi_size, iodesc + offsetof(struct nm6408_desc, image) + 4);
			iowrite32(axi_offset, iodesc + offsetof(struct nm6408_desc, image) + 0);
			iowrite32(axi_offset, iodesc + offsetof(struct nm6408_desc, entry));

			tmp32 = tmp.init_code.crc32;
			iowrite32(tmp32, iodesc + offsetof(struct nm6408_desc, crc));
		} while (0);

		if (rv)
			break;

		buffer = kmalloc(axi_size, GFP_KERNEL);
		if (!buffer) {
			spin_lock_irqsave(&brd->lock, flags);
#if LEGACY
				brd->irq = brd->pdev->irq;
#else
				brd->irq = pci_irq_vector(brd->pdev, 0);
#endif
			spin_unlock_irqrestore(&brd->lock, flags);

			rv = -ENOMEM;
			break;
		}

		rv = copy_from_user(buffer, (void __user *)tmp.init_code.data, tmp.init_code.size);
		if (rv) {
			spin_lock_irqsave(&brd->lock, flags);
#if LEGACY
				brd->irq = brd->pdev->irq;
#else
				brd->irq = pci_irq_vector(brd->pdev, 0);
#endif
			spin_unlock_irqrestore(&brd->lock, flags);

			kfree(buffer);

			rv = -EFAULT;
			break;
		}

		iomem = brd->region_virt[NM6408_BAR_CTRL] + axi_offset;
		axi_size = 0;

		while (axi_size < tmp.init_code.size) {
			iowrite32(*((u32 *)(buffer + axi_size)), iomem);

			iomem += 4;
			axi_size+=4;
		}

		mb(); // nead read after write ???
		kfree(buffer);

		iomem = brd->region_virt[NM6408_BAR_CTRL] + axi_addr; // CCPU_REGS_FW_CORESTART
		iowrite32(~0, iomem);
		iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_BOARD_READY;

		i = 1000;

		while (0xA5A5A5A5 != ioread32(iomem)) {
			if (!i)
				break;

			msleep(1);
			i--;
		}

		if (0 == i) {
			printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 09).\n", brd->name);
			rv = -EIO;
			break;
		}

		iowrite32(0x5A5A5A5A, iomem);

#if defined(CONFIG_E2K)
		ioread32(iomem); // barrier to ensure write
#endif

		if (0x5A5A5A5A != ioread32(iomem)) {
			printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 10).\n", brd->name);
			rv = -EIO;
			break;
		}

nm6408_init_exit:
		nm6408_print_dbg("%s init - filp = %lx, board = %d, cluster = %d, local axi addr = 0x%lx, global axi addr = 0x%lx,"
				 " global axi offset = 0x%lx, local region size = 0x%lx, global region size = 0x%lx",
				 brd->name, filp, brd->indx, xfer->cluster_indx, xfer->cluster_offset[xfer->cluster_indx],
				 xfer->str_pos, xfer->cur_pos, brd->region_len[NM6408_BAR_MEM], NM6408_INBOUND_AXI_SIZE);

		pci_save_state(brd->pdev);

		iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK;
		iowrite32(WDT_UNLOCK_TEMPLATE, iomem);

#if defined(CONFIG_E2K)
		ioread32(iomem); // barrier to ensure write
#endif

		if (ioread32(iomem) == 0) {
			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_RIS;
			if (ioread32(iomem) & WDT_INT_STATUS) {
				printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 11).\n",
				brd->name);

				rv = -EIO;
				break;
			}

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CTRL;
			iowrite32(0x0, iomem);

			mb();

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOAD;
			iowrite32(WDT_LOAD_1_SEC, iomem);

			mb();

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CLR;
			iowrite32(WDT_INT_STATUS, iomem);

			mb();

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CTRL;
			iowrite32(WDT_CTRL_INTEN, iomem);
		}
		else {
			printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 12).\n",
			brd->name);

			rv = -EIO;
			break;
		}

		iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK;
		iowrite32(~WDT_UNLOCK_TEMPLATE, iomem);

		spin_lock_irqsave(&brd->lock, flags);
#if LEGACY
			brd->irq = brd->pdev->irq;
#else
			brd->irq = pci_irq_vector(brd->pdev, 0);
#endif
		spin_unlock_irqrestore(&brd->lock, flags);

		brd->timer.expires = jiffies + (NM6408_XFER_TIMEOUT * HZ) / 1000;

		add_timer(&brd->timer);
		brd->is_alive = 1;

		rv = 0;
		break;
	case NM6408_DO_RESET:
		spin_lock_irqsave(&brd->lock, flags);

		if (0 == brd->irq) {
			spin_unlock_irqrestore(&brd->lock, flags);
			msleep(100);
			spin_lock_irqsave(&brd->lock, flags);
		}

		if (brd->irq < 0) {
			struct pci_dev  *port = brd->pdev->bus->self;
			u16		val;

			spin_unlock_irqrestore(&brd->lock, flags);

			pci_cfg_access_lock(brd->pdev);
			msleep(1000);

			pci_read_config_word(port, PCI_BRIDGE_CONTROL, &val);
			val |= (PCI_BRIDGE_CTL_BUS_RESET);
			pci_write_config_word(port, PCI_BRIDGE_CONTROL, val);

			msleep(1000);

			pci_read_config_word(port, PCI_BRIDGE_CONTROL, &val);
			val &= ~(PCI_BRIDGE_CTL_BUS_RESET);
			pci_write_config_word(port, PCI_BRIDGE_CONTROL, val);

			msleep(1000);
			pci_cfg_access_unlock(brd->pdev);
		}
		else {
			brd->irq = -1;
			spin_unlock_irqrestore(&brd->lock, flags);
		}

		brd->is_alive = 0;
		del_timer_sync(&brd->timer);

		rv = pci_read_config_dword(
				brd->pdev,
				PCI_BASE_ADDRESS_0 + 4*brd->region_indx[NM6408_BAR_CTRL],
				&bar_ctrl
		);

		if (!rv && ((bar_ctrl & PCI_BASE_ADDRESS_MEM_MASK) ==
		    (brd->bar_ctrl & PCI_BASE_ADDRESS_MEM_MASK))) {
			u32 tmp32;

			pci_save_state(brd->pdev);

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK;
			iowrite32(WDT_UNLOCK_TEMPLATE, iomem);

#if defined(CONFIG_E2K)
			ioread32(iomem); // barrier to ensure write
#endif

			if (ioread32(iomem) == 0) {
				iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CTRL;
				iowrite32(0x0, iomem);

				mb();

				iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOAD;
				iowrite32(WDT_LOAD_1_SEC, iomem);

				mb();

				iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CLR;
				iowrite32(WDT_INT_STATUS, iomem);

			}
			else {
				spin_lock_irqsave(&brd->lock, flags);
#if LEGACY
					brd->irq = brd->pdev->irq;
#else
					brd->irq = pci_irq_vector(brd->pdev, 0);
#endif
				spin_unlock_irqrestore(&brd->lock, flags);

				printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 13).\n",
				brd->name);

				rv = -EIO;
				break;
			}

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK;
			iowrite32(~WDT_UNLOCK_TEMPLATE, iomem);			

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_BOARD_READY; // CCPU_REGS_FW_CORESTART
			tmp32 = ioread32(iomem);

			if (tmp32 == 0)
				goto nm6408_reset_exit;
#if 0
			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_BOARD_READY;
			tmp32 = ioread32(iomem);
#endif
			if (tmp32 != 0x5A5A5A5A) {
				spin_lock_irqsave(&brd->lock, flags);
#if LEGACY
					brd->irq = brd->pdev->irq;
#else
					brd->irq = pci_irq_vector(brd->pdev, 0);
#endif
				spin_unlock_irqrestore(&brd->lock, flags);

				printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 14).\n",
				brd->name);

				rv = -EBUSY;
				break;
			}

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_SWI_COMMAND;
			iowrite32(SWI_BOARD_RESET, iomem);

			mb(); // may be need read, race on PCIe ???

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_GIC_SGIR;
			iowrite32(GIC_SGIR_SWI0, iomem);

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_SWI_COMMAND;
			i = 1000;

			while (ioread32(iomem) != 0) {
				if (!i)
					break;

				msleep(1);
				i--;
			}

			if (ioread32(iomem) != 0) {
				iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK;
				iowrite32(WDT_UNLOCK_TEMPLATE, iomem);

#if defined(CONFIG_E2K)
				ioread32(iomem); // barrier to ensure write
#endif

				if (ioread32(iomem) == 0) {
					iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_RIS;
					if (ioread32(iomem) & WDT_INT_STATUS) {
						printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 15).\n",
						brd->name);

						rv = -EIO;
						break;
					}

					iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CTRL;
					iowrite32(0x0, iomem);

					mb();

					iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOAD;
					iowrite32(WDT_LOAD_1_SEC, iomem);

					mb();

					iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CLR;
					iowrite32(WDT_INT_STATUS, iomem);

					mb();

					iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CTRL;
					iowrite32(WDT_CTRL_INTEN | WDT_CTRL_RSTEN, iomem);

					mb();

					iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_MIS;
					i = 10;

					rv = 0;

					while (!(ioread32(iomem) & WDT_INT_STATUS)) {
						if (!i) {
							iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CTRL;
							iowrite32(0x0, iomem);

							mb();

							iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CLR;
							iowrite32(WDT_INT_STATUS, iomem);

							printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 16).\n",
							brd->name);

							rv = -EIO;
							break;
						}

						msleep(100);
						i--;
					}

					if (rv)
						break;
				}
				else  {
					printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 17).\n",
					brd->name);

					rv = -EIO;
					break;
				}

				iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK;
				iowrite32(~WDT_UNLOCK_TEMPLATE, iomem);
			}

			rv = 0;

			do  {
				struct pci_dev	*port = brd->pdev->bus->self;
				u16		val;

				pci_cfg_access_lock(brd->pdev);

				pci_read_config_word(port, port->pcie_cap + PCI_EXP_LNKCTL, &val);
				val |= (PCI_EXP_LNKCTL_LD);
				pci_write_config_word(port, port->pcie_cap + PCI_EXP_LNKCTL, val);

				msleep(2000);

				pci_read_config_word(port, port->pcie_cap + PCI_EXP_LNKCTL, &val);
				val &= ~(PCI_EXP_LNKCTL_LD);
				val |= (PCI_EXP_LNKCTL_RL);
				pci_write_config_word(port, port->pcie_cap + PCI_EXP_LNKCTL, val);

				i = 10;

				while (1) {
					pci_read_config_word(port, port->pcie_cap + PCI_EXP_LNKSTA, &val);

					if (i < 0)
						break;

					if ((val & PCI_EXP_LNKSTA_LT) == 0) {
						val &= PCI_EXP_LNKSTA_CLS;

						if (val == PCI_EXP_LNKSTA_CLS_5_0GB)
							break;
						else if (i == 0)
							break;

						pci_read_config_word(port, port->pcie_cap + PCI_EXP_LNKCTL, &val);
						val |= (PCI_EXP_LNKCTL_RL | PCI_EXP_LNKCTL_ES);
						pci_write_config_word(port, port->pcie_cap + PCI_EXP_LNKCTL, val);
					}

					msleep(100);
					i--;
				}

				if ((val & PCI_EXP_LNKSTA_LT) != 0) {
					printk(KERN_WARNING "%s: retrain link error.\n", brd->name);
					rv = -EIO;
				}

				pci_cfg_access_unlock(brd->pdev);
			} while (0);

			if (rv)
				break;
		}

		i = 10;

		do {
			msleep(100);
			rv = pci_read_config_word(brd->pdev, PCI_VENDOR_ID, &vendor_id);
			i--;

			if (!i)
				break;
		} while (rv || (NM6408_VENDOR_ID != vendor_id));

		if (rv || (NM6408_VENDOR_ID != vendor_id)) {
			printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 18).\n", brd->name);
			rv = -EIO;
			break;
		}

		pci_restore_state(brd->pdev);

		for (i = 100; i > 0; i--) {
			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_CRG_CRGSR;

			if (CRG_CRGSR_SWRSTACK & ioread32(iomem)) {
				i = 0;
				break;
			}

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_CRG_CRGCR;

			if (CRG_CRGCR_SWRSTREQ & ioread32(iomem)) {
				msleep(10);
				continue;
			}

			break;
		}

		if (0 == i) {
			printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 19).\n", brd->name);
			rv = -EIO;
			break;
		}

		rv = 0;

		do {
			u32 tmp32;

			iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_CCMB_FW_MAGIC;
			tmp32 = ioread32(iomem + offsetof(struct nm6408_magic, magic_addr));
#if 0
			if (tmp32 != CCPU_REGS_FW_CORESTART) {
				printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 20).\n", brd->name);
				rv = -EIO;
				break;
			}
#endif
			tmp32 = ioread32(iomem + offsetof(struct nm6408_magic, data_point));
			iomem = brd->region_virt[NM6408_BAR_CTRL] + tmp32;
			tmp32 = ioread32(iomem + offsetof(struct nm6408_desc, is_ready));

			i = 1000;

			while (!tmp32) {
				if (i)
					msleep(1);
				else
					break;

				tmp32 = ioread32(iomem + offsetof(struct nm6408_desc, is_ready));
				i--;
			}

			if (!tmp32) {
				printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 21).\n", brd->name);
				rv = -EIO;
				break;
			}

			iowrite32(0, iomem + offsetof(struct nm6408_desc, is_ready));

#if defined(CONFIG_E2K)
			ioread32(iomem + offsetof(struct nm6408_desc, is_ready)); // barrier to ensure write
#endif

			if (ioread32(iomem + offsetof(struct nm6408_desc, is_ready))) {
				printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 22).\n", brd->name);
				rv = -EIO;
				break;
			}
		} while (0);

		if (rv)
			break;

nm6408_reset_exit:
		nm6408_print_dbg("%s reset - filp = %lx, board = %d, cluster = %d, local axi addr = 0x%lx, global axi addr = 0x%lx,"
				 " global axi offset = 0x%lx, local region size = 0x%lx, global region size = 0x%lx",
				 brd->name, filp, brd->indx, xfer->cluster_indx, xfer->cluster_offset[xfer->cluster_indx],
				 xfer->str_pos, xfer->cur_pos, brd->region_len[NM6408_BAR_MEM], NM6408_INBOUND_AXI_SIZE);

		spin_lock_irqsave(&brd->lock, flags);
#if LEGACY
			brd->irq = brd->pdev->irq;
#else
			brd->irq = pci_irq_vector(brd->pdev, 0);
#endif
		spin_unlock_irqrestore(&brd->lock, flags);

		rv = 0;
		break;
	default:
		rv = -ENOTTY;
		break;
	}

	up(&brd->semap);

	nm6408_print_dbg("%s: end - pid = %d, tgid = 0x%x, filp = %lx, board = %d, command = 0x%x",
			brd->name, current->pid, current->tgid, filp, brd->indx, cmd);

	return rv;
}

#if defined(__LP64__) || defined(_LP64) // (__SIZEOF_POINTER__ == 8)
static loff_t nm6408_llseek(struct file *filp, loff_t offset, int whence)
{
	struct nm6408_xfer	*xfer = filp->private_data;
	struct nm6408_brd	*brd = xfer->brd;
	loff_t			axi_addr, axi_offset, axi_size;
	size_t			file_pos;
	int			i;

	nm6408_print_dbg("%s: begin - pid = %d, tgid = 0x%x, filp = %lx, board = %d",
			 brd->name, current->pid, current->tgid, filp, brd->indx);

	if (down_interruptible(&brd->semap)) {
		printk(KERN_INFO "%s: interrupt - pid = %d, tgid = 0x%x, board = %d.\n",
			brd->name, current->pid, current->tgid, brd->indx);

		return -ERESTARTSYS;
	}

	if (!brd->is_alive) {
		printk(KERN_WARNING "%s: need reload driver of board (case - 23).\n", brd->name);

		axi_offset = -EPERM;
		goto err_llseek_end;
	}

	if (offset & 0x3) {
		axi_offset = -EINVAL;
		goto err_llseek_end;
	}

	for (i = 0; i < NM6408_INBOUND_AXI_NMB; i++) {
		if ((xfer->str_pos >= xfer->cluster_base[i]) &&
		    (xfer->str_pos <= xfer->cluster_limit[i]))
			break;
	}

	if (i == NM6408_INBOUND_AXI_NMB) {
		axi_offset = -EFAULT;
		goto err_llseek_end;
	}

	axi_size = (loff_t)xfer->cluster_limit[i];
	axi_offset = (loff_t)xfer->cur_pos;
	axi_addr = (loff_t)xfer->cluster_base[i];

	switch (whence) {
	case SEEK_SET:
		if (offset < 0)
			axi_offset = axi_addr;
		else if (offset < NM6408_INBOUND_AXI_SIZE)
			axi_offset = axi_addr + offset;
		else
			axi_offset = offset; // change cluster

		break;
	case SEEK_CUR:
		if ((axi_offset + offset) < axi_addr)
			axi_offset = axi_addr;
		else if ((axi_offset + offset) < (axi_size + 1))
			axi_offset = axi_offset + offset;
		else
			axi_offset = axi_size & ~0x3;

		break;
	case SEEK_END:
		if (((axi_size + 1) + offset) < axi_addr)
			axi_offset = axi_addr;
		else if (offset < 0)
			axi_offset = (axi_size + 1) + offset;
		else
			axi_offset = axi_size & ~0x3;

		break;
	default:
		axi_offset = -EINVAL;
		goto err_llseek_end;
	}

	file_pos = (size_t)axi_offset;

	for (i = 0; i < NM6408_INBOUND_AXI_NMB; i++) {
		if ((file_pos >= xfer->cluster_base[i]) &&
		    (file_pos <= xfer->cluster_limit[i]))
			break;
	}

	if (i == NM6408_INBOUND_AXI_NMB) {
		axi_offset = -EFAULT;
		goto err_llseek_end;
	}

	xfer->str_pos = xfer->cluster_base[i];

	nm6408_print_dbg("%s: filp = %lx, board = %d, cluster = %d, global axi address = 0x%lx,"
			" prev file offset = 0x%lx, new file offest = 0x%lx", brd->name, filp,
			 brd->indx, i, xfer->cluster_base[i], xfer->cur_pos, file_pos);

	xfer->cur_pos = file_pos;

err_llseek_end:

	up(&brd->semap);

	nm6408_print_dbg("%s: end - pid = %d, tgid = 0x%x, filp = %lx, board = %d",
			brd->name, current->pid, current->tgid, filp, brd->indx);

	return axi_offset;
}
#else
static loff_t nm6408_llseek(struct file *filp, loff_t offset, int whence)
{
	struct nm6408_xfer	*xfer = filp->private_data;
	struct nm6408_brd	*brd = xfer->brd;
	size_t			cur_offs = offset;
	int			i, err = 0;

	nm6408_print_dbg("%s: begin - pid = %d, tgid = 0x%x, filp = %lx, board = %d",
			 brd->name, current->pid, current->tgid, filp, brd->indx);

	if (down_interruptible(&brd->semap)) {
		printk(KERN_INFO "%s: interrupt - pid = %d, tgid = 0x%x, board = %d.\n",
				 brd->name, current->pid, current->tgid, brd->indx);

		return -ERESTARTSYS;
	}

	if (!brd->is_alive) {
		printk(KERN_WARNING "%s: need reload driver of board (case - 24).\n", brd->name);

		err = -EPERM;
		goto err_llseek_end;
	}

	if ((cur_offs & 0x3) || (whence != SEEK_SET)) {
 		err = -EINVAL;
 		goto err_llseek_end;
	}

	if (cur_offs < NM6408_INBOUND_AXI_SIZE)
		cur_offs = xfer->str_pos + cur_offs;

 	for (i = 0; i < NM6408_INBOUND_AXI_NMB; i++) {
		if ((cur_offs >= xfer->cluster_base[i]) &&
		    (cur_offs <= xfer->cluster_limit[i]))
			break;
	}

	if (i == NM6408_INBOUND_AXI_NMB) {
		err = -EFAULT;
		goto err_llseek_end;
	}

	xfer->str_pos = xfer->cluster_base[i];

	nm6408_print_dbg("%s: filp = %lx, board = %d, cluster = %d, global axi address = 0x%lx,"
			 " prev file offset = 0x%lx, new file offest = 0x%lx", brd->name, filp,
			 brd->indx, i, xfer->str_pos, xfer->cur_pos, cur_offs);

	xfer->cur_pos = cur_offs;

err_llseek_end:

	up(&brd->semap);

	nm6408_print_dbg("%s: end - pid = %d, tgid = 0x%x, filp = %lx, board = %d",
			 brd->name, current->pid, current->tgid, filp, brd->indx);

	if (!err)
		return (xfer->cur_pos - xfer->str_pos);
	else
		return err;
}
#endif

static void nm6408_rd32_buff(void *buff, void __iomem *iomem, size_t cnt)
{
#if 1
	while (cnt) {
		*(u32 *)buff = ioread32(iomem);

		buff += 4;
		iomem += 4;
		cnt -= 4;
	}
#else
	memcpy_fromio(buff, iomem, cnt);
#endif
	return;
}

static void nm6408_wr32_buff(void *buff, void __iomem *iomem, size_t cnt)
{
#if 1
	while (cnt) {
		iowrite32(*(u32 *)buff, iomem);

		buff += 4;
		iomem += 4;
		cnt -= 4;
	}
#else
	memcpy_toio(iomem, buff, cnt);
#endif
	return;
}

static int nm6408_rw_page(struct nm6408_xfer *xfer, void *buff, size_t cnt)
{
	struct nm6408_brd	*brd = xfer->brd;
	unsigned long		flags;
	void __iomem		*iomem;

	spin_lock_irqsave(&brd->lock, flags);
	if (brd->irq <= 0) {
		spin_unlock_irqrestore(&brd->lock, flags);
		return -EIO;
	}

	iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_EP_INREGIONS +
		8 * brd->region_indx[NM6408_BAR_MEM];

	iowrite32(xfer->cur_pos & ~(brd->region_len[NM6408_BAR_MEM] - 1), iomem);

#if defined(CONFIG_E2K)
	ioread32(iomem); // barrier to ensure write
#endif

	if (ioread32(iomem) != (xfer->cur_pos & ~(brd->region_len[NM6408_BAR_MEM] - 1))) {
		brd->irq = -1;
		spin_unlock_irqrestore(&brd->lock, flags);

		printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 25).\n", brd->name);
		return -EIO;
	}
	spin_unlock_irqrestore(&brd->lock, flags);

	iomem = brd->region_virt[NM6408_BAR_MEM] + (xfer->cur_pos & (brd->region_len[NM6408_BAR_MEM] - 1));

	if (NM6408_READ == xfer->rw) {
		nm6408_print_dbg("%s: copy data from board to host -"
				" file offset = 0x%lx, size = 0x%lx",
				brd->name, xfer->cur_pos, cnt);

		nm6408_rd32_buff(buff, iomem, cnt);
	}
	else {
		nm6408_print_dbg("%s: copy data from host to board -"
				" file offset = 0x%lx, size = 0x%lx",
				brd->name, xfer->cur_pos, cnt);

		nm6408_wr32_buff(buff, iomem, cnt);
	}

	xfer->cur_pos += cnt;

	return 0;
}

static ssize_t nm6408_xfer(struct nm6408_xfer *xfer, char __user *buff, size_t cnt)
{
	struct nm6408_brd	*brd = xfer->brd;
	unsigned long		flags;
	struct page		**pages = NULL;
	int			fpage = 0, lpage = 0;
	unsigned int		npages = 0;
	size_t			delta;
	int			i, rv = 0;

	spin_lock_irqsave(&brd->lock, flags);
	if (brd->irq <= 0) {
		spin_unlock_irqrestore(&brd->lock, flags);

		rv = -EIO;
		goto err_xfer_unlock;
	}
	spin_unlock_irqrestore(&brd->lock, flags);

	if (xfer->cur_pos > xfer->str_pos)
		delta = xfer->cur_pos - xfer->str_pos;
	else
		delta = xfer->str_pos - xfer->cur_pos;

	if (!cnt || (delta >= NM6408_INBOUND_AXI_SIZE)) {
		cnt = 0;
		goto err_xfer_unlock;
	}

	if (cnt > (NM6408_INBOUND_AXI_SIZE - delta))
		cnt = NM6408_INBOUND_AXI_SIZE - delta;

	if (((unsigned long)buff & 0x3) || (cnt & 0x3)) {
		rv = -EINVAL;
		goto err_xfer_unlock;
	}

	rv = nm6408_check_axi_addr(xfer->cur_pos, xfer->str_pos, cnt, xfer->rw);
	if (rv) {
		rv = -EINVAL;
		goto err_xfer_unlock;
	}

#if defined (RHEL_RELEASE_CODE)
#if RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(8,0)
	if (!access_ok(buff, cnt))
#else
	if (!access_ok((NM6408_READ == xfer->rw) ? VERIFY_WRITE : VERIFY_READ, buff, cnt))
#endif
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 0, 0)
	if (!access_ok((NM6408_READ == xfer->rw) ? VERIFY_WRITE : VERIFY_READ, buff, cnt))
#else
	if (!access_ok(buff, cnt))
#endif
#endif
	{
		rv = -EFAULT;
		goto err_xfer_unlock;
	}

	npages = (((unsigned long)buff & (PAGE_SIZE - 1)) + cnt + (PAGE_SIZE - 1)) >> PAGE_SHIFT;
	pages = kmalloc(npages * sizeof(struct page *), GFP_KERNEL);
	if (NULL == pages) {
		npages = 0;

		rv = -ENOMEM;
		goto err_xfer_unlock;
	}

	nm6408_print_dbg("%s: %s - user address = %lx, count = 0x%lx, pages = %u",
			brd->name, (NM6408_READ == xfer->rw) ? "read" : "write",
			buff, cnt, npages);

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 8, 1)
	down_read(&current->mm->mmap_sem);
#else
	mmap_read_lock(current->mm);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 6, 1)
#if LEGACY
	i = get_user_pages(
			   current, current->mm,
			   (unsigned long)buff & PAGE_MASK, npages,
			   (NM6408_READ == xfer->rw) ? FOLL_WRITE | FOLL_GET : FOLL_GET,
			   pages, NULL
	);
#else
	i = get_user_pages(
			   current, current->mm,
			   (unsigned long)buff & PAGE_MASK, npages,
			   (NM6408_READ == xfer->rw) ? 1 : 0, 0,
			   pages, NULL
	);
#endif
#else
	i = get_user_pages(
			   (unsigned long)buff & PAGE_MASK, npages,
			   (NM6408_READ == xfer->rw) ? FOLL_WRITE | FOLL_GET : FOLL_GET,
			   pages, NULL
	);
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 8, 1)
	up_read(&current->mm->mmap_sem);
#else
	mmap_read_unlock(current->mm);
#endif

	if (i != npages) {
		if (i < 0)
			npages = 0;
		else
			npages = i;

		rv = -EFAULT;
		goto err_xfer_unlock;
	}

	if ((1 == npages) || ((unsigned long)buff & (PAGE_SIZE - 1)))
		fpage = 1;

	if ((npages > 1) && (((unsigned long)buff + cnt) & (PAGE_SIZE - 1)))
		lpage = 1;

	if (0 != (npages - (fpage + lpage))) {
		if (sg_alloc_table(&xfer->sc_table, npages - (fpage + lpage), GFP_KERNEL)) {
			rv = -ENOMEM;
			goto err_xfer_unlock;
		}

		for_each_sg(xfer->sc_table.sgl, xfer->sc_entry, npages - (fpage + lpage), i)
			sg_set_page(xfer->sc_entry, pages[i + fpage], PAGE_SIZE, 0);

		xfer->sc_entry = xfer->sc_table.sgl;
		xfer->sc_nents = dma_map_sg(&brd->pdev->dev, xfer->sc_table.sgl, npages - (fpage + lpage),
					    (NM6408_READ == xfer->rw) ? DMA_FROM_DEVICE : DMA_TO_DEVICE);
		if (!xfer->sc_nents) {
			rv = -EFAULT;
			goto err_xfer_free;
		}
	}

	if (fpage) {
		void *page_addr;
		size_t xfer_size;

		if (npages > 1)
			xfer_size = PAGE_SIZE - ((unsigned long)buff & (PAGE_SIZE - 1));
		else
			xfer_size = cnt;

		nm6408_print_dbg("%s: cpu copy page[%u] - page offset = 0x%lx , count = 0x%lx",
				brd->name, 0, (unsigned long)buff & (PAGE_SIZE - 1), xfer_size);

		page_addr = kmap(pages[0]);
		rv = nm6408_rw_page(xfer, page_addr + ((unsigned long)buff & (PAGE_SIZE - 1)), xfer_size);
		kunmap(pages[0]);

		if (rv)
			goto err_xfer_end;
	}

	if (0 != (npages - (fpage + lpage))) {
		spin_lock_irqsave(&brd->lock, flags);
		if (brd->irq <= 0) {
			xfer->stat = XFER_COMPLETE_ERROR;
			brd->xfer_cur = NULL;
		}
		else {
			xfer->stat = 0;
			brd->xfer_cur = xfer;
		}
		spin_unlock_irqrestore(&brd->lock, flags);

		nm6408_print_dbg("%s: dma copy pages from page[%u] to page[%u], page size = 0x%lx",
				brd->name, fpage, npages - lpage - 1, PAGE_SIZE);

		queue_work(brd->work_queue, &xfer->work);
		wait_event(brd->wait_queue, xfer->stat != 0);
		cancel_work_sync(&xfer->work);

		spin_lock_irqsave(&brd->lock, flags);
		if (xfer->stat != XFER_COMPLETE_SUCCESS) {
			if (xfer->stat == XFER_COMPLETE_TIMEOUT) {
				spin_unlock_irqrestore(&brd->lock, flags);
				msleep(100);
				spin_lock_irqsave(&brd->lock, flags);
			}

			spin_unlock_irqrestore(&brd->lock, flags);

			rv = -EIO;
			goto err_xfer_end;
		}
		spin_unlock_irqrestore(&brd->lock, flags);
	}

	if (lpage) {
		void *page_addr;

		nm6408_print_dbg("%s: cpu copy page[%u] - page offset = 0x%lx , count = 0x%lx",
				brd->name, npages - 1, 0, ((unsigned long)buff + cnt) & (PAGE_SIZE - 1));

		page_addr = kmap(pages[npages - 1]);
		rv = nm6408_rw_page(xfer, page_addr, ((unsigned long)buff + cnt) & (PAGE_SIZE - 1));
		kunmap(pages[npages - 1]);

		if (rv)
			goto err_xfer_end;
	}

err_xfer_end:

	if (0 != (npages - (fpage + lpage)))
		dma_unmap_sg(&brd->pdev->dev, xfer->sc_table.sgl, npages - (fpage + lpage),
			(NM6408_READ == xfer->rw) ? DMA_FROM_DEVICE : DMA_TO_DEVICE);

err_xfer_free:

	if (0 != (npages - (fpage + lpage)))
		sg_free_table(&xfer->sc_table);

err_xfer_unlock:

	for (i = 0; i < npages; i++) {
		if (NM6408_READ == xfer->rw)
			SetPageDirty(pages[i]);

		put_page(pages[i]);
	}

	if (NULL != pages)
		kfree(pages);

	if (!rv)
		return cnt;
	else
		return rv;
}

static ssize_t nm6408_read(struct file *filp, char __user *buff, size_t cnt, loff_t *offp)
{
	struct nm6408_xfer	*xfer = filp->private_data;
	struct nm6408_brd	*brd = xfer->brd;
	ssize_t			rv;

	nm6408_print_dbg("%s: begin - pid = %d, tgid = 0x%x, filp = %lx, board = %d",
			brd->name, current->pid, current->tgid, filp, brd->indx);

	if (down_interruptible(&brd->semap)) {
		printk(KERN_INFO "%s: interrupt - pid = %d, tgid = 0x%x, board = %d.\n",
			brd->name, current->pid, current->tgid, brd->indx);

		return -ERESTARTSYS;
	}

	if (!brd->is_alive) {
		printk(KERN_WARNING "%s: need reload driver of board (case - 26).\n", brd->name);

		rv = -EPERM;
		goto err_read_end;
	}

	xfer->dma_offset = 0;
	xfer->rw = NM6408_READ;

	rv = nm6408_xfer(xfer, buff, cnt);

err_read_end:

	up(&brd->semap);

	nm6408_print_dbg("%s: end - pid = %d, tgid = 0x%x, file = %lx, board = %d",
			brd->name, current->pid, current->tgid, filp, brd->indx);

	return rv;
}

static ssize_t nm6408_write(struct file *filp, const char __user *buff, size_t cnt, loff_t *offp)
{
	struct nm6408_xfer	*xfer = filp->private_data;
	struct nm6408_brd	*brd = xfer->brd;
	ssize_t			rv;

	nm6408_print_dbg("%s: begin - pid = %d, tgid = 0x%x, file = %lx, board = %d",
			brd->name, current->pid, current->tgid, filp, brd->indx);

	if (down_interruptible(&brd->semap)) {
		printk(KERN_INFO "%s: interrupt - pid = %d, tgid = 0x%x, board = %d.\n",
			brd->name, current->pid, current->tgid, brd->indx);

		return -ERESTARTSYS;
	}

	if (!brd->is_alive) {
		printk(KERN_WARNING "%s: need reload driver of board (case - 27).\n", brd->name);

		rv = -EPERM;
		goto err_write_end;
	}

	xfer->dma_offset = 0;
	xfer->rw = NM6408_WRITE;

	rv = nm6408_xfer(xfer, (char __user *)buff, cnt);

err_write_end:

	up(&brd->semap);

	nm6408_print_dbg("%s: end - pid = %d, tgid = 0x%x, file = %lx, board = %d",
			brd->name, current->pid, current->tgid, filp, brd->indx);

	return rv;
}

static void nm6408_vma_open(struct vm_area_struct *vma)
{
	struct nm6408_brd *brd = vma->vm_private_data;

	atomic_inc(&brd->count);

	return;
}

static void nm6408_vma_close(struct vm_area_struct *vma)
{
	struct nm6408_brd *brd = vma->vm_private_data;

	atomic_dec(&brd->count);

	return;
}

static struct vm_operations_struct nm6408_vma_ops = {
	.open = nm6408_vma_open,
	.close = nm6408_vma_close,
};

static int nm6408_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct nm6408_xfer	*xfer = filp->private_data;
	struct nm6408_brd	*brd = xfer->brd;
	void __iomem		*iomem;
	resource_size_t		psize;
	phys_addr_t		paddr;
	u32			base, offset;
	int			i, rv = 0;

	if (down_interruptible(&brd->semap))
		return -ERESTARTSYS;

	if (!brd->is_alive || (brd->region_count < NM6408_BAR_NMB)) {
		rv = -EPERM;
		goto err_mmap_end;
	}

	if (atomic_read(&brd->count)) {
		rv = -EBUSY;
		goto err_mmap_end;
	}

	iomem = brd->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_EP_INREGIONS +
		8 * brd->region_indx[NM6408_BAR_MMAP];

	for (i = 0; i < NM6408_INBOUND_AXI_NMB; i++) {
		if ((xfer->cluster_base[i] <= (vma->vm_pgoff << PAGE_SHIFT)) &&
		    (xfer->cluster_limit[i] >= (vma->vm_pgoff << PAGE_SHIFT))) {
			base = vma->vm_pgoff << PAGE_SHIFT;
			base &= ~(brd->region_len[NM6408_BAR_MMAP] - 1);
			offset = (vma->vm_pgoff << PAGE_SHIFT) - base;

			break;
		}
	}
		
	if (i == NM6408_INBOUND_AXI_NMB) {
		rv = -EINVAL;
		goto err_mmap_end;
	}

	psize = brd->region_len[NM6408_BAR_MMAP] - offset;
	paddr = brd->region_phys[NM6408_BAR_MMAP] + offset;

	if ((vma->vm_end - vma->vm_start) > psize) {
		rv = -EFAULT;
		goto err_mmap_end;
	}

	iowrite32(base, iomem);

#if defined(CONFIG_E2K)
	ioread32(iomem); // barrier to ensure write
#endif

	if (base != ioread32(iomem)) {
		rv = -EIO;
		goto err_mmap_end;
	}

	vma->vm_flags |= (VM_IO | VM_DONTEXPAND | VM_DONTDUMP);
	vma->vm_page_prot = vm_get_page_prot(vma->vm_flags);
	vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);

	vma->vm_pgoff = (paddr >> PAGE_SHIFT);

	rv = io_remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
				vma->vm_end - vma->vm_start, vma->vm_page_prot);
	if (rv) {
		rv = -EAGAIN;
		goto err_mmap_end;
	}

	vma->vm_ops = &nm6408_vma_ops;
	vma->vm_private_data = brd;
	nm6408_vma_open(vma);

err_mmap_end:

	up(&brd->semap);
	return rv;
}

static const struct file_operations nm6408_fops = {
	.owner		= THIS_MODULE,
	.read		= nm6408_read,
	.write		= nm6408_write,
	.open		= nm6408_open,
	.release	= nm6408_release,
	.mmap		= nm6408_mmap,
	.llseek		= nm6408_llseek,
	.unlocked_ioctl	= nm6408_ioctl,
};

static int nm6408_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
	struct nm6408_brd	*nm6408;
	struct device		*dev = &pdev->dev;
	dev_t			node;
	void			*ptr;
	u32			tmp;
	int			i, rc;

	dev_info(dev, "device probe pdev=%p (%02x:%02x.%01x), [%04x, %04x] (rev %x).\n",
		pdev, pdev->bus->number, PCI_SLOT(pdev->devfn), PCI_FUNC(pdev->devfn),
		(int)pdev->vendor, (int)pdev->device, (int)pdev->revision);

	nm6408 = kzalloc(sizeof(struct nm6408_brd), GFP_KERNEL);
	if (!nm6408) {
		dev_err(dev, "memory allocation failed.\n");
		return -ENOMEM;
	}

	atomic_set(&nm6408->count, 0);
	nm6408->indx = -1;

	for (i = 0; i < NM6408_BRD_NUM; i++) {
		if (!test_and_set_bit(i, &nm6408_brds_indx)) {
			nm6408->indx = i;

			nm6408->name = kasprintf(GFP_KERNEL, "%s_%2.2d", NM6408_BRD_NAME, i);
			if (!nm6408->name) {
				dev_err(dev, "memory allocation failed.\n");
				rc = -ENOMEM;
				goto err_free_mem;
			}

			nm6408->bus_nmb = (unsigned int)pdev->bus->number;
			nm6408->dev_nmb = (unsigned int)PCI_SLOT(pdev->devfn);
			nm6408->func_nmb = (unsigned int)PCI_FUNC(pdev->devfn);

			break;
		}
	}

	if (nm6408->indx < 0) {
		dev_err(dev, "number of devices exceeded limit.\n");
		rc = -EPERM;
		goto err_free_mem;
	}

	rc = pci_enable_device_mem(pdev);
	if (rc) {
		dev_err(dev, "enable pci device failed.\n");
		goto err_free_mem;
	}

	nm6408->pdev = pdev;

	for (i = 0; i < NM6408_BAR_NMB; i++) {
		nm6408->region_virt[i] = 0;
		nm6408->region_indx[i] = -1;
	}

	nm6408->region_count = 0;

	for (i = 0; i < PCI_ROM_RESOURCE; i++) {
		if (!pci_resource_len(pdev, i))
			continue;

		if (!(pci_resource_flags(pdev, i) & IORESOURCE_MEM))
			continue;

		nm6408->region_len[nm6408->region_count] = pci_resource_len(pdev, i);
		nm6408->region_flags[nm6408->region_count] = pci_resource_flags(pdev, i);

		rc = pci_request_region(pdev, i, nm6408->name);
		if (rc) {
			dev_err(dev, "pci request region %d failed.\n", i);
			break;
		}

		nm6408->region_indx[nm6408->region_count] = i;
		nm6408->region_phys[nm6408->region_count] = pci_resource_start(pdev, i);

		if (!(pci_resource_flags(pdev, i) & IORESOURCE_PREFETCH)) {
			rc = pci_read_config_dword(pdev, PCI_BASE_ADDRESS_0 + 4 * i, &tmp);
			if (rc || (nm6408->region_count == NM6408_BAR_CTRL)) {
				if (rc) {
					dev_err(dev, "pci read BAR %d failed.\n", i);
					break;
				}

				nm6408->bar_ctrl = tmp;
			}

			nm6408->region_virt[nm6408->region_count] = pci_ioremap_bar(pdev, i);
			if (!nm6408->region_virt[nm6408->region_count]) {
				dev_err(dev, "pci ioremap region %d failed.\n", i);
				break;
			}

			dev_info(dev, "brd_indx = %d, BAR%d at %pR->0x%p.\n", nm6408->indx, i,
				&pdev->resource[i], nm6408->region_virt[nm6408->region_count]);
		}

		if (NM6408_INBOUND_AXI_SIZE & (nm6408->region_len[nm6408->region_count] - 1)) {
			dev_err(dev, "pci region %d not fit.\n", i);
			break;
		}

		nm6408->region_count++;

		if (nm6408->region_count == NM6408_BAR_NMB)
			break;
	}

	if (nm6408->region_count < NM6408_BAR_MMAP) {
		dev_err(dev, "pci probe failed.\n");
		rc = -EBUSY;
		goto err_pci_release_ioumap;
	}

	if ((CCPU_REG_AXI_ADDR + CCPU_REG_AXI_SIZE) > nm6408->region_len[NM6408_BAR_CTRL]) {
		dev_err(dev, "pci probe - not access to registers.\n");
		rc = -EPERM;
		goto err_pci_release_ioumap;
	}

	tmp = ioread32(nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_EP_INREGIONS +
			8 * nm6408->region_indx[NM6408_BAR_CTRL]);

	if (tmp != CCPU_INBOUND_AXI_BASE) {
		dev_err(dev, "pci probe - axi address not match.\n");
		rc = -EIO;
		goto err_pci_release_ioumap;
	}

	if (!dma_set_mask(dev, DMA_BIT_MASK(64))) {
		rc = dma_set_coherent_mask(dev, DMA_BIT_MASK(64));
		if (rc) {
			dev_err(dev, "unable set dma64 mode.\n");
			rc = -EIO;
			goto err_pci_release_ioumap;
		}
	}
	else if (!dma_set_mask(dev, DMA_BIT_MASK(32))) {
		rc = dma_set_coherent_mask(dev, DMA_BIT_MASK(32));
		if (rc) {
			dev_err(dev, "unable set dma32 mode.\n");
			rc = -EIO;
			goto err_pci_release_ioumap;
		}
	}
	else {
		dev_err(dev, "unable set dma mode.\n");
		rc = -EIO;
		goto err_pci_release_ioumap;
	}

	nm6408->desc_cpu_addr = dma_alloc_coherent(dev, PAGE_SIZE, &nm6408->desc_bus_addr, GFP_KERNEL);
	if (!nm6408->desc_cpu_addr) {
		dev_err(dev, "unable get dma buffer.\n");
		rc = -ENOMEM;
		goto err_pci_release_ioumap;
	}

	nm6408->desc_num = PAGE_SIZE / sizeof(struct dma_desc);
	sema_init(&nm6408->semap, 1);

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 15, 1)
	init_timer(&nm6408->timer);
	nm6408->timer.data = (unsigned long)nm6408;
	nm6408->timer.function = _check_timer;
#else
	timer_setup(&nm6408->timer, _check_timer, 0);
#endif

	spin_lock_init(&nm6408->lock);
	init_waitqueue_head(&nm6408->wait_queue);

	nm6408->work_queue = create_workqueue(nm6408->name);
	if (!nm6408->work_queue) {
		dev_err(dev, "unable create workqueue.\n");
		rc = -EPERM;
		goto err_dma_free_coherent;
	}

	cdev_init(&nm6408->cdev, &nm6408_fops);
	nm6408->cdev.owner = THIS_MODULE;

	node = MKDEV(nm6408_major, nm6408_minor + nm6408->indx);

	rc = cdev_add(&nm6408->cdev, node, 1);
	if (rc) {
		dev_err(dev, "chrdev add failed.\n");
		goto err_workqueue_destroy;
	}

	ptr = device_create(nm6408_class, dev, node, NULL, "%s", nm6408->name);
	if (IS_ERR(ptr)) {
		dev_err(dev, "device create failed.\n");
		rc = PTR_ERR(ptr);
		goto err_chrdev_delete;
	}

#if LEGACY
	nm6408->irq = pdev->irq;
#else
	switch (irq_type) {
	case 1:
		rc = pci_alloc_irq_vectors(pdev, 1, 1, PCI_IRQ_MSI);
		break;
	case 2:
		rc = pci_alloc_irq_vectors(pdev, 1, 1, PCI_IRQ_LEGACY);
		break;
	default:
		rc = -EINVAL; // PCI_IRQ_ALL_TYPES
		break;
	}

	if (rc < 0) {
		dev_err(dev, "unable alloc irq vectors.\n");
		goto err_chrdev_destroy;
	}

	nm6408->irq = pci_irq_vector(pdev, 0);
#endif

	pci_set_drvdata(pdev, nm6408);

	rc = request_irq(nm6408->irq, nm6408_irq_handler, IRQF_SHARED, nm6408->name, nm6408);
	if (rc) {
		dev_err(dev, "unable request irq.\n");
		pci_set_drvdata(pdev, NULL);
#if !LEGACY
		pci_free_irq_vectors(pdev);
#endif
		nm6408->irq = -1;
		goto err_chrdev_destroy;
	}

	pci_set_master(pdev);
	pci_save_state(pdev);

	nm6408->is_alive = 0;

	return 0;

err_chrdev_destroy:
	device_destroy(nm6408_class, node);

err_chrdev_delete:
	cdev_del(&nm6408->cdev);

err_workqueue_destroy:
	destroy_workqueue(nm6408->work_queue);

err_dma_free_coherent:
	dma_free_coherent(dev, PAGE_SIZE, nm6408->desc_cpu_addr, nm6408->desc_bus_addr);

err_pci_release_ioumap:
	for (i = 0; i < NM6408_BAR_NMB; i++) {
		if (nm6408->region_virt[i] != 0)
			pci_iounmap(pdev, nm6408->region_virt[i]);

		if (nm6408->region_indx[i] >= 0)
			pci_release_region(pdev, nm6408->region_indx[i]);
	}

	pci_disable_device(pdev);

err_free_mem:
	if (nm6408->name)
		kfree(nm6408->name);

	if (nm6408->indx >= 0)
		clear_bit(nm6408->indx, &nm6408_brds_indx);

	kfree(nm6408);

	return rc;
}

static void nm6408_remove(struct pci_dev *pdev)
{
	struct nm6408_brd	*nm6408 = pci_get_drvdata(pdev);
	struct device		*dev = &pdev->dev;
	dev_t			node = MKDEV(nm6408_major, nm6408_minor + nm6408->indx);
	unsigned long		flags;
	int			i;

        dev_info(dev, "device remove pdev=%p (%02x:%02x.%01x), [%04x, %04x] (rev %x).\n",
		pdev, pdev->bus->number, PCI_SLOT(pdev->devfn), PCI_FUNC(pdev->devfn),
		(int)pdev->vendor, (int)pdev->device, (int)pdev->revision);

	spin_lock_irqsave(&nm6408->lock, flags);
	nm6408->irq = -1;
	spin_unlock_irqrestore(&nm6408->lock, flags);

#if LEGACY
	free_irq(pdev->irq, nm6408);;
#else
	free_irq(pci_irq_vector(pdev, 0), nm6408);
#endif

	flush_workqueue(nm6408->work_queue);
	del_timer_sync(&nm6408->timer);

	if (nm6408->is_alive) {
		pci_restore_state(pdev);

		iowrite32(WDT_UNLOCK_TEMPLATE, nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK);
		mb();
		iowrite32(0x0, nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CTRL);
		mb();
		iowrite32(WDT_LOAD_1_SEC, nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOAD);
		mb();
		iowrite32(WDT_INT_STATUS, nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CLR);
		mb();
		iowrite32(~WDT_UNLOCK_TEMPLATE, nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK);

		nm6408->is_alive = 0;
	}

#if !LEGACY
	pci_free_irq_vectors(pdev);
#endif

	for (i = 0; i < NM6408_BAR_NMB; i++) {
		if (nm6408->region_virt[i] != 0)
			pci_iounmap(pdev, nm6408->region_virt[i]);

		if (nm6408->region_indx[i] >= 0)
			pci_release_region(pdev, nm6408->region_indx[i]);
	}

	if (pci_is_enabled(pdev))
		pci_disable_device(pdev);

	destroy_workqueue(nm6408->work_queue);
	dma_free_coherent(dev, PAGE_SIZE, nm6408->desc_cpu_addr, nm6408->desc_bus_addr);

	device_destroy(nm6408_class, node);
	cdev_del(&nm6408->cdev);

	kfree(nm6408->name);
	clear_bit(nm6408->indx, &nm6408_brds_indx);

	pci_set_drvdata(pdev, NULL);
	kfree(nm6408);

	return;
}

#ifdef CONFIG_PM
static int nm6408_suspend(struct pci_dev *pdev, pm_message_t state)
{
	struct nm6408_brd	*nm6408 = pci_get_drvdata(pdev);
	unsigned long		flags;

	if (!nm6408) {
		printk(KERN_INFO NM6408_BRD_NAME ": device pointer is NULL.\n");
		return 0;
	}

	printk(KERN_WARNING "%s: going to suspend.\n", nm6408->name);

	spin_lock_irqsave(&nm6408->lock, flags);
	if (nm6408->irq <= 0) {
		spin_unlock_irqrestore(&nm6408->lock, flags);
		return -EIO;
	}
	nm6408->irq = -1;
	spin_unlock_irqrestore(&nm6408->lock, flags);

	do {
		void __iomem	*iomem;
		int		i;

		down(&nm6408->semap);
		nm6408->is_alive = 0;
		up(&nm6408->semap);

		del_timer_sync(&nm6408->timer);

		iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK;
		iowrite32(WDT_UNLOCK_TEMPLATE, iomem);

#if defined(CONFIG_E2K)
                ioread32(iomem); // barrier to ensure write
#endif

		if (ioread32(iomem) == 0) {
			iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CTRL;
			iowrite32(0x0, iomem);

			mb();

			iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOAD;
			iowrite32(WDT_LOAD_1_SEC, iomem);

			mb();

			iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CLR;
			iowrite32(WDT_INT_STATUS, iomem);
		}
		else {
			printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 28).\n", nm6408->name);
			return -EIO;
		}

		mb();

		iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK;
		iowrite32(~WDT_UNLOCK_TEMPLATE, iomem);

		iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_BOARD_READY;
		if (ioread32(iomem) != 0x5A5A5A5A)
			break;

		iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_CSC_PCIE_REG2;
		iowrite32(ioread32(iomem) & ~PCIE_REG2_PM_AUTOACK, iomem);

		mb();

		iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_SWI_COMMAND;
		iowrite32(SWI_BOARD_SUSPEND, iomem);

		mb();

		iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_GIC_SGIR;
		iowrite32(GIC_SGIR_SWI0, iomem);

		iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_SWI_COMMAND;
                i = 1000;

		while (ioread32(iomem) != SWI_BOARD_RESUME) {
			if (!i)
				break;

			msleep(1);
			i--;
		}

		if (ioread32(iomem) != SWI_BOARD_RESUME) {
			printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 29).\n", nm6408->name);
			return -EIO;
		}
	} while (0);

	pci_save_state(nm6408->pdev);
	pci_disable_device(nm6408->pdev);
	pci_set_power_state(nm6408->pdev, PCI_D3hot);
	pci_cfg_access_lock(nm6408->pdev);

	return 0;
}

static int nm6408_resume(struct pci_dev *pdev)
{
	struct nm6408_brd	*nm6408 = pci_get_drvdata(pdev);
	unsigned long		flags;
	int			rc;

	if (!nm6408) {
		printk(KERN_INFO NM6408_BRD_NAME ": device pointer is NULL.\n");
		return 0;
	}

	printk(KERN_WARNING "%s: going to resume.\n", nm6408->name);

	pci_cfg_access_unlock(nm6408->pdev);
	pci_set_power_state(nm6408->pdev, PCI_D0);
	pci_restore_state(nm6408->pdev);

	rc = pci_enable_device_mem(nm6408->pdev);
	if (rc) {
		printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 30).\n", nm6408->name);
		return rc;
	}

	pci_save_state(nm6408->pdev);

	do {
		void __iomem	*iomem;
		int		i;

		iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_BOARD_READY;
		if (ioread32(iomem) != 0x5A5A5A5A)
			break;

		iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_MEMS_SWI_COMMAND;
		i = 1000;

		while (ioread32(iomem) != 0xFFFFFFFF) {
			if (!i)
				break;

			msleep(1);
			i--;
		}

                if (ioread32(iomem) != 0xFFFFFFFF) {
                        printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 31).\n", nm6408->name);
                        return -EIO;
                }

		iowrite32(0, iomem);
		mb();

		iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK;
		iowrite32(WDT_UNLOCK_TEMPLATE, iomem);

#if defined(CONFIG_E2K)
		ioread32(iomem); // barrier to ensure write
#endif

		if (ioread32(iomem) == 0) {
			iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CTRL;
			iowrite32(0x0, iomem);

			mb();

			iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOAD;
			iowrite32(WDT_LOAD_1_SEC, iomem);

			mb();

			iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CLR;
			iowrite32(WDT_INT_STATUS, iomem);

			mb();

			iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_CTRL;
			iowrite32(WDT_CTRL_INTEN, iomem);
		}
		else {
			printk(KERN_WARNING "%s: need hardware reset and/or reload driver of board (case - 32).\n", nm6408->name);
			return -EIO;
		}

		mb();

		iomem = nm6408->region_virt[NM6408_BAR_CTRL] + CCPU_REGS_WDT_LOCK;
		iowrite32(~WDT_UNLOCK_TEMPLATE, iomem);

		nm6408->timer.expires = jiffies + (NM6408_XFER_TIMEOUT * HZ) / 1000; 
		add_timer(&nm6408->timer);

		down(&nm6408->semap);
		nm6408->is_alive = 1;
		up(&nm6408->semap);
	} while (0);

	spin_lock_irqsave(&nm6408->lock, flags);
#if LEGACY
		nm6408->irq = pdev->irq;
#else
		nm6408->irq = pci_irq_vector(pdev, 0);
#endif
	spin_unlock_irqrestore(&nm6408->lock, flags);

	return 0;
}
#endif /* CONFIG_PM */

static const struct pci_device_id nm6408_ids[] = {
	{ PCI_DEVICE(NM6408_VENDOR_ID, NM6408_DEVICE_ID), },
	{ /* end: all zeroes */ },
};
MODULE_DEVICE_TABLE(pci, nm6408_ids);

static struct pci_driver nm6408_driver = {
	.name		= NM6408_BRD_NAME,
	.probe		= nm6408_probe,
	.remove		= nm6408_remove,
	.shutdown	= nm6408_remove,
	.id_table	= nm6408_ids,
#ifdef CONFIG_PM
	.suspend	= nm6408_suspend,
	.resume		= nm6408_resume,
#endif /* CONFIG_PM */
};

static int __init nm6408_init(void)
{
	int	i, rc;

	printk(KERN_INFO "%s - driver version: %s\n", NM6408_BRD_NAME, DRIVER_VERSION);

	nm6408_class = class_create(THIS_MODULE, NM6408_BRD_NAME);
	if (IS_ERR(nm6408_class)) {
		printk(KERN_INFO NM6408_BRD_NAME ": class create error.\n");
		rc = PTR_ERR(nm6408_class);
		goto err_class_create;
	}

	rc = alloc_chrdev_region(&nm6408_dev, NM6408_BRD_MINOR, NM6408_BRD_NUM, NM6408_BRD_NAME);
	if (rc) {
		printk(KERN_INFO NM6408_BRD_NAME ": chrdev region allocate error.\n");
		goto err_alloc_region;
	}

	nm6408_major = MAJOR(nm6408_dev);
	nm6408_minor = MINOR(nm6408_dev);

	for(i = 0; i < NM6408_BRD_NUM; i++)
		clear_bit(i, &nm6408_brds_indx);

	rc = pci_register_driver(&nm6408_driver);
	if (rc) {
		printk(KERN_INFO NM6408_BRD_NAME ": pci driver register error.\n");
		goto err_driver_register;
	}

	return 0;

err_driver_register:
	unregister_chrdev_region(nm6408_dev, NM6408_BRD_NUM);

err_alloc_region:
	class_destroy(nm6408_class);

err_class_create:
	return rc;
}
module_init(nm6408_init);

static void __exit nm6408_exit(void)
{
	pci_unregister_driver(&nm6408_driver);
	unregister_chrdev_region(nm6408_dev, NM6408_BRD_NUM);
	class_destroy(nm6408_class);

	return;
}
module_exit(nm6408_exit);

MODULE_AUTHOR("RC Module <nm-support@module.ru>");
MODULE_DESCRIPTION("Driver for " NM6408_BRD_NAME " board");

MODULE_LICENSE("Dual BSD/GPL");
MODULE_VERSION(DRIVER_VERSION);

