//////////////////////////////////////////////////////////////////////////
//                                                                      //
// services.cpp -                                                       //
//                                                                      //
//      Tool for NM and ARM programs to access host devices             //
//                                                                      //
//      Note: For NM, the implementation takes into account the sparse  //
//            representation of characters (see charToSparse ).         //
//            Response part to targets - io_stub.c                      //
//     ops->write_mem_block,   ops->read_mem_block,                      //
//     ops->type_access,      get_region_RPC                               //
//////////////////////////////////////////////////////////////////////////
#include <cassert>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <locale.h>

#include "host_duties.h"
#include "libload_lin.h"
#include "io_host_lin.h"
#include "io_host.h"

using namespace std;

void __attribute__((constructor)) robust_mutex_constructor(void);
void __attribute__((destructor)) robust_mutex_destructor(void);

static pthread_mutex_t		print_mutex;
static pthread_mutex_t		works_mutex;
static pthread_mutexattr_t	mutexattr;

static int _pthread_mutex_lock(pthread_mutex_t *mutex)
{
	int rv;

	rv = pthread_mutex_lock(mutex);
	if (rv == EOWNERDEAD)
		rv = pthread_mutex_consistent(mutex);

	return rv;
}

static int _pthread_mutex_unlock(pthread_mutex_t *mutex)
{
	return pthread_mutex_unlock(mutex);
}


//#if IO_SERV_DBG
static char found[256];

static void OUT_DBG(char *found, FILE *F)
{
	if (F == NULL) {
		printf("%s", found);
		fflush(stdout);
	}
	else {
		fprintf(F, "%s", found);
		fflush(F);
	}

	return;
}
//#endif // IO_SERV_DBG

#ifdef __cplusplus
extern "C"
{
#endif

int print_mutex_lock(void)
{
	return _pthread_mutex_lock(&print_mutex);
}

int print_mutex_unlock(void)
{
	return _pthread_mutex_unlock(&print_mutex);
}

#ifdef __cplusplus
}; // extern "C"
#endif

static int wstat2lstat(struct stat *st, uint32_t *stat, int ch);
void *IOServiceDispatcher(void *v_trg);
static ssize_t sparseCharToChar(char *dst, PL_Word *src, size_t ch_count);
static ssize_t charToSparseChar(PL_Word *dst, char *src, size_t ch_count);

int IO_Service::mutexLock()
{
	struct ll_operation *ops = (struct ll_operation *)ll_ops;
	PL_Word hostWantMutex;
	PL_Word boardWantMutex;
	PL_Word whoIsWaitMutex;

	hostWantMutex = MUTEX_TRUE;
	whoIsWaitMutex = HOST_THREAD;
	int res;

	if ((res=ops->write_mem_block(a, &hostWantMutex, datSecAddr + trg_hostWantMutex * bINw, 1)) != PL_OK) {
#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		printf(found, "mutexLock: failed wtire hostWantMutex res=%d\n",res);
		OUT_DBG(found, trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif

		throw "mutexLock error";
	}

	if ((res=ops->write_mem_block(a, &whoIsWaitMutex, datSecAddr + trg_whoIsWaitMutex * bINw, 1)) != PL_OK) {
#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "mutexLock: failed wtire whoIsWaitMutex res=%d\n",res);
		OUT_DBG(found, trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif

		throw "mutexLock error";
	}

	do {
		usleep(1000);

		if ((res=ops->read_mem_block(a, &whoIsWaitMutex, datSecAddr + trg_whoIsWaitMutex * bINw, 1)) != PL_OK) {
#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "mutexLock: failed read whoIsWaitMutex res=%d\n",res);
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif

			throw "mutexLock error";
		}

		if ((res=ops->read_mem_block(a, &boardWantMutex, datSecAddr + trg_boardWantMutex * bINw, 1)) != PL_OK) {
#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "mutexLock: failed read boardWantMutex res=%d\n",res);
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif

			throw "mutexLock error";
		}
	} while ((whoIsWaitMutex == HOST_THREAD) && (boardWantMutex == MUTEX_TRUE));

	return 0;
}

int
IO_Service::mutexUnlock()
{
	struct ll_operation *ops = (struct ll_operation *)ll_ops;
	PL_Word hostWantMutex;

	hostWantMutex = MUTEX_FALSE;
	int res;

	if ((res=ops->write_mem_block(a, &hostWantMutex, datSecAddr + trg_hostWantMutex * bINw, 1)) != PL_OK) {
#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "mutexUnlock: failed write hostWantMutex res=%d\n",res);
		OUT_DBG(found, trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif

		throw "mutexUnlock error";
	}

	return 0;
}

bool
IO_Service::isMutexLocked()
{
	struct ll_operation *ops = (struct ll_operation *)ll_ops;
	PL_Word hostWantMutex;
	int res;

	if ((res=ops->read_mem_block(a, &hostWantMutex, datSecAddr + trg_hostWantMutex * bINw, 1)) != PL_OK) {
#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "isMutexLocked: failed read hostWantMutex res=%d\n",res);
		OUT_DBG(found, trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif

		throw "isMutexLocked error";
	}

	return (hostWantMutex == MUTEX_TRUE);
}

void 
IO_Service::readStatus()
{
	struct ll_operation *ops = (struct ll_operation *)ll_ops;

#if IO_SERV_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "readStatus a=0x%lx bsStatus=0x%lx datSecAddr=0x%x trg_size=0x%x\n",
			a, bsStatus, datSecAddr, trg_size);
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif

	mutexLock();

	if (ops->read_mem_block(a, bsStatus, datSecAddr, trg_size) != PL_OK) {
#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "status block reading failure\n");
		OUT_DBG(found, trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif
		mutexUnlock();
		throw "status block reading failure";
	}

	mutexUnlock();

#if IO_SERV_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "SERV_RPC 0-0x%x 1-0x%x 2-0x%x 3-0x%x 4-0x%x 5-0x%x 6-0x%x 7-0x%x\n",
			bsStatus[trg_boardSend], bsStatus[trg_sended], bsStatus[trg_sendMostDistant],
			bsStatus[trg_io_debug], bsStatus[trg_hostReceive], bsStatus[trg_sendConfirmed],
			bsStatus[trg_backBufferReady], bsStatus[trg_io_debug_back]);
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif

	return;
}

void 
IO_Service::writeBackStatus()
{
	struct ll_operation *ops = (struct ll_operation *)ll_ops;
	PL_Word *dat;
	PL_Addr addr;

	dat = bsStatus + trg_hostReceive;
	addr = datSecAddr + trg_hostReceive * bINw;

#if IO_SERV_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "writeBackStatus bsStatus 4-0x%x 5-0x%x 6-0x%x 7-0x%x\n",
			bsStatus[trg_hostReceive], bsStatus[trg_sendConfirmed],
			bsStatus[trg_backBufferReady], bsStatus[trg_io_debug_back]);
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif

	mutexLock();

	if (ops->write_mem_block(a, dat, addr, trg_io_debug_back - trg_io_debug) != PL_OK) {
#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "writeBackStatus status block writing failure\n");
		OUT_DBG(found, trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif
		mutexUnlock();
		throw "status block writing failure";
	}

	mutexUnlock();

	return;
}

//
// expected, that destructor is called after finish target program
//
IO_Service::~IO_Service()
{
	struct ll_operation *ops = (struct ll_operation *)ll_ops;
	int i = 0;
#if IO_SERV_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "~IO_Service START\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif

	if (IOService_finish_start) {
		while ((!IOService_finish_end) && ((i * 10) <= destructor_time)) {
			usleep(10000); // 10 mlsec
			i++;
		}
	}
	else {
		IOService_finish(this);
	}

	if (isMutexLocked())	mutexUnlock();
	_pthread_mutex_lock(&works_mutex);
	if (bsStatus)	{ delete[] bsStatus; bsStatus=NULL;}
	if (bsBuf)	{ delete[] bsBuf; bsBuf=NULL;}
	if (SemihostFile_table)	{ delete[] SemihostFile_table; SemihostFile_table=NULL;}
	_pthread_mutex_unlock(&works_mutex);

//////////////////////////////////////////////////////////
		RPC_busy &= ~RPC_BUSY_BIT;
		if (ops->write_mem_block(a, &RPC_busy, datSecAddr + trg_rpc_busy * bINw, 1) != PL_OK) {
#if IO_SERV_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "Clear RPC_busy failure\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				throw "clear RPC_busy error";
		}
//////////////////////////////////////////////////////////
	pthread_detach(io_thread);  // forget the thread and his resorce
	
#if IO_SERV_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "~IO_Service END\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
}

static ssize_t
sparseCharToChar(char *dst, PL_Word *src, size_t ch_count)
{
	ssize_t res = 0;

	while (res < ch_count) {
		*(dst + res) = (char)(*(src + res) & 0xFF);
		res++;
	}

	return res; 
}

static ssize_t
charToSparseChar(PL_Word *dst, char *src, size_t ch_count)
{
	ssize_t res = 0;

	while (res < ch_count) {
		*(dst + res) = (PL_Word)(*(src + res)) & 0xFF;
		res++;
	}

	return res;
}

static void processPrintf(char *charBuf, size_t charCnt, FILE *fileBuf)
{
	*(charBuf + charCnt - 1) = '\0';

	_pthread_mutex_lock(&print_mutex);
	if (fileBuf != NULL) {
		fprintf(fileBuf, "%s", charBuf);
		fflush(fileBuf);
	}
	else {
		printf("%s", charBuf);
		fflush(stdout);
	}
	_pthread_mutex_unlock(&print_mutex);

	return;
}

IO_Service::IO_Service(
		PL_Access* access, FILE *_file,
		void *_llops, int *error_occur,
		ostream* _cout, ostream* _cerr,istream* _cin)
:letsShutDown(false),a(access),trg_cout(_cout),trg_file(_file),ll_ops(_llops)
{
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "START IO_Service access=0x%lx FILE=0x%lx\n",
		access, _file);
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
	DWORD status = 0;
	struct ll_operation *ops = (struct ll_operation *)ll_ops;
	if (ops == NULL) {
		ops = &PL_FunctionExport;
		ll_ops = (void *)ops;
	}

	_pthread_mutex_lock(&works_mutex); //?????????? MUTEX ??????????????

	try {
		// checking ops
		char* method_err=NULL;
		if (ops == NULL) {
			if ((error_occur) != NULL) *error_occur = IO_SERV_NO_METHOD;
			throw "IO_Service: Not allowed method to work";
		}
		if(ops->type_access == NULL) method_err=(char *)"type_access";
		else if(ops->get_region_RPC == NULL) method_err=(char *)"get_region_RPC";
		else if(ops->get_status == NULL) method_err=(char *)"get_status";
		else if(ops->read_mem_block == NULL) method_err=(char *)"read_mem_block";
		else if(ops->write_mem_block == NULL) method_err=(char *)"write_mem_block";
		if (method_err != NULL) {
			if ((error_occur) != NULL) *error_occur = IO_SERV_NO_METHOD;
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "IO_Service: No method %s\n",method_err);
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
			throw "IO_Service: Not allowed method to work";
		}
		//////////////////

		DWORD is_nm_core = 222;

		if ((io_service_fail.err =ops->type_access(a, &is_nm_core)) != PL_OK) {
#if IO_SERV_DBG 
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "Unable get type of core\n");
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif
			io_service_fail.function = BSP_access;
			if ((error_occur) != NULL) *error_occur = IO_SERV_BSP;
			throw "IO_Service: get type of core failure";
		}

		nmc_core = (is_nm_core == 0) ? false : true;
		bINw = (nmc_core) ? nmcIntegerSize : armIntegerSize;

		// check that target-prog running
		if ((io_service_fail.err = ops->get_status(a, &status)) != PL_OK) {
#if STDIO_TRACE
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "IOService: get_status error\n");
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif
			io_service_fail.function = BSP_status;
			if ((error_occur) != NULL) *error_occur = IO_SERV_BSP;
			throw "IO_Service: get program status failure";
		}
		else if (!(status)) { //target-prog is not running at all
			if((error_occur) != NULL) *error_occur = IO_SERV_NO_PROG;
			throw "IO_Service: target-prog is not running";
		}
		
		if ((io_service_fail.err =ops->get_region_RPC(a, &datSecAddr, &datSecSize)) != PL_OK) {
			if(io_service_fail.err == PL_AGAIN){
#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "IO_Service: RPC region is busy\n");
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif
				io_service_fail.function = IO_service_err;
				if ((error_occur) != NULL) *error_occur = IO_SERV_SECOND;
				throw "IO_Service: RPC region is busy";
			}
#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "Unable get control region of RPC\n");
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif
			io_service_fail.function = BSP_access;
			if ((error_occur) != NULL) *error_occur = IO_SERV_BSP;
			throw "IO_Service: get RPC region failure";
		}

#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "IO_Service %s datSecAddr=0x%x  datSecSize=0x%x\n",
			".rpc_services", datSecAddr, datSecSize);
		OUT_DBG(found, trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif

		if (!datSecAddr || (datSecSize != trg_size)) {
#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "Control region of RPC mismatch\n");
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif
			if ((error_occur) != NULL) *error_occur = IO_SERV_BSP;
			throw "IO_Service: RPC region mismatch";
		}

		// Here check free RPC busy bit 
		RPC_busy = 0;
		if ((io_service_fail.err=ops->read_mem_block(a, &RPC_busy, datSecAddr + trg_rpc_busy * bINw, 1)) != PL_OK) {
#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "Get RPC_busy failure\n");
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif
			io_service_fail.function = BSP_read;
			if ((error_occur) != NULL) *error_occur = IO_SERV_BSP;
			throw "IO_Service: get RPC_busy error";
		}

		if (RPC_busy & RPC_BUSY_BIT) {//Busy
#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "Second IO_Service with the same core\n");
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif
			if ((error_occur) != NULL) *error_occur = IO_SERV_SECOND;
			throw "IO_Service: Second IO_Service with the same core";
		}

		RPC_busy |= RPC_BUSY_BIT;
		if ((io_service_fail.err=ops->write_mem_block(a, &RPC_busy, \
				datSecAddr + trg_rpc_busy * bINw, 1)) != PL_OK) {
#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "Set RPC_busy failure\n");
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif
			io_service_fail.function = BSP_write;
			throw "IO_Service: set RPC_busy error";
		}

		// initial initialization of variables
		SemihostFile_table = NULL;
		bsStatus = NULL;
		bsBuf = NULL;
		io_return = -1;
		dispatch_start = false;
		dispatch_end=false;
		IOService_finish_start = false;
		IOService_finish_end = false;
		destructor_force = 0;
		destructor_time = 10000; // 10 seconds
		io_service_fail.function = -1;
		io_service_fail.err = 0;


		if (pthread_create(&io_thread, NULL, IOServiceDispatcher, (void *)this)){
			if ((error_occur) != NULL) *error_occur = IO_SERV_ERROR;
			throw "IO_Service: can't start IO_Service - pthread_create error";
		}

#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "IO_Service pthread_create\n");
		OUT_DBG(found, trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif

		while (!dispatch_start) {// waiting start warking ::dispatch
			_pthread_mutex_unlock(&works_mutex);  //??????????????????????
			usleep(10000); // 10 mlsec
			_pthread_mutex_lock(&works_mutex);   //???????????????????????

			if (dispatch_end){ // dispatch finished very quick
				if ((error_occur) != NULL) *error_occur = IO_SERV_ERROR;
				throw "IO_Service:can't start IO_Service - some wrong";
			}
		}
	}
		catch (const char* exception){
#if IO_SERV_DBG
		cerr << endl << "Error: " << exception << endl;
		cerr << endl << "IO_Service constructor aborted!" << endl;
#endif
		_pthread_mutex_unlock(&works_mutex);   //????????????????????????
		throw -1;
	}
	catch (...) {
#if IO_SERV_DBG
		cerr << endl << "IO_Service constructor aborted!" << endl;
#endif
		_pthread_mutex_unlock(&works_mutex);  //????????????????????????
		throw -1;
	}

	_pthread_mutex_unlock(&works_mutex);  //????????????????????????
}

void *IOServiceDispatcher(void *v_trg)
{
	IO_Service& trg= *static_cast<IO_Service*>(v_trg);
	struct ll_operation *ops = (struct ll_operation *)trg.ll_ops;

#if IO_SERV_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found,"ServiceDispatcher: _force=0x%x v_trg=0x%lx trg=0x%lx\n",
			trg.destructor_forcev_trg, &trg);
	OUT_DBG(found, trg.trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, 0x00);
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, 0x00);

	try {
		trg.SemihostFile_table = new SemihostFile_t[SEMIHOST_MAX_FILE];
		trg.SemihostFile_index = 3;

		trg.SemihostFile_table[0].fd = 0;
		trg.SemihostFile_table[0].TextMode = 1;

		trg.SemihostFile_table[1].fd = 1;
		trg.SemihostFile_table[1].TextMode = 1;

		trg.SemihostFile_table[2].fd = 2;
		trg.SemihostFile_table[2].TextMode = 1;

		for (int i = 3; i < SEMIHOST_MAX_FILE; i++) {
			trg.SemihostFile_table[i].fd = -1;
			trg.SemihostFile_table[i].TextMode = 0; 
		}
#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "ServiceDispatcher: End Install SemihostFile_table\n");
		OUT_DBG(found, trg.trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif
		if (trg.datSecSize == trg_size) {
			trg.bsStatus = new PL_Word[trg.datSecSize];
		}else{
			throw "ServiceDispatcher: Bad datSecSize";
		}

		if (((trg.io_service_fail).err=ops->read_mem_block(trg.a, &trg.bssSecAddr, trg.datSecAddr + trg_rpcSendAddr * trg.bINw, 1)) != PL_OK) {
#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "Unable get addr of RPC buffer\n");
			OUT_DBG(found, trg.trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif
			(trg.io_service_fail).function = BSP_read;
			throw "ServiceDispatcher: get buffer addr failure";
		}

		if (((trg.io_service_fail).err=ops->read_mem_block(trg.a, &trg.bssSecSize, trg.datSecAddr + trg_rpcSendSize * trg.bINw, 1)) != PL_OK) {
#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "Unable get size of RPC buffer\n");
			OUT_DBG(found, trg.trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif
			(trg.io_service_fail).function = BSP_read;
			throw "ServiceDispatcher: get buffer size failure";
		}

#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "b2hBufSize(0x%x) + backBufferSize(0x%x)=0x%x\n",
			b2hBufSize, backBufferSize, (b2hBufSize + backBufferSize));
		OUT_DBG(found, trg.trg_file);

		sprintf(found, "IO_Service %s  bssSecAddr=0x%x  bssSecSize=0x%x\n",
			".rpc_services.bss", trg.bssSecAddr, trg.bssSecSize);
		OUT_DBG(found, trg.trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif

		if(!trg.bssSecAddr || (trg.bssSecSize < (b2hBufSize + backBufferSize))) {
#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "IO_Service Buffer of RPC mismatch\n");
			OUT_DBG(found, trg.trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif
			throw "ServiceDispatcher: RPC buffer mismatch";
		}

		trg.bsBuf = new PL_Word[trg.bssSecSize];

#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "IO_Service bsStatus=0x%lx, bsBuf=0x%lx\n", 
				trg.bsStatus, trg.bsBuf);
		OUT_DBG(found, trg.trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif

//		_pthread_mutex_lock(&works_mutex);
		trg.dispatch_start = true;
//		_pthread_mutex_unlock(&works_mutex);

		trg.io_return = trg.dispatch();
#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		sprintf(found,"dispatch is finished\n");
		OUT_DBG(found, trg.trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif
	}
	catch (const char* exception) {
#if IO_SERV_DBG
		cerr << endl << "Error: " << exception << endl;
		cerr << endl << "IO_Service dispatch aborted!" << endl;
#endif
	}
	catch (...) {
#if IO_SERV_DBG
		cerr << endl << "error: IO_Service thread is aborted!" << endl;
#endif
	}

	trg.dispatch_end = true;
	_pthread_mutex_lock(&works_mutex);

		if (trg.bsStatus) {
			delete[] trg.bsStatus;
			trg.bsStatus = NULL;
		}

		if (trg.bsBuf) {
			delete[] trg.bsBuf;
			trg.bsBuf = NULL;
		}

		if (trg.SemihostFile_table) {
			delete[] trg.SemihostFile_table;
			trg.SemihostFile_table = NULL;
		}
	_pthread_mutex_unlock(&works_mutex);

	pthread_exit((void *)&trg.io_return);
}

/*
					//		Win		Lin
#define O_RDONLY	_O_RDONLY		//	+	0x00000		0x00000
#define O_WRONLY	_O_WRONLY	//	+	0x00001		0x00001
#define O_RDWR		_O_RDWR			//	+	0x00002		0x00002
#define O_APPEND	_O_APPEND		//	-	0x00008		0x02000
#define O_CREAT		_O_CREAT		//	+	0x00100		0x00100
#define O_TRUNC		_O_TRUNC		//	-	0x00200		0x01000
#define O_EXCL		_O_EXCL			//	-	0x00400		0x00200
#define O_TEXT		_O_TEXT			//	-	0x04000		0x20000
#define O_BINARY	_O_BINARY		//	-	0x08000		0x10000
*/

#define char2int(x)	(((x) + bINw - 1) / bINw)
#define word2byte(x)	((x) * (nmc_core ? 4 : 1))
#define byte2word(x)	((x) / (nmc_core ? 4 : 1))
#define lose4div(x)	((x) % (nmc_core ? 4 : 1))

int
IO_Service::dispatch()
{
	volatile struct RPC_control_region *status = (struct RPC_control_region *)bsStatus;
	volatile struct RPC_send_region *buffer = (struct RPC_send_region *)bsBuf;
	struct ll_operation *ops = (struct ll_operation *)ll_ops;

	bool proceed = true;

	while (proceed) {
		usleep(1000);

//		_pthread_mutex_lock(&works_mutex);
		proceed = !letsShutDown;
//		_pthread_mutex_unlock(&works_mutex);

		readStatus();

		if (status->hostReceive == status->boardSend)
			continue; // nothing to do

		int words;
		PL_Word *hostBuf = (PL_Word *)buffer->boardOutServBuf;
		PL_Word r;

		if (status->sended < status->sendConfirmed) {
			words = status->sendMostDistant - status->sendConfirmed;

#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "YYY sended < sendConfirmed; hostBuf=0x%lx\n", hostBuf);
			OUT_DBG(found, trg_file);

			sprintf(found, "YYY words(0x%x) = sendMostDistant(0x%x) - sendConfirmed(0x%x)\n",
					words, status->sendMostDistant, status->sendConfirmed);
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif

			if (words) {
				r = ops->read_mem_block(a, hostBuf, bssSecAddr + status->sendConfirmed * bINw, words);
				if (r != PL_OK)
					throw "error reading block from board";

				hostBuf += words;
			}

			status->sendConfirmed = 0;
		}

		words = status->sended - status->sendConfirmed;

#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "YYY sended >= sendConfirmed;  words=0x%x hostBuf=0x%lx\n", words, hostBuf);
		OUT_DBG(found, trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif

		if (words) {
			r = ops->read_mem_block(a, hostBuf, bssSecAddr + status->sendConfirmed * bINw, words);
			if ( r != PL_OK ) {
				io_service_fail.err = r;
				io_service_fail.function = BSP_read;
				throw "IO_Service dispatcher: error reading block from board";
			}
			hostBuf += words;
		}

		status->sendConfirmed = status->sended;
		PL_Word *bufferFromBoard = (PL_Word *)buffer->boardOutServBuf;
		words = hostBuf - bufferFromBoard;

#if IO_SERV_DBG
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "YYY words=0x%x bufferFromBoard=0x%lx, boardSend=0x%x, hostReceive=0x%x\n",
				words, bufferFromBoard, status->boardSend, status->hostReceive);
		OUT_DBG(found, trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif

		PL_Word *bufferToBoard = (PL_Word *)buffer->backBuffer;
		PL_Word *charBuf = new PL_Word[5 * packetLimit];
		int backCount = 0;

		while(status->hostReceive != status->boardSend) {
			IO_ServiceID pKind;
			int pSize;
			char *charTmp;

			if (!words)
				break;

			pSize = (int)(*bufferFromBoard++);
			pKind = (IO_ServiceID)(*bufferFromBoard++);
			words -= 2;

			if (pSize > (packetLimit - 2) * bINw){
				io_service_fail.err = -1;
				io_service_fail.function = IO_service_err;
				throw "IO_Service dispatcher: error of packet format";
			}

			if ((pKind != NMSYNC_SERVICE_PRINTF) && (status->backBufferReady > backBufferLimit))
				pKind = (IO_ServiceID)(~0);

#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "EEE START pKind=0x%x pSize=0x%x bufferFromBoard=0x%lx"
					" hostReceive=0x%x  boardSend=0x%x words=0x%x\n",
					pKind, pSize, bufferFromBoard, status->hostReceive, status->boardSend, words);
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif

			errno = 0;

			switch (pKind) {
			case NMSYNC_SERVICE_PRINTF:
			{
				if (!pSize)
					break;

				charTmp = (char *)charBuf;

				if (nmc_core) {
					sparseCharToChar(charTmp, bufferFromBoard, pSize);
				}
				else {
					charTmp = (char *)bufferFromBoard;
				}

				processPrintf(charTmp, pSize, trg_file);

				break;
			}
			case INDEX_OPEN:
			{
				#define FLAGS_CNT	11

				const char *flags_str[FLAGS_CNT] = {
						"O_EXCL", "O_NOCTTY", "O_NONBLOCK", "O_CREAT",
						"O_TRUNC", "O_APPEND", "O_BINARY", "O_TEXT",
						"O_WRONLY", "O_RDONLY", "O_RDWR"
				};

				const unsigned flags_vlu[FLAGS_CNT] = {
						O_EXCL, 0x000000, 0x000000, O_CREAT,
						O_TRUNC, O_APPEND, O_BINARY, O_TEXT,
						O_WRONLY, O_RDONLY, O_RDWR
				};

				int errno1;

				int local_fd=-1;
				size_t f_size; //length file name
				char *f_name;
				char *open_flags;
				int flags=0, flags_cnt, index ;
				int flag_linux;
				int mode;
				uint32_t *buf = NULL;

#if STDIO_TRACE
_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_OPEN START\n");
	OUT_DBG(found, trg_file);
_pthread_mutex_unlock(&print_mutex);
#endif
				f_size = bufferFromBoard[0];
				index = char2int(f_size) + 1; //indx_flags
				charTmp = (char *)charBuf;
				// file name
				if (nmc_core) {
					sparseCharToChar(charTmp, &bufferFromBoard[1], f_size);
				}
				else {
					charTmp = (char *)&bufferFromBoard[1];
				}

				f_name = charTmp;
				charTmp += char2int(f_size) * bINw;
				flags_cnt = bufferFromBoard[index++];
				// opening flags 
				if (nmc_core) {
					sparseCharToChar(charTmp, &bufferFromBoard[index], pSize - index - 1);
				}
				else {
					charTmp = (char *)&bufferFromBoard[index];
				}

				open_flags = charTmp;
				flags = 0;
				for (int i = 0; i < flags_cnt; i++) {
					for (int n = 0; n < FLAGS_CNT; n++) {
						if (!strcmp(charTmp, flags_str[n])) {
							flags |= flags_vlu[n];
#if STDIO_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found,"INDEX_OPEN flags_str[%d]=%s flags_vlu[%d]=%o \n",
		n,flags_str[n],n,flags_vlu[n]);
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
							break;
						}
					}
					index += char2int(strlen(charTmp) + 1);
					charTmp += char2int(strlen(charTmp) + 1) * bINw;
				}
//				mode = bufferFromBoard[pSize - 1];
				mode = bufferFromBoard[index];
				// O_BINARY and O_TEXT is not use in linux
				flag_linux = flags & 0xFFFF;

#if STDIO_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_OPEN f_name=%s open_flags=%s ,flags=%o flag_linux=%o\n",
		f_name, open_flags,flags,flag_linux);
	OUT_DBG(found, trg_file);
	sprintf(found, "INDEX_OPEN mode=%o pSize=%d mode=%o index=%d\n",
		mode, pSize, bufferFromBoard[index], index);
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				if (flag_linux & O_CREAT){
					if ((mode < 0x200)&(mode > 0)) { //0777
						errno = 0;
						local_fd = open(f_name, flag_linux, mode);
					}else{
						errno = 0;
						local_fd = open(f_name, flag_linux, 0666);
					}
				} else {
					errno = 0;
					local_fd = open(f_name, flag_linux);
				}
				errno1 = errno;
				if (local_fd < 0) {
#if STDIO_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_OPEN NULL local_fd=%d errno=%d\n",local_fd,errno1);
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
					bufferToBoard[status->backBufferReady + 0] =  (PL_Word)-1;
					bufferToBoard[status->backBufferReady + 1] = errno1;
					bufferToBoard[status->backBufferReady + 2] = !(flags & O_BINARY);
					bufferToBoard[status->backBufferReady + 3] = INDEX_OPEN;
					bufferToBoard[status->backBufferReady + 4] = 3;

					status->backBufferReady += 5;
					backCount += 5;
					break;
				}
				int remote_fd = -1;

				for (int i = SemihostFile_index, j = SEMIHOST_MAX_FILE; i <= j; i++) {
					if (i == SEMIHOST_MAX_FILE) {
						j = SemihostFile_index;
						i = 3;
					}
					if (SemihostFile_table[i].fd > -1)
						continue;

					remote_fd = i;

					SemihostFile_table[i].fd = local_fd;
					SemihostFile_table[i].TextMode = (flags & O_BINARY) ? 0 : 1;

					SemihostFile_index = i + 1;

					if (SemihostFile_index == SEMIHOST_MAX_FILE)
						SemihostFile_index = 3;

					break;
				}

				if (remote_fd < 0) { //Too many opened files
					close(local_fd);
				}

				bufferToBoard[status->backBufferReady + 0] = remote_fd;
				bufferToBoard[status->backBufferReady + 1] = (remote_fd < 0) ? ENFILE : errno;
				bufferToBoard[status->backBufferReady + 2] = !(flags & O_BINARY);
				bufferToBoard[status->backBufferReady + 3] = INDEX_OPEN;
				bufferToBoard[status->backBufferReady + 4] = 3;

				status->backBufferReady += 5;
				backCount += 5;
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found,"INDEX_OPEN opened local_fd=%d TextMode=%d END\n",
			local_fd, SemihostFile_table[local_fd].TextMode);
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				break;
			}
//////////////////////////////////////////////////////////////////////
			case INDEX_CLOSE:
			{
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_CLOSE START\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif

				int remote_fd = bufferFromBoard[0];

				if ((remote_fd <= 2) || (remote_fd >= SEMIHOST_MAX_FILE)) {
					bufferToBoard[status->backBufferReady + 0] = -1;
					bufferToBoard[status->backBufferReady + 1] = EBADF;
					bufferToBoard[status->backBufferReady + 2] = INDEX_CLOSE;
					bufferToBoard[status->backBufferReady + 3] = 2;

					status->backBufferReady += 4;
					backCount += 4;

					break;
				}
				int local_fd = SemihostFile_table[remote_fd].fd;

				SemihostFile_table[remote_fd].fd = -1;
				SemihostFile_table[remote_fd].TextMode = 0;

				if (local_fd > 2) {
					bufferToBoard[status->backBufferReady + 0] = close(local_fd);
					bufferToBoard[status->backBufferReady + 1] = errno;
				}else {
					bufferToBoard[status->backBufferReady + 0] = -1;
					bufferToBoard[status->backBufferReady + 1] = EBADF;
				}
				bufferToBoard[status->backBufferReady + 2] = INDEX_CLOSE;
				bufferToBoard[status->backBufferReady + 3] = 2;

				status->backBufferReady += 4;
				backCount += 4;
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_CLOSE END OK\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				break;
			}
//////////////////////////////////////////////////////////////////////
			case INDEX_READ:
			{
				int errno1;
#if STDIO_TRACE
_pthread_mutex_lock(&print_mutex);
sprintf(found, "INDEX_READ START\n");
OUT_DBG(found, trg_file);
_pthread_mutex_unlock(&print_mutex);
#endif
				int remote_fd = bufferFromBoard[0];
				unsigned hsize = bufferFromBoard[1];
				unsigned tsize = bufferFromBoard[2];
				unsigned count = bufferFromBoard[3];
				PL_Addr addr = bufferFromBoard[4];
#if STDIO_TRACE
_pthread_mutex_lock(&print_mutex);
sprintf(found, "INDEX_READ remote_fd=%d hsize=%d tsize=%d count=%d\n",
		remote_fd, hsize, tsize, count);
OUT_DBG(found, trg_file);
_pthread_mutex_unlock(&print_mutex);
#endif
				int head, body, tail;
				int local_fd;
				if ((remote_fd == 1) || (remote_fd == 2) || (remote_fd < 0)  || \
						(remote_fd >= SEMIHOST_MAX_FILE) || \
						((local_fd = SemihostFile_table[remote_fd].fd)<0) ) { //stdout or stderr
					_pthread_mutex_lock(&print_mutex);
					sprintf(found, "INDEX_READ ERROR fd=%d fd=1 || fd=2\n",remote_fd);
					OUT_DBG(found, trg_file);
					_pthread_mutex_unlock(&print_mutex);
					bufferToBoard[status->backBufferReady + 0] = -1;
					bufferToBoard[status->backBufferReady + 1] = EBADF;
					bufferToBoard[status->backBufferReady + 2] = 0;
					bufferToBoard[status->backBufferReady + 3] = 0;
					bufferToBoard[status->backBufferReady + 4] = INDEX_READ;
					bufferToBoard[status->backBufferReady + 5] = 4;

					status->backBufferReady += 6;
					backCount += 6;

					break;
				}

				int isText = SemihostFile_table[remote_fd].TextMode;

#if STDIO_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_READ isText=%d fd=%d  count=%d\n",
				isText, remote_fd, count); 
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				int bytes;
				int index = 0;
				unsigned int count_all = hsize + count + tsize;
				if(count_all == 0) {
					bufferToBoard[status->backBufferReady + 0] = 0;
					bufferToBoard[status->backBufferReady + 1] = 0;
					bufferToBoard[status->backBufferReady + 2] = 0;
					bufferToBoard[status->backBufferReady + 3] = 0;
					bufferToBoard[status->backBufferReady + 4] = INDEX_READ;
					bufferToBoard[status->backBufferReady + 5] = 4;

					status->backBufferReady += 6;
					backCount += 6;
					break;
				}
				// read array from host
				unsigned int count_all_byte= word2byte(count_all);
				PL_Word *read_buf = (PL_Word *)malloc(count_all_byte+4*3);  //// MALLOC  read_buf
				memset(read_buf, 0, count_all_byte+4*3);
				char * char_read_buf = (char *)read_buf;
#if STDIO_DBG
_pthread_mutex_lock(&print_mutex);
sprintf(found, "INDEX_READ hsize=%d tsize=%d count=%d count_all=%d count_all_byte=%d\n",
		hsize,tsize,count,count_all,count_all_byte);
OUT_DBG(found, trg_file);
_pthread_mutex_unlock(&print_mutex);
#endif
				if (hsize) {
					head = index;
					char_read_buf += 4 - word2byte(hsize);
					index += 1;
				}

				body = index;
				index += count / bINw;

				if (tsize) {
					tail = index;
					index += 1;
				}

				if (isText && nmc_core) {
					charTmp = (char *)malloc(index+4);  //// MALLOC  charTmp
					memset(charTmp, 0, index + 4);
					bytes = read(local_fd, charTmp, index);
					errno1 = errno;
					if (bytes > 0)
						charToSparseChar(read_buf, charTmp, bytes);
					free(charTmp);                                     //// FREE  charTmp
				}
				else { // ARM or isBinary
					
					bytes = read(local_fd, char_read_buf, count_all_byte);
					errno1 = errno;
					int losser = lose4div(bytes);
					bytes = byte2word(bytes);

					if (losser) {// NMC binary - io service error in offset. Must be 0
//						SemihostFile_table[remote_fd].error = 1;
						bufferToBoard[status->backBufferReady + 0] = -1;
						bufferToBoard[status->backBufferReady + 1] = EIO;
						bufferToBoard[status->backBufferReady + 2] = 0;
						bufferToBoard[status->backBufferReady + 3] = 0;
						bufferToBoard[status->backBufferReady + 4] = INDEX_READ;
						bufferToBoard[status->backBufferReady + 5] = 4;

						status->backBufferReady += 6;
						backCount += 6;
						if (read_buf) free(read_buf);                         ////// FREE read_buf
						break;
					}
				}
				// here bytes in full words32
				if (bytes < hsize) {
					count = 0;
					tsize = 0;
				} else if (bytes < (hsize + count)) {
					count = char2int(bytes - hsize) * bINw;
					tsize = 0;
				}
#if STDIO_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_READ head=%d tail=%d body=%d count=%d  read_buf=%s\n",
				head, tail, body, count/bINw, &(read_buf[body])); 
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				if (count > 0) {
					if((r = ops->write_mem_block(a, &(read_buf[body]), addr, count/bINw)) != PL_OK) {
						_pthread_mutex_lock(&print_mutex);
						sprintf(found, "ERROR INDEX_READ: write_mem_block\n");
						OUT_DBG(found, trg_file);
						_pthread_mutex_unlock(&print_mutex);
						if (read_buf) free(read_buf);                        ////// FREE read_buf
						bufferToBoard[status->backBufferReady + 0] = 0;
						bufferToBoard[status->backBufferReady + 1] = BSP_ERROR+r;
						bufferToBoard[status->backBufferReady + 2] = 0;
						bufferToBoard[status->backBufferReady + 3] = 0;
						bufferToBoard[status->backBufferReady + 4] = INDEX_READ;
						bufferToBoard[status->backBufferReady + 5] = 4;

						status->backBufferReady += 6;
						backCount += 6;
						break;
//						throw "ERROR INDEX_READ: write_mem_block";
					}
				}
				
				bufferToBoard[status->backBufferReady + 0] = bytes;
				bufferToBoard[status->backBufferReady + 1] = errno1;
				bufferToBoard[status->backBufferReady + 2] = (hsize)?read_buf[head]:0;
				bufferToBoard[status->backBufferReady + 3] = (tsize)?read_buf[tail]:0;
				bufferToBoard[status->backBufferReady + 4] = INDEX_READ;
				bufferToBoard[status->backBufferReady + 5] = 4;

				status->backBufferReady += 6;
				backCount += 6;
#if STDIO_DBG
if((hsize !=0) & (tail !=0)) {
_pthread_mutex_lock(&print_mutex);
/////////////////////////////////
char * c_read_buf = (char *)read_buf;
printf("INDEX_READ c_read_buf[head] ");
for(int d=4-hsize; d<4; d++) printf("read_buf[%d]=%x ",d,c_read_buf[d]);
printf("\n");
///////////////////////////////////////
		sprintf(found, "INDEX_READ   head=%x  tail=%x\n",
			(bufferToBoard[status->backBufferReady + 2]),(bufferToBoard[status->backBufferReady + 3]));
		OUT_DBG(found, trg_file);
		sprintf(found, "INDEX_READ read_buf head=%x tail=%x\n",
			((char*)read_buf)[head], ((char*)read_buf)[tail]);
		OUT_DBG(found, trg_file);
_pthread_mutex_unlock(&print_mutex);
}
#endif

#if STDIO_TRACE
_pthread_mutex_lock(&print_mutex);
sprintf(found, "INDEX_READ  (hsize)?read_buf[head]:0 =0x%x (tsize)?read_buf[tail]:0  =0x%x\n", (hsize)?read_buf[head]:0, (tsize)?read_buf[tail]:0);
OUT_DBG(found, trg_file);
sprintf(found, "INDEX_READ  END bytes=%d\n", bytes);
OUT_DBG(found, trg_file);
_pthread_mutex_unlock(&print_mutex);
#endif
				if (read_buf) free(read_buf);                                  ////// FREE read_buf
				break;
			}
//////////////////////////////////////////////////////////////////////////////////////////
			case INDEX_WRITE:
			{
				int errno1;

#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_WRITE START\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				int remote_fd = bufferFromBoard[0];
				unsigned hsize = bufferFromBoard[1];
				PL_Word head = bufferFromBoard[2];
				unsigned tsize = bufferFromBoard[3];
				PL_Word tail = bufferFromBoard[4];
				unsigned count = bufferFromBoard[5];
				PL_Addr addr = bufferFromBoard[6];
				int local_fd;

				if ((remote_fd < 3) || (remote_fd >= SEMIHOST_MAX_FILE) || \
					((local_fd = SemihostFile_table[remote_fd].fd)<0) ) {
					bufferToBoard[status->backBufferReady + 0] = -1;
					bufferToBoard[status->backBufferReady + 1] = EBADF;
					bufferToBoard[status->backBufferReady + 2] = INDEX_WRITE;
					bufferToBoard[status->backBufferReady + 3] = 2;

					status->backBufferReady += 4;
					backCount += 4;

					break;
				}

				int isText = SemihostFile_table[remote_fd].TextMode;

#if STDIO_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found,"INDEX_WRITE isText=%d fd=%d count=%d\n",
			isText, remote_fd, count);
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				int bytes=0;
				int index = 0;
				unsigned int count_all = hsize + count + tsize;
				if(count_all == 0) {
					bufferToBoard[status->backBufferReady + 0] = 0;
					bufferToBoard[status->backBufferReady + 1] = 0;
					bufferToBoard[status->backBufferReady + 2] = INDEX_WRITE;
					bufferToBoard[status->backBufferReady + 3] = 2;

					status->backBufferReady += 4;
					backCount += 4;
					break;
				}
				// read array from board
				unsigned int count_all_byte= word2byte(count_all);
				PL_Word *write_buf = (PL_Word *)malloc(count_all_byte+4*3);  //// MALLOC write_buf
				 char * char_write_buf = (char *)write_buf;
#if STDIO_DBG
_pthread_mutex_lock(&print_mutex);
sprintf(found, "INDEX_WRITE hsize=%d tsize=%d count=%d count_all=%d\n",hsize,tsize,count,count_all);
OUT_DBG(found, trg_file);
_pthread_mutex_unlock(&print_mutex);
#endif
				if (hsize) {
					write_buf[index] = head;
					char_write_buf += 4 - word2byte(hsize);
					index += 1;
				}
				if ((r= ops->read_mem_block(a, &(write_buf[index]), addr,char2int(count))) != PL_OK) {
					if (write_buf) free(write_buf);			//////// FREE write_buf
					bufferToBoard[status->backBufferReady + 0] = 0;
					bufferToBoard[status->backBufferReady + 1] = BSP_ERROR+r;
					bufferToBoard[status->backBufferReady + 2] = INDEX_WRITE;
					bufferToBoard[status->backBufferReady + 3] = 2;

					status->backBufferReady += 4;
					backCount += 4;
					break;
	//				throw "Error: read_mem_block fail";
				}
				index += char2int(count);
				if (tsize) {
					write_buf[index] = tail;
					index += 1;
				}
				
				if (isText && nmc_core) {
					charTmp = (char *)malloc(index+4);			//// MALLOC charTmp
					memset(charTmp, 0, index + 4);
					sparseCharToChar(charTmp, write_buf, index);
					bytes = write(local_fd,charTmp, index);
					errno1 = errno;
					free(charTmp);                			        ////// FREE charTmp
				}
				else {	//ARM or isBinary
					bytes = 0;
					bytes = write(local_fd, char_write_buf, word2byte(count_all));
					errno1 = errno;
					int losser = lose4div(bytes);
					bytes = byte2word(bytes);

					if (losser) {// NMC binary - io service error in offset. Must be 0
//						SemihostFile_table[remote_fd].error = 1;
						bufferToBoard[status->backBufferReady + 0] = -1;
						bufferToBoard[status->backBufferReady + 1] = EIO;
						bufferToBoard[status->backBufferReady + 2] = INDEX_WRITE;
						bufferToBoard[status->backBufferReady + 3] = 2;

						status->backBufferReady += 4;
						backCount += 4;
						if (write_buf) free(write_buf);                         ////// FREE write_buf
						break;
					}
				}

				bufferToBoard[status->backBufferReady + 0] = bytes;
				bufferToBoard[status->backBufferReady + 1] = errno1;
				bufferToBoard[status->backBufferReady + 2] = INDEX_WRITE;
				bufferToBoard[status->backBufferReady + 3] = 2;

				status->backBufferReady += 4;
				backCount += 4;

#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_WRITE  END bytes=%d error=%d\n", bytes, errno1);
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				if (write_buf) free(write_buf);                   		     ////// FREE write_buf
				break;
			}
/////////////////////////////////////////////////////////////////////////////////////////////
			case INDEX_LSEEK:
			{
				int res;
				int errno1;
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_LSEEK START\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				int remote_fd = bufferFromBoard[0];
				int offset = bufferFromBoard[1];
				int whence = bufferFromBoard[2];
				int local_fd;
				int losser;
#if STDIO_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_LSEEK START fd=%d offset=%d whence=%d\n",
		remote_fd, offset, whence);
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				if ((remote_fd <= 2) || (remote_fd >= SEMIHOST_MAX_FILE) || \
						((local_fd = SemihostFile_table[remote_fd].fd)<0) ) {
					bufferToBoard[status->backBufferReady + 0] = -1;
					bufferToBoard[status->backBufferReady + 1] = EBADF;
					bufferToBoard[status->backBufferReady + 2] = INDEX_LSEEK;
					bufferToBoard[status->backBufferReady + 3] = 2;

					status->backBufferReady += 4;
					backCount += 4;

					break;
				}
				int isText = SemihostFile_table[remote_fd].TextMode;
				// position 
				if (isText && nmc_core) {
					res = lseek(local_fd, offset, whence);
					errno1 = errno;
				}else{ // ARM or isBinary
					long  _offset = word2byte(offset);
					res = lseek(local_fd, _offset, whence);
					errno1 = errno;
					losser = lose4div(res);
					res = byte2word(res);
				}
				bufferToBoard[status->backBufferReady + 0] = res;
				bufferToBoard[status->backBufferReady + 1] = losser ? EIO : errno1;
				bufferToBoard[status->backBufferReady + 2] = INDEX_LSEEK;
				bufferToBoard[status->backBufferReady + 3] = 2;

				status->backBufferReady += 4;
				backCount += 4;
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_LSEEK  END res=%d\n", res);
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif

				break;
			}
//////////////////////////////////////////////////////////////////////
			case INDEX_UNLINK:
			{
				int res, errno1;
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_UNLINK START\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif

				if (nmc_core) {
					charTmp = (char *)charBuf;
					memset(charTmp, 0, pSize);
					sparseCharToChar(charTmp, bufferFromBoard, pSize - 1);
				} else {
					charTmp = (char *)bufferFromBoard;
				}

				res = unlink(charTmp);
				errno1 = errno;

				bufferToBoard[status->backBufferReady + 0] = res;
				bufferToBoard[status->backBufferReady + 1] = errno1;
				bufferToBoard[status->backBufferReady + 2] = INDEX_UNLINK;
				bufferToBoard[status->backBufferReady + 3] = 2;

				status->backBufferReady += 4;
				backCount += 4;
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_UNLINK END\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				break;
			}
//////////////////////////////////////////////////////////////////////
			case INDEX_LINK:
			{
				int res;
				char *old_name;
				char *new_name;
				size_t old_size;
				size_t new_size;
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_LINK START\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif

				old_size = bufferFromBoard[0];
				charTmp = (char *)charBuf;
				int index = char2int(old_size) + 1;

				if (nmc_core)
					sparseCharToChar(charTmp, &bufferFromBoard[1], old_size);
				else
					charTmp = (char *)&bufferFromBoard[1];

				old_name = charTmp;
				charTmp += char2int(old_size) * bINw;
				new_size = bufferFromBoard[index++];

				if (nmc_core)
					sparseCharToChar(charTmp, &bufferFromBoard[index], new_size);
				else
					charTmp = (char *)&bufferFromBoard[index];

				new_name = charTmp;

				bufferToBoard[status->backBufferReady + 0] = link(old_name, charTmp);
				bufferToBoard[status->backBufferReady + 1] = errno;
				bufferToBoard[status->backBufferReady + 2] = INDEX_LINK;
				bufferToBoard[status->backBufferReady + 3] = 2;

				status->backBufferReady += 4;
				backCount += 4;
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_LINK END\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				break;
			}
//////////////////////////////////////////////////////////////////////
			case INDEX_FSTAT:
			{
				int res, errno1 = 0;
				int remote_fd = bufferFromBoard[0];
				struct stat st = {0};
				int sz = 0;
				int local_fd;

#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_FSTAT START\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				
				if ((remote_fd <= 2) || (remote_fd >= SEMIHOST_MAX_FILE) || \
							((local_fd = SemihostFile_table[remote_fd].fd) < 0)	) {
					bufferToBoard[status->backBufferReady + 0] = -1;
					bufferToBoard[status->backBufferReady + 1] = EBADF;
					bufferToBoard[status->backBufferReady + 2] = 'w';
					sz = 0;
					bufferToBoard[status->backBufferReady + sz + 3] = INDEX_FSTAT;
					bufferToBoard[status->backBufferReady + sz + 4] = sz + 3;

					status->backBufferReady += (sz + 5);
					backCount += (sz + 5);

					break;
				}

				res = fstat(local_fd, &st);
				errno1 = errno;
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_FSTAT local_fd=%d st[0]=%d st[7]=%d\n",local_fd,st.st_dev,st.st_size);
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif			
				bufferToBoard[status->backBufferReady + 0] = res;
				bufferToBoard[status->backBufferReady + 1] = errno1;
				bufferToBoard[status->backBufferReady + 2] = 'l'; //Result from host-linux
				if (res == 0) {
					sz = wstat2lstat(&st, &bufferToBoard[status->backBufferReady + 3], 'l');
				} else sz=0;
				bufferToBoard[status->backBufferReady + sz + 3] = INDEX_FSTAT;
				bufferToBoard[status->backBufferReady + sz + 4] = sz + 3;

				status->backBufferReady += (sz + 5);
				backCount += (sz + 5);
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_FSTAT END\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				break;
			}
//////////////////////////////////////////////////////////////////////
			case INDEX_STAT:
			{
				int res, errno1 = 0;
				struct stat st = {0};
				char *file;
				int sz;

#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found,"INDEX_STAT START\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif

				if (nmc_core) {
					charTmp = (char *)charBuf;
					memset(charTmp, 0, pSize);
					sparseCharToChar(charTmp, bufferFromBoard, pSize - 1);
				}
				else {
					charTmp = (char *)bufferFromBoard;
				}
				file = (char *)charTmp;
				
				res = stat(charTmp, &st);
				errno1 = errno;

				bufferToBoard[status->backBufferReady + 0] = res;
				bufferToBoard[status->backBufferReady + 1] = errno1;
				bufferToBoard[status->backBufferReady + 2] = 'l';
				if (res == 0) {
					sz = wstat2lstat(&st, &bufferToBoard[status->backBufferReady + 3], 'l');
				} else sz=0;
				bufferToBoard[status->backBufferReady + sz + 3] = INDEX_STAT;
				bufferToBoard[status->backBufferReady + sz + 4] = sz + 3;

				status->backBufferReady += (sz + 5);
				backCount += (sz + 5);
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "INDEX_STAT END\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
				break;
			}
//////////////////////////////////////////////////////////////////////
			case INDEX_ISATTY:
			{
				int res, errno1 = 0;
				int remote_fd = bufferFromBoard[0];
				int local_fd;
				
#if STDIO_TRACE
				_pthread_mutex_lock(&print_mutex);
				sprintf(found, "INDEX_ISATTY START\n");
				OUT_DBG(found, trg_file);
				_pthread_mutex_unlock(&print_mutex);
#endif
				if ((remote_fd <= 2) || (remote_fd >= SEMIHOST_MAX_FILE) ||
						((local_fd = SemihostFile_table[remote_fd].fd)<0)) {
					bufferToBoard[status->backBufferReady + 0] = 0;
					bufferToBoard[status->backBufferReady + 1] = EBADF;
					bufferToBoard[status->backBufferReady + 2] = INDEX_ISATTY;
					bufferToBoard[status->backBufferReady + 3] = 2;

					status->backBufferReady += 4;
					backCount += 4;

					break;
				}

				res = isatty(local_fd);
				errno1 = errno;
				if (res != 0)  res = 1;

				bufferToBoard[status->backBufferReady + 0] = res;
				bufferToBoard[status->backBufferReady + 1] = errno1;
				bufferToBoard[status->backBufferReady + 2] = INDEX_ISATTY;
				bufferToBoard[status->backBufferReady + 3] = 2;

				status->backBufferReady += 4;
				backCount += 4;
#if STDIO_TRACE
				_pthread_mutex_lock(&print_mutex);
				sprintf(found, "INDEX_ISATTY END\n");
				OUT_DBG(found, trg_file);
				_pthread_mutex_unlock(&print_mutex);
#endif
				break;
			}
			default:
				if (pKind == (IO_ServiceID)(~0))
					break;
				printf("Unknown service called pKind=%d\n", pKind);
//				cerr << endl << "unknown service called" << endl;
				throw( "unknown service called" );
			};

//			bufferFromBoard += ((pSize + bINw - 1) / bINw);
//			words -= ((pSize + bINw - 1) / bINw);
			bufferFromBoard += char2int(pSize);
			words -= char2int(pSize);

			if (pKind != (IO_ServiceID)(~0))
				status->hostReceive++;

#if IO_SERV_DBG
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "EEE END pKind=0x%x pSize=0x%x bufferFromBoard=0x%lx"
					" hostReceive=0x%x  boardSend=0x%x words=0x%x\n",
					pKind, pSize, bufferFromBoard, status->hostReceive, status->boardSend, words);
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif
		} // END while( hostReceive != boardSend )

		if ( (status->backBufferReady - backCount) != 0){
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "!!!!!!!!!!!!!! PROBLEM  backBufferReady !!!!!!!!!!!!!!!!\n"\
						"        (status->backBufferReady - backCount) != 0\n"\
						"        status->backBufferReady=%d  backCount=%d\n",
						status->backBufferReady, backCount);
			OUT_DBG(found, trg_file);
			_pthread_mutex_unlock(&print_mutex);
		}
		if (backCount){
			r = ops->write_mem_block(a, &bufferToBoard[status->backBufferReady - backCount],
			bssSecAddr+(b2hBufSize + status->backBufferReady - backCount)*bINw, backCount);
			if (r != PL_OK){
				io_service_fail.err = r;
				io_service_fail.function = BSP_write;
				throw "IO_Service dispatcher: error writing block to board";
			}
		}

		writeBackStatus();
		if (charBuf) {
			delete[] charBuf;
			charBuf = NULL;
		}
	} //END while (proceed) 

#if IO_SERV_DBG
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "===fin===\n");
	OUT_DBG(found, trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif

	return 0;
}

static int wstat2lstat(struct stat* st, uint32_t* stat, int ch)
{
	if(ch == 'w') { // Windows struct stat
		stat[0] = st->st_dev;
		stat[1] = st->st_mode;
		stat[2] = st->st_nlink;
		stat[3] = st->st_rdev;
		stat[4] = st->st_size;
		stat[5] = st->st_atime;
		stat[6] = st->st_mtime;
		stat[7] = st->st_ctime;

		return 8;
	}

	if (ch == 'l') { // Linux struct stat
		stat[0] = st->st_dev;
		stat[1] = st->st_ino;
		stat[2] = st->st_mode;
		stat[3] = st->st_nlink;
		stat[4] = st->st_uid;
		stat[5] = st->st_gid;
//		int __pad0; //???????????????????????
		stat[6] = st->st_rdev;		// тип устройства
		stat[7] = st->st_size;		// общий размер в байтах 
		stat[8] = st->st_blksize;	// размер блока ввода-вывода  
		stat[9] = st->st_blocks;	// количество выделенных блоков 
		stat[10] = st->st_atime;	// время последнего доступа 
		stat[11] = st->st_mtime;	// время последней модификации 
		stat[12] = st->st_ctime;	// время последнего изменения 

		return 13;
	}

	return -1;
}
///////////////////////////////////////////////////////////////////////////////

int  IOService_finish(void*v_nms)
{
	IO_Service& trg = *static_cast<IO_Service*>(v_nms);
	struct ll_operation *ops = (struct ll_operation *)trg.ll_ops;
	DWORD status = 0, res_status, attr=0;
	int res = 0;
	int count;

	// start condition
	trg.IOService_finish_end = false;
	trg.IOService_finish_start = true;
	trg.io_service_fail.function = 0;
	trg.io_service_fail.err = 0;
	
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "IOService_finish start _force=%d _time=%d _end=%d\n",
		trg.destructor_force, trg.destructor_time, trg.dispatch_end);
	OUT_DBG(found, trg.trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif  

	if (trg.destructor_force == 0) {
		if ((res_status = ops->get_status(trg.a, &status)) != PL_OK) {
#if STDIO_TRACE
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "IOService_finish get_status error =%d\n", res_status);
			OUT_DBG(found, trg.trg_file);
			_pthread_mutex_unlock(&print_mutex);
#endif
			trg.io_service_fail.function = BSP_status;
			trg.io_service_fail.err = res_status;
			trg.IOService_finish_end = true;
			return -1;
		}

#if STDIO_TRACE
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "IOService_finish  status(%d) != PROGRAM_RUNNING(%d)  dispatch_end=%d\n", 
			status,PROGRAM_RUNNING, trg.dispatch_end);
		OUT_DBG(found, trg.trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif

		if(status == PROGRAM_RUNNING){
			trg.IOService_finish_end = true;
			return IO_SERV_PROG_RUN;
		}
	}
	if (!trg.dispatch_end) { // Here dispatcher is working. Need to wait or to stop it. 
#if STDIO_TRACE
		_pthread_mutex_lock(&print_mutex);
		sprintf(found, "IOService_finish dispatch_end=%d\n", trg.dispatch_end);
		OUT_DBG(found, trg.trg_file);
		_pthread_mutex_unlock(&print_mutex);
#endif

		trg.letsShutDown = true; //order to finish "dispatch"

//		res = WaitForSingleObject(trg.io_thread, trg.destructor_time); 
		//Expected after destructor_time mlsec. dispatch is finished.
		//Default 10000 mlsec
		for( count=trg.destructor_time; count>0; count-=10){ //time in mlsec
			usleep(10000); //  wait finishing IOServiceDispatcher 10mlsec
			if (trg.dispatch_end) break;
		}
		if(count <= 0){  // Timeout. Here the thread is not finished
#if STDIO_TRACE
	_pthread_mutex_lock(&print_mutex);
	sprintf(found, "IOService_finish: thread is not finished "\
				"Time_Destruct=%d\n",
				Get_IO_Serv_attribute(&trg, Att_Time_Destruct, NULL));
	OUT_DBG(found, trg.trg_file);
	_pthread_mutex_unlock(&print_mutex);
#endif
			if (trg.destructor_force) {
#if STDIO_TRACE
				_pthread_mutex_lock(&print_mutex);
				sprintf(found, "IOService_finish: Force TerminateThread\n");
				OUT_DBG(found, trg.trg_file);
				_pthread_mutex_unlock(&print_mutex);
#endif
				if (!trg.dispatch_end) { //NOT end
					pthread_cancel(trg.io_thread);
					pthread_join(trg.io_thread, NULL);
				}
				trg.IOService_finish_end = true;
			}else {// not destructor_force
				//IO_Service thread is not finished during set Time_Destruct";
				trg.IOService_finish_end = true;
				return IO_SERV_NOT_FINISHED;
			}
		} else { ;
#if STDIO_TRACE
			_pthread_mutex_lock(&print_mutex);
			sprintf(found, "IOService_finish: finish dispatch_end=%d\n", trg.dispatch_end);
			OUT_DBG(found, trg.trg_file);
			_pthread_mutex_unlock(&print_mutex);;
#endif
		}
	} // dispatch_end == true
	trg.IOService_finish_end = true;
	return 0;
}
//////////////////////////////////////////////////////////////
void robust_mutex_constructor(void)
{
	pthread_mutexattr_init(&mutexattr);

	pthread_mutexattr_setrobust(&mutexattr,PTHREAD_MUTEX_ROBUST);
	pthread_mutexattr_settype(&mutexattr,PTHREAD_MUTEX_ERRORCHECK);

	pthread_mutex_init(&print_mutex, &mutexattr);
	pthread_mutex_init(&works_mutex, &mutexattr);
}

void robust_mutex_destructor(void)
{
	pthread_mutex_destroy(&print_mutex);
	pthread_mutex_destroy(&works_mutex);

	pthread_mutexattr_destroy(&mutexattr);
}
