#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sstream>
#include <sys/time.h>
#include <net/if.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "libelf.h"
#include "libload.h"

#ifndef PLOAD_MAJOR_VERSION
#define PLOAD_MAJOR_VERSION 2
#endif

#ifndef PLOAD_MINOR_VERSION
#define PLOAD_MINOR_VERSION 1
#endif

#ifndef BUILD_DAT
#define BUILD_DAT 0
#endif

#ifndef BUILD_NUM
#define BUILD_NUM 0
#endif

#define _STRINGIFY(X) #X
#define STRINGIFY(X) _STRINGIFY(X)

#define PLOAD_VERSION STRINGIFY(PLOAD_MAJOR_VERSION) "." STRINGIFY(PLOAD_MINOR_VERSION) "." STRINGIFY(BUILD_DAT) "." STRINGIFY(BUILD_NUM)

#ifndef BUILD_Y
#define BUILD_Y 0
#endif

#ifndef BUILD_M
#define BUILD_M 0
#endif

#ifndef BUILD_D
#define BUILD_D 0
#endif

#ifndef BUILD_N
#define BUILD_N 0
#endif

#define PLOAD_MAJOR_MINOR	((PLOAD_MAJOR_VERSION << 16) | (PLOAD_MINOR_VERSION & 0xFFFF))
#define PLOAD_DATE_REVID	((((BUILD_Y << 9) | (BUILD_M << 5) | (BUILD_D << 0)) << 16) | (BUILD_N & 0xFFFF))

#ifdef __cplusplus
extern "C"
{
#endif

unsigned char local_mac_addr[EDCL_MAC_ADDR_LEN];
unsigned char remote_mac_addr[EDCL_MAC_ADDR_LEN];
int use_edcl = 0;

static pthread_mutex_t ll_mutex = PTHREAD_MUTEX_INITIALIZER;
static DWORD globalTimeout = 0;

DWORD __attribute__ ((visibility ("default"))) PL_BuildVersion = PLOAD_DATE_REVID;

int PL_GetVersion(unsigned int *version_major, unsigned int *version_minor)
{
	PL_DBG("Start.");
	PL_DBG(PLOAD_VERSION);

	if (version_major)
		*version_major = PLOAD_MAJOR_VERSION; // PLOAD_MAJOR_MINOR

	if (version_minor)
		*version_minor = PLOAD_MINOR_VERSION; // PLOAD_DATE_REVID

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

int PL_SetChannelEDCL(const unsigned char host_mac_addr[], const unsigned char board_mac_addr[])
{
	PL_DBG("Start.");

	LOCK_LL;

	for (int i = 0; i < EDCL_MAC_ADDR_LEN; i++) {
		local_mac_addr[i] = host_mac_addr[i];
		remote_mac_addr[i] = board_mac_addr[i];
	}

	use_edcl = 1;

	UNLOCK_LL;

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

int PL_GetBoardCount(unsigned int *count)
{
	int ret;

	PL_DBG("Start.");

	LOCK_LL;
	if (!!use_edcl) {
		PL_DBG("ERROR: Getting the count of boards is not supported for EDCL!");
		PL_DBG("End (With ERROR!); Return: PL_NOT_IMPLEMENTED!");

		UNLOCK_LL;
		return PL_NOT_IMPLEMENTED;
	}
	UNLOCK_LL;

	if (count) {
		*count = ll_mutex_count;

		PL_DBG("End; Return: PL_OK.");
		ret = PL_OK;
	}
	else {
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		ret = PL_ERROR;
	}

	return ret;
}

int PL_GetBoardDesc(unsigned int index, PL_Board **board)
{
	int i, ret;

	PL_DBG("Start; index = %u.", index);

	LOCK_LL;

	if (!use_edcl && (index >= NM6408_BRD_NUM)) {
		PL_DBG("ERROR: Invalid argument - 'index'!");

		ret = PL_ERROR;
		goto err_pl_exit;
	}

	if (!use_edcl) {
		for (i = 0; i < ll_mutex_count; i++) {
			if (index == ll_mutex_index[i])
				break;
		}

		if (i == ll_mutex_count) {
			PL_DBG("ERROR: Board with index = %u not found!", index);
			ret = PL_ERROR;
			goto err_pl_exit;
		}
	}

	PL_Board *newBoard;

	try {
		if (!!use_edcl)
			newBoard = new PL_Board(~0);
		else
			newBoard = new PL_Board(index);
	}
	catch (...) {
		PL_DBG("ERROR: Failed create board descriptor!");
		ret = PL_ERROR;
		goto err_pl_exit;
	}

	if (board)
		*board = newBoard;
	else
		delete newBoard;

	ret = PL_OK;

err_pl_exit:

	UNLOCK_LL;

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_CloseBoardDesc(PL_Board *board)
{
	int ret;

	PL_DBG("Start.");

	LOCK_LL;

	if (Check_board(board)) {
		PL_DBG("ERROR: Invalid 'board'!");
		ret = PL_ERROR;
		goto err_pl_exit;
	}

	if (board->Count_In_Progress) {
		PL_DBG("ERROR: Board has %u operation(s) in"
			" progress!", board->Count_In_Progress);
		ret = PL_AGAIN;
		goto err_pl_exit;
	}

	if (board->Count_PL_Access) {
		PL_DBG("ERROR: Board has %u opened processor"
			" descriptor!", board->Count_PL_Access);
		ret = PL_AGAIN;
		goto err_pl_exit;
	}

	delete board;
	ret = PL_OK;

err_pl_exit:

	UNLOCK_LL;

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else if (ret == PL_AGAIN)
		PL_DBG("End (With ERROR!); Return: PL_AGAIN!");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_GetPostStatus(PL_Board *board, int *status)
{
	PL_CoreNo	core;
	DWORD		post;
	int		ret;

	PL_DBG("Start.");

	if (!status) {
		PL_DBG("ERROR: Invalid 'status' (NULL)!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	LOCK_LL;
	ret = Check_board(board);
	if (!ret)
		board->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'board'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	core.cluster_id = NM6408_CCPU;
	core.nm_id = -1;

	ret = board->Command(&core, GET_POST_STATUS, &post);
	if (ret) {
		PL_DBG("ERROR: Unable get POST status, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	*status = (int)post;
	ret = PL_OK;

err_pl_exit1:

	LOCK_LL;
	board->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_GetSerialNumber(PL_Board *board, unsigned long *serial_number)
{
	PL_CoreNo	core;
	DWORD		sn_lo, sn_hi;
	QWORD		sn;
	int		ret;

	PL_DBG("Start.");

	if (!serial_number) {
		PL_DBG("ERROR: Invalid 'serial_number' (NULL)!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	LOCK_LL;
	ret = Check_board(board);
	if (!ret)
		board->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'board'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	if (board->is_used_edcl) {
		PL_DBG("ERROR: Getting serial number not supported for EDCL!");
		ret = PL_NOT_IMPLEMENTED;
		goto err_pl_exit1;
	}

	core.cluster_id = NM6408_CCPU;
	core.nm_id = -1;

	ret = board->Command(&core, GET_SERIAL_NMB_LO, &sn_lo);
	if (ret) {
		PL_DBG("ERROR: Unable get serial number, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = board->Command(&core, GET_SERIAL_NMB_HI, &sn_hi);
	if (ret) {
		PL_DBG("ERROR: Unable get serial number, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	sn = sn_hi & 0xFFFFFFFF;
	sn = (sn << 32) | (sn_lo & 0xFFFFFFFF);

	*serial_number = (unsigned long)sn;

	ret = PL_OK;

err_pl_exit1:

	LOCK_LL;
	board->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else if (ret == PL_NOT_IMPLEMENTED)
		PL_DBG("End (With ERROR!); Return: PL_NOT_IMPLEMENTED!");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

DWORD __attribute__ ((visibility ("default"))) PL_FirmwareVersion = 0;

int PL_GetFirmwareVersion(PL_Board *board, unsigned int *version_major,
			  unsigned int *version_minor)
{
	PL_CoreNo	core;
	DWORD		version;
	int		ret;

	PL_DBG("Start.");

	if (!version_major || !version_minor) {
		PL_DBG("ERROR: Invalid 'version_major' and/or"
			" 'version_minor' (NULL)!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	LOCK_LL;
	ret = Check_board(board);
	if (!ret)
		board->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'board'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	if (board->is_used_edcl) {
		PL_DBG("ERROR: Getting firmware version not supported for EDCL!");
		ret = PL_NOT_IMPLEMENTED;
		goto err_pl_exit1;
	}

	core.cluster_id = NM6408_CCPU;
	core.nm_id = -1;

	ret = board->Command(&core, GET_VERSION_MAJOR, &version);
	if (ret) {
		PL_DBG("ERROR: Unable get version major, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	*version_major = (version >> 16) & 0xFFFF;
	*version_minor = version & 0xFFFF;

	ret = board->Command(&core, GET_VERSION_MINOR, &version);
	if (ret) {
		PL_DBG("ERROR: Unable get version minor, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	PL_FirmwareVersion = version;
	ret = PL_OK;

err_pl_exit1:

	LOCK_LL;
	board->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else if (ret == PL_NOT_IMPLEMENTED)
		PL_DBG("End (With ERROR!); Return: PL_NOT_IMPLEMENTED!");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_ResetBoard(PL_Board *board)
{
	PL_CoreNo	core;
	int		ret;

	PL_DBG("Start.");

	LOCK_LL;
	ret = Check_board(board);
	if (!ret)
		board->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'board'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	core.cluster_id = NM6408_CCPU;
	core.nm_id = -1;

	ret = board->Command(&core, SW_RESET_BOARD, NULL);
	if (ret) {
		PL_DBG("ERROR: Unable reset board, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = PL_OK;

err_pl_exit1:

	LOCK_LL;
	board->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_LoadInitCode(PL_Board *board)
{
	PL_CoreNo	core;
	DWORD           *load_prm[2];
	int		ret;

	PL_DBG("Start.");

	LOCK_LL;
	ret = Check_board(board);
	if (!ret)
		board->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'board'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	core.cluster_id = NM6408_CCPU;
	core.nm_id = -1;

	const char *DistribPath;
	char FullName[PATH_MAX];
	PL_Word *FileBuf;
	ssize_t FileSize;
	int	FileDesc;

	DistribPath = getenv(ENV_PATH_NAME);
	if (!DistribPath)
		DistribPath = LOCAL_PATH_NAME;

	if (!board->is_used_edcl)
		sprintf(FullName, "%s/bin/%s", DistribPath, INIT_PCIE_FILE_NAME);
	else
		sprintf(FullName, "%s/bin/%s", DistribPath, INIT_EDCL_FILE_NAME);

	FileDesc = open(FullName, O_RDONLY, 0);
	if (FileDesc < 0) {
		PL_DBG("ERROR: Unable open file \"%s\"!", FullName);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	FileSize = lseek(FileDesc, 0, SEEK_END);
	if (FileSize <= 0) {
		close(FileDesc);

		if (!board->is_used_edcl)
			PL_DBG("ERROR: Unable get size of file \"%s\"!", INIT_PCIE_FILE_NAME);
		else
			PL_DBG("ERROR: Unable get size of file \"%s\"!", INIT_EDCL_FILE_NAME);

		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	if (lseek(FileDesc, 0, SEEK_SET) < 0) {
		close(FileDesc);

		if (!board->is_used_edcl)
			PL_DBG("ERROR: Unable set position in file \"%s\"!", INIT_PCIE_FILE_NAME);
		else
			PL_DBG("ERROR: Unable set position in file \"%s\"!", INIT_EDCL_FILE_NAME);

		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	FileBuf = new (std::nothrow) PL_Word[((FileSize + 3) >> 2)];
	if (!FileBuf) {
		close(FileDesc);

		PL_DBG("ERROR: Unable allocate memory!");
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	if (read(FileDesc, FileBuf, FileSize) < FileSize) {
		close(FileDesc);
		delete[] FileBuf;

		PL_DBG("ERROR: Failure read file!");
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	close(FileDesc);

	load_prm[0] = (DWORD *)FileBuf;
	load_prm[1] = (DWORD *)(((FileSize + 3) >> 2) << 2);

	ret = board->Command(&core, LOAD_INIT_CODE, (DWORD *)load_prm);
	if (ret) {
		delete[] FileBuf;

		PL_DBG("ERROR: Unable load init code, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	delete[] FileBuf;
	ret = PL_OK;

err_pl_exit1:

	LOCK_LL;
	board->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_GetAccess(PL_Board *board, PL_CoreNo *coreNo, PL_Access **access)
{
	int ret;

	PL_DBG("Start.");

	if (!access || !coreNo) {
		PL_DBG("ERROR: Invalid 'access' and/or 'coreNo' (NULL)!");

		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	PL_DBG("CoreNo; cluster_id = %d, nm_id = %d.",
		coreNo->cluster_id, coreNo->nm_id);

	LOCK_LL;

	if (Check_board(board)) {
		PL_DBG("ERROR: Invalid 'board'!");
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	PL_Access *newAccess;

	try {
		newAccess = new PL_Access(coreNo, board);
	}
	catch (...) {
		PL_DBG("ERROR: Failed create processor descriptor!");
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	board->Count_PL_Access++;
	*access = newAccess;

	ret = PL_OK;

err_pl_exit1:

	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_TypeAccess(PL_Access *access, DWORD *isNMC)
{
	int ret;

	PL_DBG("Start.");

	if (!isNMC) {
		PL_DBG("ERROR: Invalid 'isNMC' parameter (NULL)!"); 
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
		return PL_ERROR;
	}

	LOCK_LL;

	if (Check_access(access)) {
		PL_DBG("ERROR: Invalid 'access'!");
		ret = PL_ERROR;
		goto err_pl_exit;
	}

	if (access->coreNo.nm_id != -1)
		*isNMC = 1;
	else
		*isNMC = 0;

	ret = PL_OK;

err_pl_exit:

	UNLOCK_LL;

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_CloseAccess(PL_Access *access)
{
	int ret;

	PL_DBG("Start.");

	LOCK_LL;

	if (Check_access(access)) {
		PL_DBG("ERROR: Invalid 'access'!");
		ret = PL_ERROR;
		goto err_pl_exit;
	}

	if (access->Count_In_Progress) {
		PL_DBG("ERROR: Access has %u operation(s) in"
			" progress!", access->Count_In_Progress);
		ret = PL_AGAIN;
		goto err_pl_exit;
	}

	access->board->Count_PL_Access--;
	delete access;

	ret = PL_OK;

err_pl_exit:

	UNLOCK_LL;

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else if (ret == PL_AGAIN)
		PL_DBG("End (With ERROR!); Return: PL_AGAIN!");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_SetProgramArgs(PL_Access *, const char *) __attribute__ ((visibility ("default")));
int PL_SetProgramArgs(PL_Access *access, const char *mainargs)
{
	int i, j, ret;

	PL_DBG("Start.");

	if (!mainargs) {
		PL_DBG("ERROR: Invalid 'mainargs' (NULL)!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	LOCK_LL;
	ret = Check_access(access);
	if (!ret)
		access->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'access'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	libload_mutex_lock(access->board->b_mutex_map,
			access->coreNo.cluster_id, access->coreNo.nm_id);

	DWORD status;

	ret = access->board->Command(&access->coreNo, GET_LOADER_STATUS, &status);
	if (ret) {
		PL_DBG("ERROR: Unable get loader status, ret = %d", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	if (access->MainARGC || (status != PROGRAM_NO_LOAD)) {
		PL_DBG("ERROR: Program is started or args be set early!");
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	access->MainARGS = new (std::nothrow) char[strlen(mainargs) + 1];
	if (access->MainARGS == NULL) {
		PL_DBG("ERROR: ERROR not memory allocate to string!");
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	for (i = 0, j = 0; (i < strlen(mainargs)) && (access->MainARGC < MAIN_ARGS_MAX); i++) {
		if (mainargs[i] != ' ') {
			access->MainARGS[i] = mainargs[i];
			continue;
		}
		else if (mainargs[j] == ' ') {
			j = i + 1;
			continue;
		}

		access->MainARGV[access->MainARGC] = &access->MainARGS[j];
		access->MainARGS[i] = '\0';

		access->MainARGC++;
		j = i + 1;
	}

	if (i == strlen(mainargs)) {
		if (mainargs[j] != '\0') {
			access->MainARGV[access->MainARGC] = &access->MainARGS[j];
			access->MainARGS[i] = '\0';

			access->MainARGC++;
		}
		else {
			access->MainARGV[access->MainARGC] = NULL;

			if (!access->MainARGC) {
				delete[] access->MainARGS;
				access->MainARGS = NULL;
			}
		}
	}
	else {
		delete[] access->MainARGS;

		access->MainARGS = NULL;
		access->MainARGC = 0;

		PL_DBG("ERROR: ERROR not memory allocate to arguments!");

		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = PL_OK;

err_pl_exit1:

	libload_mutex_unlock(access->board->b_mutex_map,
			access->coreNo.cluster_id, access->coreNo.nm_id);

	LOCK_LL;
	access->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_LoadProgramFileArgs(PL_Access *, const char *, const char *) __attribute__ ((visibility ("default")));
int PL_LoadProgramFileArgs(PL_Access *access, const char *filename, const char *mainArgs) 
{
	int ret;

	if ((ret = PL_SetProgramArgs(access, mainArgs)) == PL_OK)
		ret = PL_LoadProgramFile(access, filename);

	return ret;
}

int PL_LoadProgramArgs(PL_Access *, const void *, unsigned int, const char *) __attribute__ ((visibility ("default")));
int PL_LoadProgramArgs(PL_Access *access, const void *addrProgram, unsigned int sizeProgram, const char *mainArgs)
{
	int ret;

	if ((ret = PL_SetProgramArgs(access, mainArgs)) == PL_OK)
		ret = PL_LoadProgram(access, addrProgram, sizeProgram);

	return ret;
}

static int PL_StartCore(PL_Access *access, PL_Addr entry, DWORD status)
{
	int ret;

	DWORD start_addr = entry;
	ret = access->board->Command(&access->coreNo, SET_ENTRY_POINT, &start_addr);
	if (ret) {
		PL_DBG("ERROR: Failed set entry point, ret = %d!", ret);
		return 1;
	}

	DWORD new_status = PROGRAM_STARTED;
	ret = access->board->Command(&access->coreNo, SET_LOADER_STATUS, &new_status);
	if (ret) {
		PL_DBG("ERROR: Failed set loader status, ret = %d!", ret);
		return 2;
	}

	DWORD command = LIBLOAD_RUN_PROGRAM;
	if ((access->coreNo.cluster_id == NM6408_CCPU) || ((access->coreNo.nm_id == -1) && (status == PROGRAM_STOPPED))) {
		ret = access->board->Command(&access->coreNo, SET_LOADER_COMMAND, &command);

		if (ret) {
			PL_DBG("ERROR: Failed set loader command, ret = %d!", ret);
			return 3;
		}
	}
	else {
		if ((access->coreNo.nm_id != -1) && (status == PROGRAM_STOPPED))
			ret = access->board->Command(&access->coreNo, SW_GEN_NMI, &start_addr);
		else
			ret = access->board->Command(&access->coreNo, SW_START_CORE, &start_addr);

		if (ret) {
			PL_DBG("ERROR: Failed start core, ret = %d!", ret);
			return 3;
		}
	}

	return 0;
}

int PL_RunProgram(PL_Access *access, PL_Addr startAddr)
{
	int ret;

	LOCK_LL;
	ret = Check_access(access);
	if (!ret)
		access->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'access'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	libload_mutex_lock(access->board->b_mutex_map,
			access->coreNo.cluster_id, access->coreNo.nm_id);

	DWORD status;

	ret = access->board->Command(&access->coreNo, GET_LOADER_STATUS, &status);
	if (ret) {
		PL_DBG("ERROR: Unable get loader status, ret = %d", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	if ((status != PROGRAM_NO_LOAD) && (status != PROGRAM_STOPPED)) {
		PL_DBG("ERROR: Program in progress"
			" (status = 0x%8.8X)!", status);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = PL_StartCore(access, startAddr, status);
	if (ret) {
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = PL_OK;

err_pl_exit1:

	libload_mutex_unlock(access->board->b_mutex_map,
			access->coreNo.cluster_id, access->coreNo.nm_id);

	LOCK_LL;
	access->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

static int PL_LoadELF(PL_Access *access, ELF *ElfFile, PL_Addr *entry, bool run_it)
{
	int ret;

	LOCK_LL;
	ret = Check_access(access);
	if (!ret)
		access->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'access'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	libload_mutex_lock(access->board->b_mutex_map,
			   access->coreNo.cluster_id, access->coreNo.nm_id);

	DWORD status;

	ret = access->board->Command(&access->coreNo, GET_LOADER_STATUS, &status);
	if (ret) {
		PL_DBG("ERROR: Unable get loader status, ret = %d", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	if ((status != PROGRAM_NO_LOAD) && (status != PROGRAM_STOPPED)) {
		PL_DBG("ERROR: Program in progress"
			" (status = 0x%8.8X)!", status);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	try {
		if (ElfFile->Kind() != ELF_K_ELF)
			throw ELF_ERROR("Wrong kind of file");

		Elf32_Ehdr *Ehdr = ElfFile->GetEhdr();

		if ((access->coreNo.nm_id != -1) &&
		    !((Ehdr->e_machine == EM_NM640X) &&
		    ((Ehdr->e_flags == EM_NM640X_NMC4) ||
		    (Ehdr->e_flags == EM_NM640X_NMC4F)))) {
			PL_DBG("e_machine = 0x%x; e_flags = 0x%x",
				Ehdr->e_machine, Ehdr->e_flags);
			throw ELF_ERROR("Wrong NM machine type");
		}

		if ((access->coreNo.nm_id == -1) &&
		    (Ehdr->e_machine != EM_ARM)) {
			PL_DBG("e_machine = 0x%x", Ehdr->e_machine);
			throw ELF_ERROR("Wrong ARM machine type");
		}

		if (Ehdr->e_type != ET_EXEC)
			throw ELF_ERROR("Bad file type");

		if (!Ehdr->e_phnum)
			throw ELF_ERROR("Program has not program segments");

		Elf32_Phdr *Phdr = ElfFile->GetPhdr();
		PL_Addr Start_Addr = Ehdr->e_entry;

		PL_DBG("Entry point 0x%8.8X", Start_Addr);

		if (access->coreNo.nm_id != -1)
			access->entry_point = (QWORD)Start_Addr << 2;
		else
			access->entry_point = (QWORD)Start_Addr;

		for (DWORD i = 0; i < Ehdr->e_phnum; i++) {
			if (Phdr[i].p_type != PT_LOAD)
				continue;

			if ((access->coreNo.nm_id != -1) &&
			    !(Phdr[i].p_memsz >> 2)) {
				continue;
			}

			if ((access->coreNo.nm_id == -1) &&
			    (Phdr[i].p_memsz == 0)) {
				continue;
			}

#ifdef LIBLOAD_DEBUG
			PL_DBG("Loading segment %u:", i);
			PL_DBG("Address:            0x%8.8X",
				Phdr[i].p_paddr);

			if (access->coreNo.nm_id != -1) {
				PL_DBG("Length (in file):   0x%8.8X",
					Phdr[i].p_filesz >> 2);
				PL_DBG("Length (in memory): 0x%8.8X",
					Phdr[i].p_memsz >> 2);
			}
			else {
				PL_DBG("Length (in file):   0x%8.8X",
					Phdr[i].p_filesz);
				PL_DBG("Length (in memory): 0x%8.8X",
					Phdr[i].p_memsz);
			}
#endif

			QWORD load_addr;

			if (access->coreNo.nm_id != -1)
				load_addr = (QWORD)Phdr[i].p_paddr << 2;
			else
				load_addr = (QWORD)Phdr[i].p_paddr;

			QWORD check_size = (QWORD)((Phdr[i].p_memsz + 3) >> 2) << 2;

#ifdef FILL_WITH_ZERO
			QWORD load_size = (QWORD)((Phdr[i].p_memsz + 3) >> 2) << 2;
			PL_DBG("Segments fill with zero at end (blockSize = 0x%8.8LX).",
				load_size);
#else
			QWORD load_size = (QWORD)((Phdr[i].p_filesz + 3) >> 2) << 2;
			PL_DBG("Segments not fill with zero at end (blockSize = 0x%8.8LX).",
				load_size);
#endif

			if (!load_size)
				continue;

			DWORD *Buffer;

			try {
				Buffer = new DWORD[load_size >> 2];
			}
			catch (...) {
				if (access->MainARGC) {
					delete[] access->MainARGS;
					access->MainARGS = NULL;
				}

				access->MainARGC = 0;
				access->entry_point = (QWORD)ACCESS_NO_ENTRY;

				PL_DBG("ERROR: ERROR not memory allocate to buffer!");
				ret = PL_ERROR;
				goto err_pl_exit1;
			}

#ifdef FILL_WITH_ZERO
			ElfFile->LoadProgramSegment(&Phdr[i], (char *)Buffer);
#else
			ElfFile->LoadProgramSegment(&Phdr[i], (char *)Buffer, 1);
#endif

#if 0 // ???
			ELF::ToM(
				ELF32_T_WORD,
				(unsigned char *)Buffer,
				(unsigned char *)Buffer,
				ELFDATA2LSB,
				load_size >> 2);
#endif

			ret = access->Xfer(Buffer, load_addr,
					   load_size, check_size, 2);

			delete[] Buffer;

			if (ret < 0)
				throw ELF_ERROR("Bad address ranges in program");
			else if (ret)
				throw ELF_ERROR("Failed write to memoty");
		}

		ELF_SCN *scn_rpc = 0, *scn_syms = 0, *scn_stack = 0;
		QWORD stack_addr = 0, stack_size = 0, stack_base = 0;
		PL_Addr dataRPC[2] = {0, 0};

		for (ELF_SCN *scn = ElfFile->NextScn(0); run_it && (scn != 0); scn = ElfFile->NextScn(scn)) {
			if ((scn_stack || scn_syms) && scn_rpc)
				break;

			Elf32_Shdr *shdr = scn->GetShdr();

			if ((shdr->sh_type == SHT_SYMTAB) && (strcmp(scn->GetName(), ".symtab") == 0)) {
				if (access->coreNo.nm_id == -1)
					continue;

				ELF_DATA *d = scn->GetData(0);
				Elf32_Sym *syms = (Elf32_Sym *)d->d_buf;
				Elf32_Word nums = d->d_size / ELF32_T_SYM_FSIZE;

				for (Elf32_Word i = 0; i < nums; i++) {
					if ((syms[i].st_shndx == SHN_UNDEF) || (syms[i].st_shndx >= SHN_LORESERVE))
						continue;

					unsigned char bind = ELF32_ST_BIND(syms[i].st_info);
					if ((bind != STB_GLOBAL) && (bind != STB_WEAK))
						continue;

					char *name = ElfFile->StrPtr(shdr->sh_link, syms[i].st_name);

					if (strcmp(name, "STACK") == 0) {
						stack_base = (QWORD)syms[i].st_value << 2;

						scn_syms = scn;
						break;
					}
				}

				continue;
			}

			if ((shdr->sh_type == SHT_NOBITS) && (strcmp(scn->GetName(), ".stack") == 0)) {
				if (access->coreNo.nm_id != -1)
					continue;

				stack_size = (QWORD)shdr->sh_size;
				stack_addr = (QWORD)shdr->sh_addr;

				scn_stack = scn;
				continue;
			}

			if ((shdr->sh_type == SHT_PROGBITS) && (strcmp(scn->GetName(), ".rpc_services" ) == 0)) {
				scn_rpc = scn;

				dataRPC[0] = shdr->sh_addr * 1;
				dataRPC[1] = shdr->sh_size / 4;

				continue;
			}
		}

		if (access->MainARGC && (scn_stack || scn_syms)) {
			QWORD ch_sizeof = 0, ch_offset = 2 * sizeof(DWORD);
			const char *FileName = ElfFile->GetName();
			int SparseChar = (access->coreNo.nm_id == -1) ? 1 : sizeof(DWORD);
			char *ch_buffer = NULL;
			QWORD args_addr = 0;

			if (!FileName)
				FileName = "unknown";

			if (scn_stack) {
				args_addr = stack_addr;
				ch_sizeof = stack_size / 8;
			}
			else {
				args_addr = stack_base;
				ch_sizeof = MAIN_ARGS_BUF;
			}

			if (ch_sizeof > ((QWORD)(access->MainARGC + 8) * sizeof(DWORD)))
				ch_buffer = new (std::nothrow) char[ch_sizeof];

			if (ch_buffer == NULL) {
				PL_DBG("ERROR: Unable allocate memory to buffer of args");
				throw 1;
			}

			memset(ch_buffer, '\0', ch_sizeof);

			ch_sizeof -= (QWORD)(access->MainARGC + 8) * sizeof(DWORD);
			if (((ch_sizeof >> 2) << 2) < ((QWORD)(strlen(FileName) + 1) * SparseChar)) {
				delete[] ch_buffer;

				PL_DBG("ERROR: Unable set file name to buffer of args");
				throw 1;
			}

			for (int i = 0; i <= strlen(FileName); i++)
				*(ch_buffer + ch_offset + i * SparseChar) = *(FileName + i);

			ch_offset += ((((QWORD)(strlen(FileName) + 1) * SparseChar) + 3) >> 2) << 2;
			ch_sizeof -= ((((QWORD)(strlen(FileName) + 1) * SparseChar) + 3) >> 2) << 2;

			for (int i = 0; i < access->MainARGC; i++) {
				char *argn = access->MainARGV[i];
				access->MainARGV[i] = (char *)((args_addr + ch_offset) / SparseChar);

				if (((ch_sizeof >> 2) << 2) < ((QWORD)(strlen(argn) + 1) * SparseChar)) {
					delete[] ch_buffer;

					PL_DBG("ERROR: Unable set string to buffer of args");
					throw 1;
				}

				for (int j = 0; j <= strlen(argn); j++)
					*(ch_buffer + ch_offset + j * SparseChar) = *(argn + j);

				ch_offset += ((((QWORD)(strlen(argn) + 1) * SparseChar) + 3) >> 2) << 2;
				ch_sizeof -= ((((QWORD)(strlen(argn) + 1) * SparseChar) + 3) >> 2) << 2;
			}

			ch_sizeof = ch_offset;
			*(DWORD *)(ch_buffer + ch_offset) = (DWORD)((args_addr + 2 * sizeof(DWORD)) / SparseChar);
			ch_offset += sizeof(DWORD);

			for (int i = 0; i < access->MainARGC; i++) {
				*(DWORD *)(ch_buffer + ch_offset) = *(DWORD *)&access->MainARGV[i];
				ch_offset += sizeof(DWORD);
			}

			*(DWORD *)(ch_buffer + ch_offset) = 0;
			ch_offset += sizeof(DWORD);

			if (ch_offset & 0x7) {
				*(DWORD *)(ch_buffer + ch_offset) = 0;
				ch_offset += sizeof(DWORD);
			}

			*(DWORD *)(ch_buffer + ch_offset) = (DWORD)((args_addr + ch_sizeof) / SparseChar);
			ch_offset += sizeof(DWORD);

			*(DWORD *)(ch_buffer + ch_offset) = (DWORD)(access->MainARGC + 1);
			ch_offset += sizeof(DWORD);

			*(DWORD *)(ch_buffer + 0 * sizeof(DWORD)) = (DWORD)((args_addr + ch_offset) / SparseChar);
			*(DWORD *)(ch_buffer + 1 * sizeof(DWORD)) = (DWORD)(MAIN_ARGS_MGC);

			ret = access->Xfer((DWORD *)ch_buffer, args_addr, ch_offset, ch_offset, 1);
			delete[] ch_buffer;

			if (ret < 0)
				throw ELF_ERROR("Bad address ranges in program");
			else if (ret)
				throw ELF_ERROR("Failed write to memoty");

			delete[] access->MainARGS;

			access->MainARGS = NULL;
			access->MainARGC = 0;
		}

		if (access->MainARGC) {
			if (!run_it) {
				delete[] access->MainARGS;

				access->MainARGS = NULL;
				access->MainARGC = 0;
			}
			else
				throw ELF_ERROR("Program has not section to set arguments");
		}

		if (scn_rpc) {
			ret = access->board->Command(&access->coreNo, SET_RPC_REGION, (DWORD *)dataRPC);
			if (ret < 0) {
				PL_DBG("ERROR: Unable set rpc region, ret = %d", ret);
				throw ELF_ERROR("Some program works with the access now");
			}
			else if (ret) {
				PL_DBG("ERROR: Failed set rpc region, ret = %d", ret);
				throw 2;
			}
		}
		else {
			ret = access->board->Command(&access->coreNo, CLR_RPC_REGION, (DWORD *)dataRPC);
			if (ret < 0) {
				PL_DBG("ERROR: Unable clear rpc region, ret = %d", ret);
				throw ELF_ERROR("Some program works with the access now");
			}
			else if (ret) {
				PL_DBG("ERROR: Failed clear rpc region, ret = %d", ret);
				throw 2;
			}
		}

		if (run_it) {
			ret = PL_StartCore(access, Start_Addr, status);
			if (ret) {
				PL_DBG("ERROR: Failed  start program, ret = %d!", ret);
				throw 3;
			}
		}

		*entry = Start_Addr;
	}
	catch (ELF_ERROR& ErrorCodeElf) {
		if (access->MainARGC) {
			delete[] access->MainARGS;
			access->MainARGS = NULL;
		}

		access->MainARGC = 0;
		access->entry_point = (QWORD)ACCESS_NO_ENTRY;

		const char *ErrorCodeString;
		std::ostringstream os;
		os << ErrorCodeElf;
		std::string s;
		s = os.str();
		ErrorCodeString = s.c_str();

		PL_DBG("ERROR: ELF ERROR: %s!", ErrorCodeString);

		if (!strcmp(ErrorCodeString, "Bad address ranges in program"))
			ret = PL_BADADDRESS;
		else if (!strcmp(ErrorCodeString, "Some program works with the access now"))
			ret = PL_AGAIN;
		else
			ret = PL_ERROR;

		goto err_pl_exit1;
	}
	catch (...) {
		if (access->MainARGC) {
			delete[] access->MainARGS;
			access->MainARGS = NULL;
		}

		access->MainARGC = 0;
		access->entry_point = (QWORD)ACCESS_NO_ENTRY;

		PL_DBG("ERROR: ERROR detected!");
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = PL_OK;

err_pl_exit1:

	libload_mutex_unlock(access->board->b_mutex_map,
			access->coreNo.cluster_id, access->coreNo.nm_id);

	LOCK_LL;
	access->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

	return ret;
}

struct ProgramBuffer : std::streambuf
{
	ProgramBuffer(char* begin, char* end)
	{
		this->setg(begin, begin, end);
	}

	pos_type seekoff(off_type off, std::ios_base::seekdir dir, std::ios_base::openmode which = std::ios_base::in)
	{
		if (dir == std::ios_base::cur)
			gbump(off);
		else if (dir == std::ios_base::end)
			setg(eback(), egptr() + off, egptr());
		else if (dir == std::ios_base::beg)
			setg(eback(), eback() + off, egptr());
		return gptr() - eback();
	}

	pos_type seekpos(pos_type sp, std::ios_base::openmode which)
	{
		return seekoff(sp - pos_type(off_type(0)), std::ios_base::beg, which);
	}
};

int PL_LoadProgram(PL_Access *access, const void *addrProgram, unsigned int sizeProgram)
{
	int ret;

	PL_DBG("Start.");

	char *StartProgram = (char *)addrProgram;
	ELF *ElfFile = NULL;
	PL_Addr entry;

	if (!addrProgram || !sizeProgram) {
		PL_DBG("ERROR: Invalid 'addrProgram' and/or 'sizeProgram' (NULL)!");
		ret = PL_ERROR;
		goto err_pl_exit;
	}
	else {
		ProgramBuffer sbuf(StartProgram, StartProgram + sizeProgram);
		std::iostream ios(&sbuf);

		try {
			ElfFile = new ELF(&ios, ELF_C_READ);
		}
		catch (ELF_ERROR& ErrorCodeElf) {
			const char *ErrorCodeString;
			std::ostringstream os;
			os << ErrorCodeElf;
			std::string s;
			s = os.str();
			ErrorCodeString = s.c_str();

			PL_DBG("ERROR: ELF ERROR: %s!", ErrorCodeString);
			ret = PL_FILE;
			goto err_pl_exit;
		}
		catch (...) {
			PL_DBG("ERROR: ERROR detected!");
			ret = PL_ERROR;
			goto err_pl_exit;
		}

		ret = PL_LoadELF(access, ElfFile, &entry, true);
		delete ElfFile;
	}

err_pl_exit:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else if (ret == PL_BADADDRESS)
		PL_DBG("End (With ERROR!); Return: PL_BADADDRESS!");
	else if (ret == PL_AGAIN)
		PL_DBG("End (With ERROR!); Return: PL_AGAIN!");
	else if (ret == PL_FILE)
		PL_DBG("End (With ERROR!); Return: PL_FILE!");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

static int _PL_LoadProgramFile(PL_Access *access, const char *filename, PL_Addr *startAddr, bool start)
{
	int ret;

	PL_DBG("Start.");

	const char *FileName = filename;
	ELF *ElfFile = NULL;

	if (!filename) {
		PL_DBG("ERROR: Invalid 'filename' (NULL)!");
		ret = PL_ERROR;
		goto err_pl_exit;
	}

	try {
		ElfFile = new ELF(FileName, ELF_C_READ);
	}
	catch (ELF_ERROR& ErrorCodeElf) {
		const char *ErrorCodeString;
		std::ostringstream os;
		os << ErrorCodeElf;
		std::string s;
		s = os.str();
		ErrorCodeString = s.c_str();

		PL_DBG("ERROR: ELF ERROR: %s!", ErrorCodeString);
		ret = PL_FILE;
		goto err_pl_exit;
	}
	catch (...) {
		PL_DBG("ERROR: ERROR detected!");
		ret = PL_ERROR;
		goto err_pl_exit;
	}

	ret = PL_LoadELF(access, ElfFile, startAddr, start);

	delete ElfFile;

err_pl_exit:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else if (ret == PL_BADADDRESS)
		PL_DBG("End (With ERROR!); Return: PL_BADADDRESS!");
	else if (ret == PL_AGAIN)
		PL_DBG("End (With ERROR!); Return: PL_AGAIN!");
	else if (ret == PL_FILE)
		PL_DBG("End (With ERROR!); Return: PL_FILE!");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_LoadProgramFile(PL_Access *access, const char *filename)
{
	int ret;
	PL_Addr entry;

	PL_DBG("Start.");
	ret = _PL_LoadProgramFile(access, filename, &entry, true);
	PL_DBG("End.");

	return ret;
}

int PL_LoadProgramFileWithoutRun(PL_Access *access, const char *filename, PL_Addr *startAddr)
{
	int ret;
	PL_Addr entry;

	PL_DBG("Start.");

	if (!startAddr) {
		PL_DBG("ERROR: Invalid 'startAddr' (NULL)!");
		return PL_ERROR;
	}

	ret = _PL_LoadProgramFile(access, filename, &entry, false);

	if (ret == PL_OK)
		*startAddr = entry;

	PL_DBG("End.");

	return ret;
}

static int PL_XferMemBlock(PL_Access *access, PL_Word *block, PL_Addr address, DWORD len, int RW)
{
	QWORD laddr, csize, lsize;
	int ret;

	PL_DBG("Start.");

	if (!block) {
		PL_DBG("ERROR: Invalid 'block' (NULL)!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}
	else if ((QWORD)block & 0x3) {
		PL_DBG("ERROR: Invalid 'block' (align 4)!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	LOCK_LL;
	ret = Check_access(access);
	if (!ret)
		access->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'access'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	libload_mutex_lock(access->board->b_mutex_map,
			   access->coreNo.cluster_id, access->coreNo.nm_id);

	DWORD status;

	ret = access->board->Command(&access->coreNo, GET_LOADER_STATUS, &status);
	if (ret) {
		PL_DBG("ERROR: Unable get loader status, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	DWORD saddr;

	ret = access->board->Command(&access->coreNo, GET_ENTRY_POINT, &saddr);
	if (ret) {
		PL_DBG("ERROR: Unable get entry point, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	access->entry_point = (QWORD)ACCESS_NO_ENTRY;

	if (access->coreNo.nm_id != -1) {
		laddr = (QWORD)address << 2;

		if (status == PROGRAM_STARTED)
			access->entry_point = (QWORD)saddr << 2;
	}
	else {
		laddr = (QWORD)address;

		if (status == PROGRAM_STARTED)
			access->entry_point = (QWORD)saddr;
	}

	if (len) {
		csize = (QWORD)len << 2;
		lsize = (QWORD)len << 2;
	}
	else {
		csize = 4;
		lsize = 0;
	}

	ret = access->Xfer((DWORD *)block, laddr, lsize, csize, RW);

	if (ret < 0) {
		PL_DBG("ERROR: Bad address ranges");
		ret = PL_BADADDRESS;
		goto err_pl_exit1;
	}
	else if (ret) {
		PL_DBG("ERROR: Failed xfer memory!");
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = PL_OK;

err_pl_exit1:

	libload_mutex_unlock(access->board->b_mutex_map,
			     access->coreNo.cluster_id, access->coreNo.nm_id);

	LOCK_LL;
	access->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else if (ret == PL_BADADDRESS)
		PL_DBG("End (With ERROR!); Return: PL_BADADDRESS!");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_ReadMemBlock(PL_Access *access, PL_Word *block, PL_Addr address, DWORD len)
{
	int ret;

	PL_DBG("Start.");
	ret = PL_XferMemBlock(access, block, address, len, 0);
	PL_DBG("End.");

	return ret;
}

int PL_WriteMemBlock(PL_Access *access, PL_Word *block, PL_Addr address, DWORD len)
{
	int ret;

	PL_DBG("Start.");
	ret = PL_XferMemBlock(access, block, address, len, 1);
	PL_DBG("End.");

	return ret;
}

static int PL_XferRegister(PL_Access *access, PL_Word *value, PL_Addr address, int RW)
{
	DWORD		tmp_buf[2];
	int		ret;

	PL_DBG("Start.");

	LOCK_LL;
	ret = Check_access(access);
	if (!ret)
		access->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'access'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	if (access->coreNo.nm_id != -1) {
		PL_DBG("ERROR: Invalid CoreNo!");
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	tmp_buf[0] = (DWORD)address;

	if (RW) {
		tmp_buf[1] = *((DWORD *)value);

		ret = access->board->Command(&access->coreNo, WRITE_ARM_REGISTER, tmp_buf);
	}
	else {
		ret = access->board->Command(&access->coreNo, READ_ARM_REGISTER, tmp_buf);

		*((DWORD *)value) = tmp_buf[1];
	}

	if (ret < 0) {
		PL_DBG("ERROR: Bad address ranges");
		ret = PL_BADADDRESS;
		goto err_pl_exit1;
	}
	else if (ret) {
		PL_DBG("ERROR: Unable get access to register, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = PL_OK;

err_pl_exit1:

	LOCK_LL;
	access->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else if (ret == PL_BADADDRESS)
		PL_DBG("End (With ERROR!); Return: PL_BADADDRESS!");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_ReadRegister(PL_Access *access, PL_Word *returnValue, PL_Addr address)
{
	PL_Word	value;
	int	ret;

	PL_DBG("Start.");

	if (!returnValue) {
		PL_DBG("ERROR: Invalid 'returnValue' (NULL)!");
		return PL_ERROR;
	}

	ret = PL_XferRegister(access, &value, address, 0);

	if (ret == PL_OK)
		*returnValue = value;

	PL_DBG("End.");

	return ret;
}

int PL_WriteRegister(PL_Access *access, PL_Word value, PL_Addr address)
{
	int ret;

	PL_DBG("Start.");
	ret = PL_XferRegister(access, &value, address, 1);
	PL_DBG("End.");

	return ret;
}

static int PL_XferMemAsRegister(PL_Access *access, PL_Word *block, PL_Addr address, DWORD length, int RW)
{
	QWORD addr, size;
	DWORD *tmp[3];
	int ret;

	PL_DBG("Start.");

	if (!block || ((QWORD)block & 0x3)) {
		PL_DBG("ERROR: Invalid 'block'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	LOCK_LL;
	ret = Check_access(access);
	if (!ret)
		access->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'access'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	addr = address;
	size = length;

	tmp[0] = (DWORD *)addr;
	tmp[1] = (DWORD *)size;
	tmp[2] = (DWORD *)block;

	if (RW)
		ret = access->board->Command(&access->coreNo, WRITE_MEM_AS_REG, (DWORD *)tmp);
	else
		ret = access->board->Command(&access->coreNo, READ_MEM_AS_REG, (DWORD *)tmp);

	if (ret < 0) {
		PL_DBG("ERROR: Bad address ranges");
		ret = PL_BADADDRESS;
		goto err_pl_exit1;
	}
	else if (ret) {
		PL_DBG("ERROR: Unable get access to memory, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = PL_OK;

err_pl_exit1:

	LOCK_LL;
	access->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else if (ret == PL_BADADDRESS)
		PL_DBG("End (With ERROR!); Return: PL_BADADDRESS!");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;

}

int PL_WriteMemAsRegister(PL_Access *, PL_Word *, PL_Addr, DWORD) __attribute__ ((visibility ("default")));
int PL_WriteMemAsRegister(PL_Access *access, PL_Word *block, PL_Addr address, DWORD len)
{
	int ret;

	PL_DBG("Start.");
	ret = PL_XferMemAsRegister(access, block, address, len, 1);
	PL_DBG("End.");

	return ret;
}

int PL_ReadMemAsRegister(PL_Access *, PL_Word *, PL_Addr, DWORD) __attribute__ ((visibility ("default")));
int PL_ReadMemAsRegister(PL_Access *access, PL_Word *block, PL_Addr address, DWORD len)
{
	int ret;

	PL_DBG("Start.");
	ret = PL_XferMemAsRegister(access, block, address, len, 0);
	PL_DBG("End.");

	return ret;
}

int PL_Sync(PL_Access *access, PL_Word value, PL_Word *returnValue)
{
	int ret;

	PL_DBG("Start.");
	ret = PL_SyncArray(access, value, 0, 0, returnValue, NULL, NULL);
	PL_DBG("End.");

	return ret;
}

int PL_SyncArray(PL_Access *access,
		PL_Word value, PL_Addr outAddress, DWORD outLen,
		PL_Word *returnValue, PL_Addr *inAddress, DWORD *inLen)
{
	PL_Word		sync_to_pc[4];
	PL_Word		sync_from_pc[4];
	int		ret = 0;

	PL_DBG("Start.");

	LOCK_LL;
	ret = Check_access(access);
	if (!ret)
		access->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'access'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	libload_mutex_lock(access->board->b_mutex_map,
			   access->coreNo.cluster_id, access->coreNo.nm_id);
#if 0
	PL_Word	status;

	ret = access->board->Command(&access->coreNo, GET_LOADER_STATUS, (DWORD *)&status);

	if (ret) {
		PL_DBG("ERROR: Unable get loader status, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	if (status != PROGRAM_STARTED) {
		PL_DBG("ERROR: User program not be in progress"
			" (Status: 0x%8.8X)!", status);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}
#endif
	struct timeval	tv;
	unsigned int	start_time;
	unsigned int	current_time;

	gettimeofday(&tv, NULL);
	start_time = (tv.tv_sec * 1000) + (tv.tv_usec / 1000);

	PL_Word	is_busy;

	while (1) {
		ret = access->board->Command(&access->coreNo, GET_PROCESSOR_BUSY, (DWORD *)&is_busy);
		if (ret) {
			PL_DBG("ERROR: Unable get processor busy condition, ret = %d!", ret);
			ret = PL_ERROR;
			goto err_pl_exit1;
		}

		if (!is_busy)
			break;

		LOCK_LL;
		if (globalTimeout) {
			gettimeofday(&tv, NULL);
			current_time = (tv.tv_sec * 1000) + (tv.tv_usec / 1000);

			if ((current_time - start_time) > globalTimeout) {
				UNLOCK_LL;

				PL_DBG("ERROR: Timeout get processor to sync!");
				ret = PL_TIMEOUT;
				goto err_pl_exit1;
			}
		}
		UNLOCK_LL;

		libload_mutex_unlock(access->board->b_mutex_map,
				access->coreNo.cluster_id, access->coreNo.nm_id);

		usleep(1000);

		libload_mutex_lock(access->board->b_mutex_map,
				access->coreNo.cluster_id, access->coreNo.nm_id);
	}

	is_busy = 1;

	ret = access->board->Command(&access->coreNo, SET_PROCESSOR_BUSY, (DWORD *)&is_busy);
	if (ret) {
		PL_DBG("ERROR: Unable set processor busy condition, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	while (1) {
		ret = access->board->Command(&access->coreNo, GET_SYNC_FLAG_TO_PC, (DWORD *)&sync_to_pc[SYNC_FLAG]);
		if (ret) {
			PL_DBG("ERROR: Unable get sync flag to PC, ret = %d!", ret);
			ret = PL_ERROR;
			goto err_pl_exit1;
		}

		if (sync_to_pc[SYNC_FLAG])
			break;

		LOCK_LL;
		if (globalTimeout) {
			gettimeofday(&tv, NULL);
			current_time = (tv.tv_sec * 1000) + (tv.tv_usec / 1000);

			if ((current_time - start_time) > globalTimeout) {
				UNLOCK_LL;

				is_busy = 0;

				ret = access->board->Command(&access->coreNo, SET_PROCESSOR_BUSY, (DWORD *)&is_busy);
				if (ret) {
					PL_DBG("ERROR: Unable clear processor busy condition, ret = %d!", ret);
					ret = PL_ERROR;
				}
				else {
					PL_DBG("ERROR: Timeout set sync flag to PC!");
					ret = PL_TIMEOUT;
				}

				goto err_pl_exit1;
			}
		}
		UNLOCK_LL;

		libload_mutex_unlock(access->board->b_mutex_map,
				access->coreNo.cluster_id, access->coreNo.nm_id);

		usleep(1000);

		libload_mutex_lock(access->board->b_mutex_map,
				access->coreNo.cluster_id, access->coreNo.nm_id);
	}

	ret = access->board->Command(&access->coreNo, GET_SYNC_ARRAY_TO_PC, (DWORD *)&sync_to_pc[SYNC_VAL]);
	if (ret) {
		PL_DBG("ERROR: Unable get sync array to PC, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	if (returnValue)
		*returnValue = sync_to_pc[SYNC_VAL];

	if (inAddress)
		*inAddress = sync_to_pc[SYNC_ADDR];

	if (inLen)
		*inLen = sync_to_pc[SYNC_LEN];

	sync_from_pc[SYNC_VAL] = value;
	sync_from_pc[SYNC_ADDR] = outAddress;
	sync_from_pc[SYNC_LEN] = outLen;

	ret = access->board->Command(&access->coreNo, SET_SYNC_ARRAY_FROM_PC, (DWORD *)&sync_from_pc[SYNC_VAL]);
	if (ret) {
		PL_DBG("ERROR: Unable set sync array from PC, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	sync_from_pc[SYNC_FLAG] = 1;

	ret = access->board->Command(&access->coreNo, SET_SYNC_FLAG_FROM_PC, (DWORD *)&sync_from_pc[SYNC_FLAG]);
	if (ret) {
		PL_DBG("ERROR: Unable set sync flag from PC, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	while (1) {
		ret = access->board->Command(&access->coreNo, GET_SYNC_FLAG_FROM_PC, (DWORD *)&sync_from_pc[SYNC_FLAG]);
		if (ret) {
			PL_DBG("ERROR: Unable get sync flag from PC, ret = %d!", ret);
			ret = PL_ERROR;
			goto err_pl_exit1;
		}

		if (!sync_from_pc[SYNC_FLAG])
			break;
#if 1
		LOCK_LL;
		if (globalTimeout) {
			gettimeofday(&tv, NULL);
			current_time = (tv.tv_sec * 1000) + (tv.tv_usec / 1000);

			if ((current_time - start_time) > globalTimeout) {
				UNLOCK_LL;

				PL_DBG("ERROR: Timeout clear sync flag from PC!");
				ret = PL_TIMEOUT;
				goto err_pl_exit1;
			}
		}
		UNLOCK_LL;
#endif
		libload_mutex_unlock(access->board->b_mutex_map,
				access->coreNo.cluster_id, access->coreNo.nm_id);

		usleep(1000);

		libload_mutex_lock(access->board->b_mutex_map,
				access->coreNo.cluster_id, access->coreNo.nm_id);
	}

	sync_to_pc[SYNC_FLAG] = 0;

	ret = access->board->Command(&access->coreNo, SET_SYNC_FLAG_TO_PC, (DWORD *)&sync_to_pc[SYNC_FLAG]);
	if (ret) {
		PL_DBG("ERROR: Unable clear sync flags to PC, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	is_busy = 0;

	ret = access->board->Command(&access->coreNo, SET_PROCESSOR_BUSY, (DWORD *)&is_busy);
	if (ret) {
		PL_DBG("ERROR: Unable clear processor busy condition, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = PL_OK;

err_pl_exit1:

	libload_mutex_unlock(access->board->b_mutex_map,
			access->coreNo.cluster_id, access->coreNo.nm_id);

	LOCK_LL;
	access->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else if (ret == PL_TIMEOUT)
		PL_DBG("End (With ERROR!); Return: PL_TIMEOUT.");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_SetTimeout(DWORD timeout)
{
	PL_DBG("Start.");

	LOCK_LL;
	globalTimeout = timeout;
	UNLOCK_LL;

	PL_DBG("End; Return: PL_OK.");
	return PL_OK;
}

int PL_GetRegionRPC(PL_Access *access, PL_Addr *addrRPC, PL_Addr *sizeRPC)
{
	PL_Addr		dataRPC[2];
	int		ret;

	PL_DBG("Start.");

	if (!addrRPC || !sizeRPC) {
		PL_DBG("ERROR: Invalid 'addrRPC' and/or 'sizeRPC' (NULL)!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	LOCK_LL;
	ret = Check_access(access);
	if (!ret)
		access->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'access'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	libload_mutex_lock(access->board->b_mutex_map,
			access->coreNo.cluster_id, access->coreNo.nm_id);

	DWORD status;

	ret = access->board->Command(&access->coreNo, GET_LOADER_STATUS, &status);
	if (ret) {
		PL_DBG("ERROR: Unable get loader status, ret = %d", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	if (status == PROGRAM_NO_LOAD) {
		PL_DBG("ERROR: Program is not loaded yet"
			" (status = 0x%8.8X)!", status);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = access->board->Command(&access->coreNo, GET_RPC_REGION, (DWORD *)dataRPC);
	if (ret < 0) {
		PL_DBG("ERROR: RPC region is busy, ret = %d!", ret);
		ret = PL_AGAIN;
		goto err_pl_exit1;
	}
	else if (ret) {
		PL_DBG("ERROR: Unable get RPC region, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	*addrRPC = dataRPC[0];
	*sizeRPC = dataRPC[1];

	ret = PL_OK;

err_pl_exit1:

	libload_mutex_unlock(access->board->b_mutex_map,
			access->coreNo.cluster_id, access->coreNo.nm_id);

	LOCK_LL;
	access->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else if (ret == PL_AGAIN)
		PL_DBG("End (With ERROR!); Return: PL_AGAIN!");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_GetStatus(PL_Access *access, PL_Word *status)
{
	int		ret;

	PL_DBG("Start.");

	if (!status) {
		PL_DBG("ERROR: Invalid 'status' (NULL)!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	LOCK_LL;
	ret = Check_access(access);
	if (!ret)
		access->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'access'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	ret = access->board->Command(&access->coreNo, GET_LOADER_STATUS, (DWORD *)status);
	if (ret) {
		PL_DBG("ERROR: Unable get loader status, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = PL_OK;

err_pl_exit1:

	LOCK_LL;
	access->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

int PL_GetResult(PL_Access *access, PL_Word *result)
{
	int		ret;

	PL_DBG("Start.");

	if (!result) {
		PL_DBG("ERROR: Invalid 'result' (NULL)!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	LOCK_LL;
	ret = Check_access(access);
	if (!ret)
		access->Count_In_Progress++;
	UNLOCK_LL;

	if (ret) {
		PL_DBG("ERROR: Invalid 'access'!");
		ret = PL_ERROR;
		goto err_pl_exit2;
	}

	PL_Word status;

	ret = access->board->Command(&access->coreNo, GET_LOADER_STATUS, (DWORD *)&status);
	if (ret) {
		PL_DBG("ERROR: Unable get loader status, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	if ((status != PROGRAM_STOPPED) && (status != PROGRAM_ABORTED)) {
		PL_DBG("ERROR: User program not complete"
			" (Status: 0x%8.8X)!", status);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = access->board->Command(&access->coreNo, GET_PROGRAM_RESULT, (DWORD *)result);
	if (ret) {
		PL_DBG("ERROR: Unable get program result, ret = %d!", ret);
		ret = PL_ERROR;
		goto err_pl_exit1;
	}

	ret = PL_OK;

err_pl_exit1:

	LOCK_LL;
	access->Count_In_Progress--;
	UNLOCK_LL;

err_pl_exit2:

#ifdef LIBLOAD_DEBUG
	if (ret == PL_OK)
		PL_DBG("End; Return: PL_OK.");
	else
		PL_DBG("End (With ERROR!); Return: PL_ERROR!");
#endif

	return ret;
}

struct ll_operation PL_FunctionExport __attribute__((visibility ("default"))) = {
	PL_GetVersion,
	PL_GetBoardCount,
	PL_SetChannelEDCL,
	PL_GetBoardDesc,
	PL_CloseBoardDesc,
	PL_GetSerialNumber,
	PL_GetFirmwareVersion,
	PL_ResetBoard,
	PL_LoadInitCode,
	PL_GetAccess,
	PL_TypeAccess,
	PL_CloseAccess,
	PL_SetProgramArgs,
	PL_LoadProgram,
	PL_LoadProgramFile,
	PL_ReadMemBlock,
	PL_WriteMemBlock,
	PL_ReadRegister,
	PL_WriteRegister,
	PL_Sync,
	PL_SyncArray,
	PL_SetTimeout,
	PL_GetRegionRPC,
	PL_GetStatus,
	PL_GetResult
};

#ifdef __cplusplus
} // extern "C"
#endif
