#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <time.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <linux/if_ether.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <linux/udp.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netdb.h>
#include <ifaddrs.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdbool.h>
#include <stdint.h>

#include "libload_edcl.h"

#ifdef __cplusplus
extern "C"
{
#endif

struct Hdr {
	struct ethhdr eth;
	struct ip ip;
	struct udphdr udp;
} __attribute__((packed));

#define __swap32(value)                                 \
	((((uint32_t)((value) & 0x000000FF)) << 24) |   \
	 (((uint32_t)((value) & 0x0000FF00)) << 8) |    \
	 (((uint32_t)((value) & 0x00FF0000)) >> 8) |    \
	 (((uint32_t)((value) & 0xFF000000)) >> 24))

static size_t edcl_get_maxpacket()
{
	return ETH_FRAME_LEN - sizeof(struct Hdr) - sizeof(struct EdclPacket);
}

static bool edcl_get_host_endianness() {
	union {
		uint32_t	i;
		char		c[4];
	} e = { 0x01000000 };

	if (e.c[0]) {
		return ENDIAN_BIG;
	} else {
		return ENDIAN_LITTLE;
	}
}

static void edcl_set_swap_need_flag(libload_edcl_t* chip_conf) {
	if (chip_conf->endian == edcl_get_host_endianness()) {
		chip_conf->need_swap = false;
	} else {
		chip_conf->need_swap = true;
	}
}

static unsigned int edcl_seq(struct EdclPacket const* p) {
	return (ntohl(p->control) >> 18) & 0x3FFF;
}

static unsigned int edcl_rwnak(struct EdclPacket const* p) {
	return (ntohl(p->control) >> 17) & 1;
}

static unsigned int edcl_len(struct EdclPacket const* p) {
	return (ntohl(p->control) >> 7) & 0x3FF;
}

static unsigned int edcl_rwflag(struct EdclPacket const* p) {
	return (ntohl(p->control) >> 17) & 1;
}

static unsigned int edcl_control(unsigned seq, unsigned flag, unsigned len) {
	return htonl(
		(seq & 0x3FFF) << 18 |
		((!!flag) & 1) << 17 |
		(len & 0x3FF) << 7);
}

static int edcl_init(char* ifname, libload_edcl_t* chip_conf)
{
	struct sockaddr_ll address = { 0 };
	struct ifreq iface;
	int s_fd;

	s_fd = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_IP));
	if(s_fd < 0)
		return -1;

	strncpy(iface.ifr_name, ifname, sizeof(iface.ifr_name));

	if(ioctl(s_fd,SIOCGIFINDEX,&iface)) {
		close(s_fd);
		return -1;
	}

	address.sll_family = AF_PACKET;
	address.sll_protocol = htons(ETH_P_IP);
	address.sll_ifindex = iface.ifr_ifindex;
	address.sll_hatype = 1;
	address.sll_pkttype = PACKET_HOST;

	if(bind(s_fd, (struct sockaddr*)&address, sizeof(address))) {
		close(s_fd);
		return -1;
	}

	if(ioctl(s_fd, SIOCGIFHWADDR, &iface)) {
		close(s_fd);
		return -1;
	}

	memcpy(
		chip_conf->local_mac,
		&iface.ifr_hwaddr.sa_data,
		sizeof(chip_conf->local_mac));

	return s_fd;
}

static ssize_t edcl_send(int s_fd, libload_edcl_t* chip_conf, void* data, size_t len)
{
	struct Hdr* hdr;
	unsigned int dst_ip;
	time_t t1;
	ssize_t n;

	if (s_fd < 0) {
		errno = EINVAL;
		return -1;
	}

	hdr = (struct Hdr*)malloc(ETH_FRAME_LEN + ETH_FCS_LEN);
	if (!hdr) {
		errno = ENOMEM;
		return -1;
	}

	t1 = time(0);

	do { // Flush buffer!!!
		if((time(0) - t1) > 10) {
			errno = ETIMEDOUT;
			free(hdr);

			return -1;
		}

		n = recv(s_fd, hdr, ETH_FRAME_LEN, MSG_DONTWAIT);
	} while (n > 0);

	memcpy(
		hdr->eth.h_dest,
		chip_conf->remote_mac,
		sizeof(chip_conf->remote_mac)
	);

	memcpy(
		hdr->eth.h_source,
		chip_conf->local_mac,
		sizeof(chip_conf->local_mac)
	);

	hdr->eth.h_proto = htons(ETH_P_IP);

	hdr->ip.ip_v = 4;
	hdr->ip.ip_hl = 5;
	hdr->ip.ip_p = 0x11;
	hdr->ip.ip_ttl = 1;

	dst_ip = inet_addr("192.168.1.0");
	dst_ip = (dst_ip & ~htonl(0xFF)) | htonl(chip_conf->remote_mac[5]);

	hdr->ip.ip_src.s_addr = inet_addr("192.168.0.1");
	hdr->ip.ip_dst.s_addr = dst_ip;

	hdr->udp.source = htons(chip_conf->local_port);
	hdr->udp.dest = htons(chip_conf->remote_port);

	memcpy(hdr + 1, data, len);

	n = send(s_fd, hdr, sizeof(*hdr) + len, 0);
	if (n < ((ssize_t)sizeof(*hdr))) {
		free(hdr);
		return -1;
	}

	free(hdr);
	return n - sizeof(*hdr);
}

static ssize_t edcl_recv(int s_fd, libload_edcl_t* chip_conf, void* data, size_t len)
{
	struct Hdr* hdr;
	unsigned int src_ip;
	time_t t1;

	if (s_fd < 0) {
		errno = EINVAL;
		return -1;
	}

	hdr = (struct Hdr*)malloc(ETH_FRAME_LEN + ETH_FCS_LEN);
	if (!hdr) {
		errno = ENOMEM;
		return -1;
	}

	t1 = time(0);

	for(;;) {
		fd_set rdset;
		struct timeval timeout;
		ssize_t n;

		timeout.tv_sec = 0;
		timeout.tv_usec = 1000000;

		FD_ZERO(&rdset);
		FD_SET(s_fd, &rdset);

		if (select(s_fd + 1, &rdset, 0, 0, &timeout) <= 0) {
			errno = ETIMEDOUT;

			free(hdr);
			return -1;
		}

		n = recv(s_fd, hdr, ETH_FRAME_LEN, 0);
		if (n < ((ssize_t)sizeof(*hdr))) {
			free(hdr);
			return -1;
		}

		src_ip = inet_addr("192.168.1.0");
		src_ip = (src_ip & ~htonl(0xFF)) | htonl(chip_conf->remote_mac[5]);

		if ((memcmp(chip_conf->remote_mac, hdr->eth.h_source, sizeof(hdr->eth.h_source)) == 0) &&
		   (hdr->ip.ip_src.s_addr == src_ip) && (hdr->ip.ip_dst.s_addr == inet_addr("192.168.0.1")) &&
		   (hdr->ip.ip_p == 0x11) && (ntohs(hdr->udp.dest) == chip_conf->local_port)) {
			memcpy(data, hdr + 1, min(n - sizeof(*hdr), len));

			free(hdr);
			return n - sizeof(*hdr);
		}

		if((time(0) - t1) > 10) {
			errno = ETIMEDOUT;

			free(hdr);
			return -1;
		}
	}
}

static int error_count = 0;

static ssize_t edcl_transaction(int s_fd, libload_edcl_t* chip_conf, unsigned int *seqnum,
				struct EdclPacket* rq, size_t rqlen, struct EdclPacket* rs, size_t rslen)
{
	ssize_t n;
	int i, tmout = 0;

	for (i = 0; i < 10; i++) {
		if (edcl_send(s_fd, chip_conf, rq, rqlen) <= 0)
			return -1;

		n = edcl_recv(s_fd, chip_conf, rs, rslen);
		if (n <= 0) {
			if (n == 0)
				return 0;

			if (errno == ETIMEDOUT) { // Dropped request or acknowledge
				tmout = 1;
				continue;
			}

			break;
		}

		/* Workaround: may has shadow effect while work with registers  */
		if(edcl_rwnak(rs)) { // Response did be lost.
			*seqnum = edcl_seq(rs);

			if (!tmout)
				break;

			error_count++;

			rq->control &= ~edcl_control(0x3FFF, 0, 0);
			rq->control |= edcl_control(*seqnum, 0, 0);

			if (!edcl_rwflag(rq)) {
				printf("WARN: message lost count = %d, addr = 0x%x, len = 0x%x\n",
					error_count, ntohl(rq->address), edcl_len(rq));

				tmout = 0;
				continue;
			}

			return n;
		}

		*seqnum = (edcl_seq(rq) + 1) & 0x3FFF;

		if(edcl_seq(rs) != edcl_seq(rq)) {
			errno = EFAULT;
			return -1;
		}

		return n;
	}

	errno = ECOMM;
	return 0;
}

int edcl_get_seqnum(int s_fd, libload_edcl_t* chip_conf, unsigned int *seqnum)
{
	struct EdclPacket rs, rq;
	ssize_t ret;
	int cnt = 10;

	rq.control = edcl_control(~(*seqnum), 0, 0);
	rq.address = htonl(rq.address);
rep:
	ret = edcl_send(s_fd, chip_conf, &rq, sizeof(rq));
	if (ret <= 0)
		return -1;

	ret = edcl_recv(s_fd, chip_conf, &rs, sizeof(rs));
	if (ret <= 0) {
		if (ret == 0)
			return -1;
		else
			cnt--;

		if (cnt && (errno == ETIMEDOUT))
			goto rep;

		return -1;
	}

	if ((rs.address != rq.address) || (edcl_len(&rs) != edcl_len(&rq))) {
		errno = EPROTO;
		return -1;
	}

	if (*seqnum != edcl_seq(&rs)) {
		errno = EPROTO;
		return -1;
	}

	if(!edcl_rwnak(&rs))
		*seqnum = (*seqnum + 1) & 0x3FFF;

	return 0;
}

int edcl_test_init(char* ifname, libload_edcl_t* chip_conf, unsigned int *seqnum)
{
	struct EdclPacket rs, rq;
	ssize_t ret;
	int s_fd, cnt = 10;
	
	s_fd = edcl_init(ifname, chip_conf);
	if (s_fd < 0)
		return -1;

	edcl_set_swap_need_flag(chip_conf);

	rq.control = edcl_control(~(*seqnum), 0, 0);
	rq.address = htonl(rq.address);
rep:
	ret = edcl_send(s_fd, chip_conf, &rq, sizeof(rq));
	if (ret <= 0) {
		close(s_fd);
		return -1;
	}

	ret = edcl_recv(s_fd, chip_conf, &rs, sizeof(rs));
	if (ret <= 0) {
		if (ret == 0) {
			close(s_fd);
			return -1;
		}

		cnt--;

		if (cnt && (errno == ETIMEDOUT))
			goto rep;

		close(s_fd);
		return -1;
	}

	if ((rs.address == rq.address) && (edcl_len(&rs) == edcl_len(&rq))) {
		if(edcl_rwnak(&rs))
			*seqnum = edcl_seq(&rs);
		else
			*seqnum = (edcl_seq(&rs) + 1) & 0x3FFF;

		return s_fd;
	}

	errno = EPROTO;
	close(s_fd);

	return -1;
}

int edcl_read(int s_fd, libload_edcl_t* chip_conf, unsigned int *seqnum,
	      unsigned int address, void* obuf, size_t len)
{
	struct EdclPacket *rs, rq;
	ssize_t n;

	if((len > edcl_get_maxpacket()) || (len > chip_conf->maxpayload)) {
		errno = EINVAL;
		return -1;
	}

	rq.control = edcl_control(*seqnum, 0, len);
	rq.address = htonl(address);

	rs = (struct EdclPacket*)malloc(sizeof(*rs) + len);
	if (!rs) {
		errno = ENOMEM;
		return -1;
	}

	n = edcl_transaction(s_fd, chip_conf, seqnum, &rq, sizeof(rq), rs, sizeof(*rs) + len);
	if(n < ((ssize_t)(len + sizeof(*rs)))) {
		free(rs);
		return -1;
	}

	if((rs->address != rq.address) || (edcl_len(rs) != len)) {
		errno = EPROTO;

		free(rs);
		return -1;
	}

	if(chip_conf->need_swap)  {
		uint32_t *srcptr = (uint32_t *)(rs + 1);
		uint32_t *destptr = (uint32_t *)obuf;
  		int num = len / sizeof(uint32_t);
		int i;

		if(len % sizeof(uint32_t) != 0) {
			errno = EFAULT;

			free(rs);
			return -1;
		}	

		for(i = 0; i < num; ++i) {
			*destptr++ = __swap32(*srcptr);
			srcptr++;
		}
	}
	else {
		memcpy(obuf, rs + 1, len);
	}

	free(rs);
	return 0;
}

int edcl_write(int s_fd, libload_edcl_t* chip_conf, unsigned int *seqnum,
	       unsigned int address, void* ibuf, size_t len)
{
	struct EdclPacket *rq, rs;
	ssize_t n;

	if((len > edcl_get_maxpacket()) || (len > chip_conf->maxpayload)) {
		errno = EINVAL;
		return -1;
	}

	rq = (struct EdclPacket*)malloc(sizeof(*rq) + len);
	if (!rq) {
		errno = ENOMEM;
		return -1;
	}

	rq->control = edcl_control(*seqnum, 1, len);
	rq->address = htonl(address);

	if(chip_conf->need_swap) {
		uint32_t *srcptr = (uint32_t *)ibuf;
		uint32_t *destptr = (uint32_t *)(rq + 1);
  		int num = len / sizeof(uint32_t);
		int i;

		if(len % sizeof(uint32_t) != 0) {
			errno = EFAULT;

			free(rq);
			return -1;
  		}

  		for(i = 0; i < num; ++i) {
  			*destptr++ = __swap32(*srcptr);
			srcptr++;
		}
	}
	else {
		memcpy(rq + 1, ibuf, len);
	}

	n = edcl_transaction(s_fd, chip_conf, seqnum, rq, sizeof(*rq) + len, &rs, sizeof(rs));
	if(n < ((ssize_t)sizeof(rs))) {
		free(rq);
		return -1;
	}

	if((rs.address != rq->address) || (edcl_len(&rs) != 0)) {
		errno = EPROTO;

		free(rq);
		return -1;
	}

	free(rq);
	return 0;
}

#ifdef __cplusplus
}
#endif


