#include <sys/types.h>
#include <sys/syscall.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <semaphore.h>
#include <dirent.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libload_mutex.h"

#ifdef __cplusplus
extern "C"
{
#endif

#include "nm6408load.h"
#include "nm6408_map.h"
#include "nm6408_ioctl.h"

libload_mutex_t	*ll_mutex_map[NM6408_BRD_NUM];
int		ll_mutex_index[NM6408_BRD_NUM];
int		ll_mutex_count;

void __attribute__((constructor)) libload_mutex_constructor(void);
void __attribute__((destructor)) libload_mutex_destructor(void);

#undef LIBLOAD_MUTEX_DEBUG

#ifdef LIBLOAD_MUTEX_DEBUG
#define mutex_print_dbg(format, ...)	printf("%s#%d: " format ".\n", \
					__FUNCTION__, __LINE__, ##__VA_ARGS__)
#else
#define mutex_print_dbg(format, ...)
#endif

static int shm_mutex_open(char *name, libload_mutex_t **mutex_map)
{
	int		fd;
	libload_mutex_t	*_mutex_map;
	int		err = 0;

	mutex_print_dbg("begin - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	fd = shm_open(name, O_RDWR, 0666);
	if (fd < 0) {
		err = errno;
		mutex_print_dbg("shm_open - error = %d, pid = %u,"
				" tid = 0x%lx", err, getpid(),
				syscall(SYS_gettid));

		return err;
	}

	_mutex_map = (libload_mutex_t *)mmap(NULL, sizeof(*_mutex_map),
					     PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);

	if (_mutex_map == MAP_FAILED) {
		err = errno;
		mutex_print_dbg("mmap -  error = %d, pid = %u,"
				" tid = 0x%lx", err, getpid(),
				syscall(SYS_gettid));

		close(fd);
		return err;
	}

	close(fd);

	if (!_mutex_map->is_valid) {
		err = EAGAIN;
		mutex_print_dbg("mutex - error = %d, pid = %u,"
				" tid = 0x%lx", err, getpid(),
				syscall(SYS_gettid));

		munmap(_mutex_map, sizeof(*_mutex_map));
		return err;
	}

	*mutex_map = _mutex_map;

	mutex_print_dbg("end - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	return err;
}

static int shm_mutex_create(char *name, libload_mutex_t **mutex_map)
{
	int			fd;
	libload_mutex_t		*_mutex_map;
	pthread_mutexattr_t	mutexattr;
	mode_t			save_mask;
	int			i, cluster, core;
	int			rv, err = 0;

	mutex_print_dbg("begin - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	save_mask = umask(0);
	fd = shm_open(name, O_RDWR|O_CREAT|O_EXCL, 0666);
	umask(save_mask);

	if (fd < 0) {
		err = errno;
		mutex_print_dbg("shm_open - warn = %d, pid = %u,"
				" tid = 0x%lx", err, getpid(),
				syscall(SYS_gettid));

		return err;
	}

	ftruncate(fd, sizeof(*_mutex_map));

	_mutex_map = (libload_mutex_t *)mmap(NULL, sizeof(*_mutex_map),
					     PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);

	if (_mutex_map == MAP_FAILED) {
		err = errno;
		mutex_print_dbg("mmap - error = %d, pid = %u,"
				" tid = 0x%lx", err, getpid(),
				syscall(SYS_gettid));

		close(fd);
		shm_unlink(name);
		return err;
	}

	close(fd);

	pthread_mutexattr_init(&mutexattr);

	pthread_mutexattr_setpshared(&mutexattr,PTHREAD_PROCESS_SHARED);
	pthread_mutexattr_setrobust(&mutexattr,PTHREAD_MUTEX_ROBUST);
	pthread_mutexattr_settype(&mutexattr,PTHREAD_MUTEX_ERRORCHECK);

	for (cluster = 0; cluster < LIBLOAD_CLUSTER_NUMBER; cluster++) {
		for (core = 0; core < LIBLOAD_CORE_NUMBER; core++) {
			rv = pthread_mutex_init(&_mutex_map->n_lock[cluster][core],
						&mutexattr);

			if (rv)
				goto err_mutex_init;
		}
	}

	for (cluster = 0; cluster < LIBLOAD_CLUSTER_NUMBER; cluster++) {
		rv = pthread_mutex_init(&_mutex_map->a_lock[cluster], &mutexattr);

		if (rv)
			goto err_mutex_init;
	}

	rv = pthread_mutex_init(&_mutex_map->c_lock, &mutexattr);

	if (!rv)
		rv = pthread_mutex_init(&_mutex_map->b_lock, &mutexattr);

err_mutex_init:

	if (rv) {
		err = rv;
		mutex_print_dbg("mutex_init - error = %d, pid = %u,"
				" tid = 0x%lx", err, getpid(),
				syscall(SYS_gettid));

		shm_unlink(name);
		munmap(_mutex_map, sizeof(*_mutex_map));
		return err;
	}

	for (i = 0; i < LIBLOAD_PROCESSOR_NUMBER; i++) {
		_mutex_map->rpc_pid[i] = getpid();
		_mutex_map->rpc_tid[i] = syscall(SYS_gettid);
		_mutex_map->rpc_rdy[i] = 0;
	}

	_mutex_map->reset_count = 0;
	_mutex_map->seqnc_count = 0;

	_mutex_map->is_valid = 1;
	*mutex_map = _mutex_map;

	mutex_print_dbg("end - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	return err;
}

int libload_mutex_open(char *name, libload_mutex_t **mutex_map)
{
	libload_mutex_t	*_mutex_map = (libload_mutex_t *)MAP_FAILED;
	int		rv;

	mutex_print_dbg("begin - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	rv = shm_mutex_create(name, &_mutex_map);
	if (rv == EEXIST) {
		do {
			usleep(100);
			rv = shm_mutex_open(name, &_mutex_map);
		} while (rv == EAGAIN);
	}

	*mutex_map = _mutex_map;

	mutex_print_dbg("end - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	return rv;
}

int libload_mutex_close(libload_mutex_t **mutex_map)
{
	libload_mutex_t	*_mutex_map = *mutex_map;
	int		rv, err = 0;

	mutex_print_dbg("begin - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	if ((_mutex_map == MAP_FAILED) || (_mutex_map == NULL)) {
		err = EINVAL;
		mutex_print_dbg("mutex - error = %d, pid = %u,"
				" tid = 0x%lx", err, getpid(),
				syscall(SYS_gettid));

		return err;
	}

	rv = munmap(_mutex_map, sizeof(*_mutex_map));
	if (rv) {
		err = errno;
		mutex_print_dbg("munmap - error = %d, pid = %u,"
				" tid = 0x%lx", err, getpid(),
				syscall(SYS_gettid));

		return err;
	}

	*mutex_map = (libload_mutex_t *)MAP_FAILED;

	mutex_print_dbg("end - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	return err;
}

int libload_mutex_lock(libload_mutex_t *mutex_map, int cluster, int core)
{
	int	rv, err = 0;

	mutex_print_dbg("begin - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	if ((mutex_map == MAP_FAILED) || (mutex_map == NULL)) {
		err = EINVAL;
		mutex_print_dbg("mutex - error = %d, pid = %u,"
				" tid = 0x%lx", err, getpid(),
				syscall(SYS_gettid));

		return err;
	}

	if ((cluster == -1) && (core == -1))
		rv = pthread_mutex_lock(&mutex_map->b_lock);
	else if (cluster == NM6408_CCPU)
		rv = pthread_mutex_lock(&mutex_map->c_lock);
	else if (core == -1)
		rv = pthread_mutex_lock(&mutex_map->a_lock[cluster]);
	else
		rv = pthread_mutex_lock(&mutex_map->n_lock[cluster][core]);

	if (rv) {
		if (rv == EOWNERDEAD) {
			mutex_print_dbg("mutex_lock - warn = %d, pid = %u,"
					" tid = 0x%lx", rv, getpid(),
					syscall(SYS_gettid));

			if ((cluster == -1) && (core == -1))
				pthread_mutex_consistent(&mutex_map->b_lock);
			else if (cluster == NM6408_CCPU)
				pthread_mutex_consistent(&mutex_map->c_lock);
			else if (core == -1)
				pthread_mutex_consistent(&mutex_map->a_lock[cluster]);
			else
				pthread_mutex_consistent(&mutex_map->n_lock[cluster][core]);
		}
		else {
			err = rv;
			mutex_print_dbg("mutex_lock error = %d, pid = %u,"
					" tid = 0x%lx", err, getpid(),
					syscall(SYS_gettid));

			return err;
		}
	}

	mutex_print_dbg("end - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	return err;
}

int libload_mutex_unlock(libload_mutex_t *mutex_map, int cluster, int core)
{
	int rv, err = 0;

	mutex_print_dbg("begin - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	if ((mutex_map == MAP_FAILED) || (mutex_map == NULL)) {
		err = EINVAL;
		mutex_print_dbg("mutex - error = %d, pid = %u,"
				" tid = 0x%lx", err, getpid(),
				syscall(SYS_gettid));

		return err;
	}

	if ((cluster == -1) && (core == -1))
		rv = pthread_mutex_unlock(&mutex_map->b_lock);
	else if (cluster == NM6408_CCPU)
		rv = pthread_mutex_unlock(&mutex_map->c_lock);
	else if (core == -1)
		rv = pthread_mutex_unlock(&mutex_map->a_lock[cluster]);
	else
		rv = pthread_mutex_unlock(&mutex_map->n_lock[cluster][core]);

	if (rv) {
		err = rv;
		mutex_print_dbg("mutex_unlock - error = %d, pid = %u,"
				" tid = 0x%lx", err, getpid(),
				syscall(SYS_gettid));

		return err;
	}

	mutex_print_dbg("end - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	return err;
}

void libload_reset_count(libload_mutex_t *mutex_map, int incr, unsigned int *count)
{
	if (!incr) 
		*count = mutex_map->reset_count;
	else
		mutex_map->reset_count += incr;

	return;
}

void libload_seqnc_count(libload_mutex_t *mutex_map, int rw, unsigned int *count)
{
	if (!!rw)
		mutex_map->seqnc_count = *count;
	else
		*count = mutex_map->seqnc_count;

	return;
}

int libload_rpc_set(libload_mutex_t *mutex_map, int cluster, int core)
{
	int i;

	if (cluster == NM6408_CCPU)
		i = LIBLOAD_PROCESSOR_NUMBER - 1;
	else if (core == -1)
		i = ((cluster + 1) * LIBLOAD_CORE_NUMBER) + cluster;
	else
		i = (cluster * ((LIBLOAD_PROCESSOR_NUMBER - 1) / LIBLOAD_CLUSTER_NUMBER)) + core;

	if (mutex_map->rpc_rdy[i] < 0) {
		int ret;

		errno = 0;
		ret = syscall(SYS_tgkill, mutex_map->rpc_pid[i], mutex_map->rpc_tid[i], 0);
		if (!(ret && (errno == ESRCH)))
			return -1;
	}
	
	mutex_map->rpc_rdy[i] = 1;
	return 0;
}

int libload_rpc_clr(libload_mutex_t *mutex_map, int cluster, int core)
{
	int i;

	if (cluster == NM6408_CCPU)
		i = LIBLOAD_PROCESSOR_NUMBER - 1;
	else if (core == -1)
		i = ((cluster + 1) * LIBLOAD_CORE_NUMBER) + cluster;
	else
		i = (cluster * ((LIBLOAD_PROCESSOR_NUMBER - 1) / LIBLOAD_CLUSTER_NUMBER)) + core;

	if (mutex_map->rpc_rdy[i] < 0) {
		int ret;

		errno = 0;
		ret = syscall(SYS_tgkill, mutex_map->rpc_pid[i], mutex_map->rpc_tid[i], 0);
		if (!(ret && (errno == ESRCH)))
			return -1;
	}

	mutex_map->rpc_rdy[i] = 0; 
	return 0;
}

int libload_rpc_get(libload_mutex_t *mutex_map, int cluster, int core)
{

	int i;

	if (cluster == NM6408_CCPU)
		i = LIBLOAD_PROCESSOR_NUMBER - 1;
	else if (core == -1)
		i = ((cluster + 1) * LIBLOAD_CORE_NUMBER) + cluster;
	else
		i = (cluster * ((LIBLOAD_PROCESSOR_NUMBER - 1) / LIBLOAD_CLUSTER_NUMBER)) + core;

	if (!mutex_map->rpc_rdy[i])
		return 1;

	if (mutex_map->rpc_rdy[i] > 0) {
		mutex_map->rpc_pid[i] = getpid();
		mutex_map->rpc_tid[i] = syscall(SYS_gettid);

		mutex_map->rpc_rdy[i] = -1;
	}
	else if (mutex_map->rpc_rdy[i] < 0) {
		int ret;

		errno = 0;
		ret = syscall(SYS_tgkill, mutex_map->rpc_pid[i], mutex_map->rpc_tid[i], 0);
		if (!(ret && (errno == ESRCH)))
			return -1;

		mutex_map->rpc_pid[i] = getpid();
		mutex_map->rpc_tid[i] = syscall(SYS_gettid);
	}

	return 0;
}

static int libload_boards_find(void)
{
	int		count, index;
	char		*str_index, *end_ptr;
	struct dirent	*entry;
	DIR		*d;

	mutex_print_dbg("begin - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	d = opendir("/dev/");
	if (!d) {
		mutex_print_dbg("opendir - error = %d, pid = %u,"
				" tid = 0x%lx", errno, getpid(),
				syscall(SYS_gettid));
		return -1;
	}

	count = 0;

	while (entry = readdir(d)) {
		if (!strncmp(entry->d_name, NM6408_BRD_NAME"_", strlen(NM6408_BRD_NAME) + 1)) {
			if (count == NM6408_BRD_NUM)
				break;

			str_index = strchr(entry->d_name, '_') + 1;

			errno = 0;
			index = (int)strtol(str_index, &end_ptr, 10);

			if (errno || (str_index == end_ptr) || (*end_ptr != '\0') ||
			    (index < 0) || (index >= NM6408_BRD_NUM)) {
				mutex_print_dbg("board name warn - pid = %u,"
						" tid = 0x%lx", getpid(),
						syscall(SYS_gettid));
				continue;
			}

			ll_mutex_index[count] = index;
			count++;
		}
	}

	closedir(d);

	mutex_print_dbg("end - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	return count;
}

static struct shmid_ds	shmem_stat;
static int 		shmem_id = -1;
static void		*shmem = ((void *)-1);

void libload_mutex_constructor(void)
{
	sem_t	*semaphor = SEM_FAILED;
	char	sem_name[NAME_MAX];
	char	shm_name[NAME_MAX];
	int	count, index;
	int	i, ret;

	mutex_print_dbg("begin - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	ll_mutex_count = 0;

	for (i = 0; i < NM6408_BRD_NUM; i++) {
		ll_mutex_map[i] = (libload_mutex_t *)MAP_FAILED;
		ll_mutex_index[i] = -1;
	}

	sprintf(sem_name, "%s_SEMAPTHOR", NM6408_BRD_NAME);

	while (semaphor == SEM_FAILED) {
		errno = 0;

		semaphor = sem_open(sem_name, O_CREAT|O_EXCL,
			   S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH, 0);

		if (errno == EEXIST)
			usleep(1000);
		else
			break;
	}

	if (semaphor == SEM_FAILED) {
		mutex_print_dbg("constructor - sem_open error = %d, pid = %u,"
				" tid = 0x%lx", errno, getpid(),
				syscall(SYS_gettid));
		return;
	}

	shmem_id = shmget(BOARD_MAGIC_CODE, 1, IPC_CREAT|S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH);
	if (shmem_id == -1) {
		mutex_print_dbg("constructor - shmget error = %d, pid = %u,"
				" tid = 0x%lx", errno, getpid(),
				syscall(SYS_gettid));

		sem_unlink(sem_name);
		sem_close(semaphor);

		return;
	}

	shmem = shmat(shmem_id, NULL, 0);
	if (shmem == ((void *)-1)) {
		mutex_print_dbg("constructor - shmat error = %d, pid = %u,"
				" tid = 0x%lx", errno, getpid(),
				syscall(SYS_gettid));

		sem_unlink(sem_name);
		sem_close(semaphor);

		return;
	}

	if (!shmctl(shmem_id, IPC_STAT, &shmem_stat)) {
		count = libload_boards_find();
		if (count < 0) {
			mutex_print_dbg("constructor, no boards found, error - pid = %u,"
					" tid = 0x%lx", getpid(), syscall(SYS_gettid));

			sem_unlink(sem_name);
			sem_close(semaphor);

			return;
		}

		for (i = 0; i < count; i++) {
			index = ll_mutex_index[i];

			sprintf(shm_name, "/%s_%2.2d_MUTEX", NM6408_BRD_NAME, index);

			ret = libload_mutex_open(shm_name, &ll_mutex_map[index]);
			if (ret) {
				mutex_print_dbg("mutex error - pid = %u,"
						" tid = 0x%lx", getpid(),
						syscall(SYS_gettid));

				ll_mutex_index[i] = -1;
			}
		}

		ll_mutex_count = count;
	}

	sem_unlink(sem_name);
	sem_close(semaphor);

	mutex_print_dbg("end - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));
}

void libload_mutex_destructor(void)
{
	sem_t	*semaphor = SEM_FAILED;
	char	sem_name[NAME_MAX];
	char	shm_name[NAME_MAX];
	int	i, index;

	mutex_print_dbg("begin - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));

	sprintf(sem_name, "%s_SEMAPTHOR", NM6408_BRD_NAME);

	while (semaphor == SEM_FAILED) {
		errno = 0;

		semaphor = sem_open(sem_name, O_CREAT|O_EXCL,
			   S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH, 0);

		if (errno == EEXIST)
			usleep(1000);
		else
			break;
	}

	if (semaphor == SEM_FAILED) {
		mutex_print_dbg("destructor - sem_open error = %d, pid = %u,"
				" tid = 0x%lx", errno, getpid(),
				syscall(SYS_gettid));
		return;
	}

	if (shmem_id != -1) {
		if (shmem != ((void *)-1)) {
			shmdt(shmem);
			shmem = ((void *)-1);
		}

		if (!shmctl(shmem_id, IPC_STAT, &shmem_stat)) {
			for (i = 0; i < ll_mutex_count; i++) {
				index = ll_mutex_index[i];

				if (index >= 0) {
					sprintf(shm_name, "/%s_%2.2d_MUTEX", NM6408_BRD_NAME, index);

					if (ll_mutex_map[index] != MAP_FAILED)
						libload_mutex_close(&ll_mutex_map[index]);

					if (!shmem_stat.shm_nattch)
						shm_unlink(shm_name);

					ll_mutex_index[i] = -1;
				}
			}

			ll_mutex_count = 0;

			if (!shmem_stat.shm_nattch)
				shmctl(shmem_id, IPC_RMID, NULL);
		}

		shmem_id = -1;
	}

	sem_unlink(sem_name);
	sem_close(semaphor);

	mutex_print_dbg("end - pid = %u, tid = 0x%lx",
			getpid(), syscall(SYS_gettid));
}

#ifdef __cplusplus
} // extern "C"
#endif
