/*
 * Load and exchange library function declaration
 * (for board NM_Quad)
 *
 * Copyright (c) 2019 RC Module
 * <nm-support@module.ru>
 */

#ifndef __NM_QUAD_LOAD_H__
#define __NM_QUAD_LOAD_H__

#include <linux/types.h>

#undef FILL_WITH_ZERO

#ifdef __cplusplus
extern "C"
{
#endif

// 32 bit unsigned variable type
typedef __u32			DWORD;

// 64 bit unsigned variable type
typedef __u64			QWORD;

// 32 bit unsigned memory element type
typedef __u32			PL_Word;

// 32 bit unsigned address type
typedef __u32			PL_Addr;

// Board descriptor
typedef struct PL_Board		PL_Board;

// Processor on board descriptor
typedef struct PL_Access	PL_Access;

// Composite core number
struct PL_CoreNo {
	int	nm_id;		// NM number (for any ARM: -1)
	int	cluster_id;	// Cluster number (for central ARM: -1)
};

typedef struct PL_CoreNo	PL_CoreNo;

// Library function return values (all library functions return result code)
enum RetValue {
	PL_OK			= 0, // OK
	PL_ERROR		= 1, // Error
	PL_TIMEOUT		= 2, // Time-out of wait for operation
	PL_FILE			= 3, // Cannot find file for load
	PL_BADADDRESS		= 4, // Bad address ranges
	PL_AGAIN		= 5, // Try again
	PL_NOT_IMPLEMENTED	= -1 // Not implemented
};

#define PROGRAM_NO		0x0
#define PROGRAM_PROGRESS	0x1
#define PROGRAM_FINISHED	0x2
#define PROGRAM_EXEPTION	0x4

/*
 * Common functions
 */

// Return version of library in variable pointed by 'version'
int PL_GetVersion(
		  unsigned int *version_major,
		  unsigned int *version_minor
) __attribute__ ((visibility ("default")));

// Return number of detected boards in variable pointed by 'count'
int PL_GetBoardCount(
		     unsigned int *count
) __attribute__ ((visibility ("default")));

// Set the channel as EDCL
int PL_SetChannelEDCL(
			const unsigned char host_mac_addr[],
			const unsigned char board_mac_addr[]
) __attribute__ ((visibility ("default")));

// Create descriptor for board with number 'index'
// Return descriptor in variable pointed by 'board'
// Boards index count from 0
int PL_GetBoardDesc(
		    unsigned int index,
		    PL_Board **board
) __attribute__ ((visibility ("default")));

// Close board descriptor and free descriptor memory
int PL_CloseBoardDesc(
		      PL_Board *board
) __attribute__ ((visibility ("default")));

/*
 * Board functions
 */

// Get serial number of board
int PL_GetSerialNumber(
		       PL_Board *board,
		       unsigned long *serial_number
) __attribute__ ((visibility ("default")));

// Get version of firmware on board
int PL_GetFirmwareVersion(
			  PL_Board *board,
			  unsigned int *version_major,
			  unsigned int *version_minor
) __attribute__ ((visibility ("default")));

// Do RESET of all processors on board
int PL_ResetBoard(
		  PL_Board *board
) __attribute__ ((visibility ("default")));

// Load init code to board
// Call this function after reset board and before loading
// user program to processors
int PL_LoadInitCode(
		    PL_Board *board
) __attribute__ ((visibility ("default")));

// Create descriptor for processor with number 'coreNo' on board
// Return processor descriptor in variable pointed by 'access'
int PL_GetAccess(
		 PL_Board *board,
		 PL_CoreNo *coreNo,
		 PL_Access **access
) __attribute__ ((visibility ("default")));

// Close processor descriptor and free descriptor memory
int PL_CloseAccess(
		   PL_Access *access
) __attribute__ ((visibility ("default")));

/*
 * Processor functions
 */

// Load user program on processor from memory and start execution
int PL_LoadProgram(
		   PL_Access *access,
		   const void *addrProgram,
		   unsigned int sizeProgram
) __attribute__ ((visibility ("default")));

// Load user program on processor and start execution
int PL_LoadProgramFile(
		       PL_Access *access,
		       const char *filename
) __attribute__ ((visibility ("default")));

int PL_LoadProgramFileWithoutRun(
			PL_Access *access,
			const char *filename,
			PL_Addr *startAddr
) __attribute__ ((visibility ("default")));

int PL_RunProgram(
		PL_Access *access,
		PL_Addr startAddr
) __attribute__ ((visibility ("default")));

// Read array from processor memory
// block	- Pointer to destination buffer in Host PC memory
// len		- Buffer size in 32 bit words
// address	- Address of source array in processor memory
int PL_ReadMemBlock(
		    PL_Access *access,
		    PL_Word *block,
		    PL_Addr address,
		    DWORD len
) __attribute__ ((visibility ("default")));

// Write array to processor memory
// block	- Pointer to source buffer in Host PC memory
// len		- Buffer size in 32 bit words
// address	- address of destination array in processor memory
int PL_WriteMemBlock(
		     PL_Access *access,
		     PL_Word *block,
		     PL_Addr address,
		     DWORD len
) __attribute__ ((visibility ("default")));

// Read register (only for ARM core)
// returnValue	- Value read from processor register
// address	- Address of register to read in processor address space
int PL_ReadRegister(
		    PL_Access *access,
		    PL_Word *returnValue,
		    PL_Addr address
) __attribute__ ((visibility ("default")));

// Write register (only for ARM core)
// value	- Value to write in processor register
// address	- Address of register to write in processor address space
int PL_WriteRegister(
		     PL_Access *access,
		     PL_Word value,
		     PL_Addr address
) __attribute__ ((visibility ("default")));

// Barrier synchronization with program on board processor
// value	- value sent to processor
// returnValue	- value received from processor
int PL_Sync(
	    PL_Access *access,
	    PL_Word value,
	    PL_Word *returnValue
) __attribute__ ((visibility ("default")));

// Barrier synchronization with program on board processor
// Additionally perform two values exchange
// value	- value sent to processor
// outAddress	- address sent to processor
// outLen	- size sent to processor
// returnValue	- value received from processor
// inAddress	- address received from processor
// inLen	- size received from processor
// Values return only if pointers are not NULL
int PL_SyncArray(
		 PL_Access *access,

		 PL_Word value,
		 PL_Addr outAddress,
		 DWORD outLen,

		 PL_Word *returnValue,
		 PL_Addr *inAddress,
		 DWORD *inLen
) __attribute__ ((visibility ("default")));

// Set time-out in milliseconds for waiting function
// Used in all PL_SyncXXX functions
int PL_SetTimeout(
		  DWORD timeout
) __attribute__ ((visibility ("default")));

// Return status bits in variable 'status'
int PL_GetStatus(
		 PL_Access *access,
		 PL_Word *status
) __attribute__ ((visibility ("default")));

// Return user program return value in variable 'result'
int PL_GetResult(
		 PL_Access *access,
		 PL_Word *result
) __attribute__ ((visibility ("default")));

#ifdef __cplusplus
} // extern "C"
#endif

#endif // __NM_QUAD_LOAD_H__
