#pragma once

#include <cstdint>
#include <stdexcept>
#include <string>
#ifdef _WIN32
#include <Windows.h>
#else
#include <dlfcn.h>
#endif

class LibloadWrapper {
private:
#ifdef _WIN32
	using InstanceType = HMODULE;
#else
	using InstanceType = void*;
#endif

	static inline auto GetInstance(const char* path) {
#ifdef _WIN32
		return LoadLibraryA(path);
#else
		return dlopen(path, RTLD_LAZY);
#endif
	}

	static inline void FreeInstance(InstanceType instance) {
#ifdef _WIN32
		FreeLibrary(instance);
#else
		dlclose(instance);
#endif
	}

	static inline auto GetAddress(InstanceType instance, const char* symbol_name) {
#ifdef _WIN32
		return GetProcAddress(instance, symbol_name);
#else
		return dlsym(instance, symbol_name);
#endif
	}

	InstanceType instance_ = nullptr;

	template <typename T>
	void Assign(const char* symbol_name, T& dst_pointer) {
		auto address = GetAddress(instance_, symbol_name);
		if(!address) {
			throw std::runtime_error("Unable to find symbol: " + std::string(symbol_name));
		}
		dst_pointer = reinterpret_cast<T>(address);
	}

public:
	enum RetValue {
		OK              = 0, // Ok
		SYSTEM          = 1, // Error
		TIMEOUT         = 2, // Timeout of wait for operation
		FILE            = 3, // Cannot find file for load
		BADADDRESS      = 4, // Bad address ranges
		NOT_IMPLEMENTED = -1 // Not implemented
	};

	struct CoreNo {
		int nm_id;      // NM number (for any ARM: -1)
		int cluster_id; // Cluster number (for centeral ARM: -1)
	};

	struct Access {};
	struct Board {};
	static constexpr std::uint32_t PROGRAM_FINISHED = 0x2;

	std::int32_t (*CloseAccess)(Access *access) = nullptr;
	std::int32_t (*CloseBoardDesc)(Board *board) = nullptr;
	std::int32_t (*GetAccess)(Board *board, const CoreNo *coreNo, Access **access) = nullptr;
	std::int32_t (*GetBoardCount)(std::uint32_t *board_count) = nullptr;
	std::int32_t (*GetBoardDesc)(std::uint32_t board_number, Board **board) = nullptr;
	std::int32_t (*LoadInitCode)(Board *board) = nullptr;
	std::int32_t (*LoadProgram)(Access *access, const void *program_data, std::uint32_t program_bytes) = nullptr;
	std::int32_t (*LoadProgramFile)(Access *access, const char *program_filename) = nullptr;
	std::int32_t (*ReadMemBlock)(Access *access, std::uint32_t *data, std::uint32_t addr, std::uint32_t size) = nullptr;
	std::int32_t (*ReadRegister)(Access *access, std::uint32_t *returnValue, std::uint32_t address){};
	std::int32_t (*ResetBoard)(Board *board) = nullptr;
	std::int32_t (*SetChannelEDCL)(const unsigned char host_mac_addr[6], const unsigned char board_mac_addr[6]){};
	std::int32_t (*Sync)(Access *access, std::int32_t value, std::int32_t *returnValue) = nullptr;
	std::int32_t (*WriteMemBlock)(Access *access, const std::uint32_t *data, std::uint32_t addr, std::uint32_t size) = nullptr;
	std::int32_t (*WriteRegister)(Access *access, std::uint32_t value, std::uint32_t address){};
	std::int32_t (*GetSerialNumber)(Board* board, unsigned long* serial_number) = nullptr;
	//std::int32_t (*GetPostStatus)(Board* board, int* status) = nullptr;

	LibloadWrapper() = default;

	
	//~LibloadWrapper() {
	//	Free();
	//}
	

	bool Load(const char *filename) {
		instance_ = GetInstance(filename);
		if(!instance_) {
			return false;
		}

		try {
			Assign("PL_CloseAccess", CloseAccess);
			Assign("PL_CloseBoardDesc", CloseBoardDesc);
			Assign("PL_GetAccess", GetAccess);
			Assign("PL_GetBoardCount", GetBoardCount);
			Assign("PL_GetBoardDesc", GetBoardDesc);
			Assign("PL_LoadInitCode", LoadInitCode);
			Assign("PL_LoadProgram", LoadProgram);
			Assign("PL_LoadProgramFile", LoadProgramFile);
			Assign("PL_ReadMemBlock", ReadMemBlock);
			Assign("PL_ReadRegister", ReadRegister);
			Assign("PL_ResetBoard", ResetBoard);
			Assign("PL_SetChannelEDCL", SetChannelEDCL);
			Assign("PL_Sync", Sync);
			Assign("PL_WriteMemBlock", WriteMemBlock);
			Assign("PL_WriteRegister", WriteRegister);
			Assign("PL_GetSerialNumber", GetSerialNumber);
			//Assign("PL_GetPostStatus", GetPostStatus);
		}
		catch(const std::exception &) {
			if(instance_) {
				FreeInstance(instance_);
				instance_ = nullptr;
			}
		}

		return IsLoaded();
	}

	void Free() {
		if(instance_) {
			FreeInstance(instance_);
			instance_ = nullptr;
		}
	}

	bool IsLoaded() const {
		return instance_ ? true : false;
	}

};
