#include <iostream>
#include <map>
#include <vector>

#include "libload_wrapper.hpp"
#include "resource.h"
#include "temperature_reader.hpp"
#include "curses.h"

#define PL_CALL(pl_result, error) if((pl_result) != 0) throw std::runtime_error(error)

struct nm6408_module
{
	std::string board_name;
	LibloadWrapper wrapper;
	std::uint32_t boards_cnt = 0;
	std::vector<LibloadWrapper::Board*> descriptors;

	nm6408_module (const std::string &board, const LibloadWrapper &lib_wrapper, const std::uint32_t boards)
	{
		board_name = board;
		wrapper = lib_wrapper;
		boards_cnt = boards;
	}
};

auto ReadTemperature(const LibloadWrapper& wrapper, LibloadWrapper::Board* board)
{
	LibloadWrapper::Access* access = nullptr;
	constexpr LibloadWrapper::CoreNo core_number = { -1, -1 };

	PL_CALL(wrapper.GetAccess(board, &core_number, &access), "ERROR: No access to CCPU");
	const auto result = TemperatureReader::ReadTemperature(wrapper, access);
	wrapper.CloseAccess(access);
	return result;
}

std::vector<int> ReadTemperatureVector(const LibloadWrapper& wrapper, const std::vector<LibloadWrapper::Board*> &nm_boards)
{
	std::vector<int> temperature_vector;
	temperature_vector.reserve(nm_boards.size());
	for (const auto board : nm_boards)
	{
		temperature_vector.push_back(ReadTemperature(wrapper, board));
	}
	return temperature_vector;
}

std::vector<std::string> SplitRows(const std::string& message)
{
	std::vector<std::string> sentences;
	std::string sentence;
	for (const char c : message)
	{
		if (c == '\n')
		{
			if (!sentence.empty())
			{
				sentences.push_back(sentence);
				sentence.clear();
			}
		}
		else
		{
			sentence += c;
		}
	}
	if (!sentence.empty())
	{
		sentences.push_back(sentence);
	}

	return sentences;
}

std::vector<nm6408_module> GetInstalledModules()
{
	const std::map<std::string ,std::string> lib_load_names =
	{
		{"MC127.05", "libmc12705load.so"},
		{"NeuroMezzanine", "libneuro_mezzanine_load.so"},
		{"NM_Card", "libnm_card_load.so"},
		{"NM_Mezzo", "libnm_mezzo_load.so"},
		{"NM_Quad", "libnm_quad_load.so"}
	};
	std::vector<nm6408_module> installed_modules;
	for (const auto &lib_load : lib_load_names)
	{
		LibloadWrapper wrapper;
		if (wrapper.Load(lib_load.second.c_str()))
		{
			std::uint32_t boards_cnt = 0;
			PL_CALL(wrapper.GetBoardCount(&boards_cnt), "Failed to get board count");
			if (boards_cnt)
			{
				installed_modules.emplace_back(lib_load.first, wrapper, boards_cnt);
			}	
		}
	}
	if (installed_modules.empty())
	{
		throw std::runtime_error("No installed libraries or boards");
	}
	return installed_modules;
}

void InitBoards(std::vector<nm6408_module> &nm6408_libs)
{
	for (auto &nm6408_lib : nm6408_libs)
	{
		for (std::uint32_t board_cnt = 0; board_cnt < nm6408_lib.boards_cnt; ++board_cnt)
		{
			LibloadWrapper::Board* board = nullptr;
			PL_CALL(nm6408_lib.wrapper.GetBoardDesc(board_cnt, &board), "ERROR: " + nm6408_lib.board_name + " Failed open board");
			nm6408_lib.descriptors.push_back(board);
		}
	}
}

void DeInitBoards(std::vector<nm6408_module> &nm6408_modules)
{
	for (auto& nm6408_module : nm6408_modules)
	{
		for (const auto & descriptor : nm6408_module.descriptors)
		{
			nm6408_module.wrapper.CloseBoardDesc(descriptor);
		}
		nm6408_module.wrapper.Free();
	}
}

std::string MakeMessage(const std::string &board_name, const std::vector<int> &temperatures)
{
	std::string message;
	for (const int temperature : temperatures)
	{
		message.append(board_name);
		message.append(" temperature ");
		message.append(std::to_string(temperature));
		//message.append(" °\n");
		message.append(" deg\n");
		//message.append("\xB0\n");
	}
	return message;
}

void PrintConsole(const std::string &message)
{
	const auto sentences = SplitRows(message);
	const std::size_t middle_row = sentences.size() / 2 + sentences.size() % 2;
	std::size_t row_cnt = 0;

	for (const std::string& sentence : sentences)
	{
		//mvprintw((getmaxy(stdscr) - sentences.size()) / 2, (getmaxx(stdscr) / 2) - 10, "degrees of Celsius:\n");
		//mvprintw((getmaxy(stdscr)/2 - (middle_row - row_cnt)), (getmaxx(stdscr) - sentence.size() / 2), "%s", sentence.c_str());
		(row_cnt <= middle_row) ? mvprintw((getmaxy(stdscr)/2 - (middle_row - row_cnt)), ((getmaxx(stdscr) - sentence.size()) / 2), "%s", sentence.c_str())
								: mvprintw((getmaxy(stdscr)/2 + (row_cnt - middle_row)), ((getmaxx(stdscr) - sentence.size()) / 2), "%s", sentence.c_str());

		
		++row_cnt;
	}
}

int main()
{
	initscr();
	noecho();
	curs_set(0);
	timeout(1000);

	std::vector<nm6408_module> nm6408_modules = GetInstalledModules();
	try
	{
		InitBoards(nm6408_modules);
		while (1)
		{
			std::string message;
			for (const auto& nm6408 : nm6408_modules)
			{
				message.append(MakeMessage(nm6408.board_name, ReadTemperatureVector(nm6408.wrapper, nm6408.descriptors)));
			}
			PrintConsole(message);
			refresh();
			if (getch() != ERR)
				break;
		}
	}
	catch (const std::exception& error) 
	{
		if (!nm6408_modules.empty())
		{
			DeInitBoards(nm6408_modules);
		}
		std::cout << error.what() << std::endl;
		endwin();
		return -1;
	}
	DeInitBoards(nm6408_modules);
	endwin();
	return 0;
}
