﻿#include <conio.h>
#include <iostream>
#include <map>
#include <vector>
#include <windows.h>

#include "libload_wrapper.hpp"
#include "resource.h"
#include "temperature_reader.hpp"

#define PL_CALL(pl_result, error) if((pl_result) != LibloadWrapper::OK) throw std::runtime_error(error)

struct nm6408_module
{
	std::string board_name;
	LibloadWrapper wrapper;
	std::uint32_t boards_cnt = 0;
	std::vector<LibloadWrapper::Board*> descriptors;

	nm6408_module (const std::string &board, const LibloadWrapper &lib_wrapper, const std::uint32_t boards)
	{
		board_name = board;
		wrapper = lib_wrapper;
		boards_cnt = boards;
	}
};

auto ReadTemperature(const LibloadWrapper& wrapper, LibloadWrapper::Board* board)
{
	LibloadWrapper::Access* access = nullptr;
	constexpr LibloadWrapper::CoreNo core_number = { -1, -1 };

	PL_CALL(wrapper.GetAccess(board, &core_number, &access), "ERROR: No access to CCPU");
	const auto result = TemperatureReader::ReadTemperature(wrapper, access);
	wrapper.CloseAccess(access);
	return result;
}

std::vector<int> ReadTemperatureVector(const LibloadWrapper& wrapper, const std::vector<LibloadWrapper::Board*> &nm_boards)
{
	std::vector<int> temperature_vector;
	temperature_vector.reserve(nm_boards.size());
	for (const auto board : nm_boards)
	{
		temperature_vector.push_back(ReadTemperature(wrapper, board));
	}
	return temperature_vector;
}

std::vector<std::string> SplitRows(const std::string& message)
{
	std::vector<std::string> sentences;
	std::string sentence;
	for (const char c : message)
	{
		if (c == '\n')
		{
			if (!sentence.empty())
			{
				sentences.push_back(sentence);
				sentence.clear();
			}
		}
		else
		{
			sentence += c;
		}
	}
	if (!sentence.empty())
	{
		sentences.push_back(sentence);
	}

	return sentences;
}

void PrintConsole(const HANDLE &screen, const std::string &message)
{
	CONSOLE_SCREEN_BUFFER_INFO screen_buffer_info;

	GetConsoleScreenBufferInfo(screen, &screen_buffer_info);

	const COORD max_size = 
	{
		static_cast<SHORT>(screen_buffer_info.srWindow.Right - screen_buffer_info.srWindow.Left + 1),
		static_cast<SHORT>(screen_buffer_info.srWindow.Bottom - screen_buffer_info.srWindow.Top + 1)
	};

	const auto sentences = SplitRows(message);
	//std::size_t max_row = sentences.size();
	const std::size_t middle_row = sentences.size() / 2 + sentences.size() % 2;
	std::size_t row_cnt = 0;
	for (const std::string& sentence : sentences)
	{
		COORD pos;

		pos.X = static_cast<SHORT>((max_size.X - sentence.size()) / 2);
		(row_cnt <= middle_row) ? pos.Y = static_cast<SHORT>(max_size.Y / 2 - (middle_row - row_cnt)) : pos.Y = static_cast<SHORT>(max_size.Y / 2 + (row_cnt - middle_row));

		SetConsoleCursorPosition(screen, pos);
		WriteConsole(screen, sentence.c_str(), sentence.size(), nullptr, nullptr);
		++row_cnt;
	}
}

std::vector<nm6408_module> GetInstalledModules()
{
	const std::map<std::string ,std::string> lib_load_names =
	{
		{"MC127.05", "mc12705_load.dll"},
		{"NeuroMezzanine", "neuro_mezzanine_load.dll"},
		{"NM_Card", "nm_card_load.dll"},
		{"NM_Mezzo", "nm_mezzo_load.dll"},
		{"NM_Quad", "nm_quad_load.dll"}
	};
	std::vector<nm6408_module> installed_modules;
	for (const auto &[board, lib] : lib_load_names)
	{
		if (LibloadWrapper wrapper; wrapper.Load(lib.c_str()))
		{
			std::uint32_t boards_cnt = 0;
			PL_CALL(wrapper.GetBoardCount(&boards_cnt), "Failed to get board count");
			if (boards_cnt)
			{
				installed_modules.emplace_back(board, wrapper, boards_cnt);
			}	
		}
	}
	if (installed_modules.empty())
	{
		throw std::runtime_error("No installed libraries or boards");
	}
	return installed_modules;
}

void InitBoards(std::vector<nm6408_module> &nm6408_libs)
{
	for (auto & [board_name, wrapper, boards_cnt, descriptors] : nm6408_libs)
	{
		for (std::uint32_t board_cnt = 0; board_cnt < boards_cnt; ++board_cnt)
		{
			LibloadWrapper::Board* board = nullptr;
			PL_CALL(wrapper.GetBoardDesc(board_cnt, &board), "ERROR: " + board_name + " Failed open board");
			descriptors.push_back(board);
		}
	}
}

void DeInitBoards(std::vector<nm6408_module> &nm6408_modules)
{
	for (auto& nm6408_module : nm6408_modules)
	{
		for (const auto & descriptor : nm6408_module.descriptors)
		{
			nm6408_module.wrapper.CloseBoardDesc(descriptor);
		}
		nm6408_module.wrapper.Free();
	}
}

std::string MakeMessage(const std::string &board_name, const std::vector<int> &temperatures)
{
	std::string message;
	for (const int temperature : temperatures)
	{
		message.append(board_name);
		message.append(" temperature ");
		message.append(std::to_string(temperature));
		message.append("\xF8\n");
	}
	return message;
}

void MoveCenter()
{
	RECT rect_client, rect_window;
	const HWND h_wnd = GetConsoleWindow();
	GetClientRect(h_wnd, &rect_client);
	GetWindowRect(h_wnd, &rect_window);
	const int pos_x = GetSystemMetrics(SM_CXSCREEN) / 2 - (rect_window.right - rect_window.left) / 2;
	const int pos_y = GetSystemMetrics(SM_CYSCREEN) / 2 - (rect_window.bottom - rect_window.top) / 2;

	MoveWindow(h_wnd, pos_x, pos_y, rect_client.right - rect_client.left, rect_client.bottom - rect_client.top, TRUE);
}

HANDLE SetConsoleParameters()
{
	SetConsoleTitle("NM6408 Temperature Monitor");
	MoveCenter();

	const HANDLE screen = GetStdHandle(STD_OUTPUT_HANDLE);

	//edit console size
	CONSOLE_SCREEN_BUFFER_INFOEX screen_buffer_infoex;

	screen_buffer_infoex.cbSize = sizeof(screen_buffer_infoex);
	GetConsoleScreenBufferInfoEx(screen, &screen_buffer_infoex);
	constexpr COORD buffer_size = { 60, 15 };
	screen_buffer_infoex.dwSize = buffer_size;
	screen_buffer_infoex.srWindow.Left = 0;
	screen_buffer_infoex.srWindow.Right = buffer_size.X;
	screen_buffer_infoex.srWindow.Top = 0;
	screen_buffer_infoex.srWindow.Bottom = buffer_size.Y;
	SetConsoleScreenBufferInfoEx(screen, &screen_buffer_infoex);

	////disable edit mode
	//DWORD prev_console_mode = 0;
	//GetConsoleMode(screen, &prev_console_mode);
	//SetConsoleMode(screen, (prev_console_mode & ~ENABLE_QUICK_EDIT_MODE));

	CONSOLE_CURSOR_INFO cursor_info;
	GetConsoleCursorInfo(screen, &cursor_info);
	cursor_info.bVisible = false; // set the cursor visibility
	SetConsoleCursorInfo(screen, &cursor_info);

	//Set font
	CONSOLE_FONT_INFOEX console_font;
	console_font.cbSize = sizeof(CONSOLE_FONT_INFOEX);
	GetCurrentConsoleFontEx(screen, false, &console_font);
	console_font.dwFontSize.X = 16;
	console_font.dwFontSize.Y = 16;
	lstrcpyW(console_font.FaceName, L"Courier New");
	SetCurrentConsoleFontEx(screen, false, &console_font);

	return screen;
}

int main()
{
	const auto screen = SetConsoleParameters();

	std::vector<nm6408_module> nm6408_modules = GetInstalledModules();
	try
	{
		InitBoards(nm6408_modules);
		while (!_kbhit())
		{
			std::string message;
			for (const auto& nm6408 : nm6408_modules)
			{
				message.append(MakeMessage(nm6408.board_name, ReadTemperatureVector(nm6408.wrapper, nm6408.descriptors)));
			}
			PrintConsole(screen, message);
			Sleep(1000);
			system("cls");  // NOLINT(concurrency-mt-unsafe)
		}
	}
	catch (const std::exception& error) 
	{
		if (!nm6408_modules.empty())
		{
			DeInitBoards(nm6408_modules);
		}
		std::cout << error.what() << std::endl;
		system("pause"); // NOLINT(concurrency-mt-unsafe)
		return -1;
	}
	DeInitBoards(nm6408_modules);
	return 0;
}