// Loader for NM6408 based board
// Output memory to dump file

#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <limits>
#include <cstddef>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "nmrun.h"

using namespace std;

extern PL_Access *Access;
extern PL_CoreNo CoreNo;

extern int Quiet;

extern char *ABS_file_name;

extern PL_Addr Dump_memory_begin;
extern DWORD Dump_memory_size;
extern unsigned int Dump_field_bits;

extern char *recv_file_name;
extern PL_Addr recv_addr;
extern DWORD recv_size;

extern char *send_file_name;
extern PL_Addr send_addr;
extern DWORD send_size;

int Dump_shared_memory(const char *ext)
{
	if (!Quiet)
	{
		cout << "Start dump of shared memory...\n";
		cout.flush();
	}

	PL_Word *Board_memory;
	try
	{
		Board_memory = new PL_Word[Dump_memory_size];
	}
	catch (...)
	{
		cout << "ERROR: Insufficient memory!\n";
		return 1;
	}

	if (PL_ReadMemBlock(Access, Board_memory, Dump_memory_begin, Dump_memory_size) != PL_OK)
	{
		cout << "ERROR: Failed read shared memory!\n";
		delete[] Board_memory;
		return 1;
	}

	unsigned int Dump_file_name_length = strlen(ABS_file_name) + strlen(ext) + 1; // With extension and null terminator

	char *Dump_file_name;
	try
	{
		Dump_file_name = new char[Dump_file_name_length];
	}
	catch (...)
	{
		cout << "ERROR: Insufficient memory!\n";
		delete[] Board_memory;
		return 1;
	}

	strcpy(Dump_file_name, ABS_file_name);
	char *ptr = strrchr(Dump_file_name, '.');
	if (ptr)
		*ptr = '\0';

	strcat(Dump_file_name, ext);

	ofstream Dump_file(Dump_file_name);
	if (!Dump_file)
	{
		cout << "ERROR: Failed open \"" << Dump_file_name << "\" for output!\n";
		delete[] Dump_file_name;
		delete[] Board_memory;
		return 1;
	}

	Dump_file << "================================= Shared memory =================================\n";

	Dump_file << hex << uppercase;

	unsigned long long byte_count = Dump_memory_size << 2;

	for (unsigned long long i = 0; i < byte_count;)
	{
		if (!(i % 32))
		{
			// Output address
			Dump_file << "\n";
			Dump_file << setw(8) << setfill('0');

			if (CoreNo.nm_id < 0)
				Dump_file << (Dump_memory_begin + i) << ":";
			else
				Dump_file << (Dump_memory_begin + (i >> 2)) << ":";
		}

		// Directly output memory
		if ((Dump_field_bits == 64) && ((byte_count - i) >= 8))
		{
			unsigned long long *w64 = (unsigned long long *)((unsigned char *)Board_memory + i);
			Dump_file << "  " << setw(16) << setfill('0') << *w64;
			i += 8;
		}
		else if ((Dump_field_bits == 32) && ((byte_count - i) >= 4))
		{
			unsigned int *w32 = (unsigned int *)((unsigned char *)Board_memory + i);
			Dump_file << "  " << setw(8) << setfill('0') << *w32;
			i += 4;
		}
		else if ((Dump_field_bits == 16) && ((byte_count - i) >= 2))
		{
			unsigned short *w16 = (unsigned short *)((unsigned char *)Board_memory + i);
			Dump_file << "  " << setw(4) << setfill('0') << *w16;
			i += 2;
		}
		else if ((Dump_field_bits == 16) || (Dump_field_bits == 32) || (Dump_field_bits == 64))
		{
			unsigned int j;
			Dump_file << "  ";

			for (j = 0; j < (byte_count - i); j++) {
				unsigned char *w8 = (unsigned char *)((unsigned char *)Board_memory + i + j);
				Dump_file << setw(2) << setfill('0') << *w8;
			}

			 i += j;
		}
		else
		{
			PL_Word bits;
			PL_Word val = *(PL_Word *)((unsigned char *)Board_memory + i);

			if ((byte_count - i) >= 4) {
				i += 4;
				bits = 32;
			}
			else {
				bits = (byte_count - i) * 8;
				i = byte_count;
			}

			PL_Word mask = (1 << Dump_field_bits) - 1;

			for (unsigned int j = 0; j < (bits / Dump_field_bits); j++)
			{
				Dump_file << " " << setw((Dump_field_bits + 3) / 4) << setfill('0') << (val & mask);
				val >>= Dump_field_bits;
			}
		}
	}

	Dump_file << "\n";
	Dump_file << flush;

	Dump_file.close();
	delete[] Dump_file_name;
	delete[] Board_memory;

	return 0;
}

int Receive_binary_dump_from_board()
{
	if (!Quiet)
		cout << "Start binary dump of shared memory...\n" << flush;

	PL_Word *Board_memory;
	try
	{
		Board_memory = new PL_Word[recv_size];
	}
	catch (...)
	{
		cout << "ERROR: Insufficient memory!\n";
		return 1;
	}

	if (PL_ReadMemBlock(Access, Board_memory, recv_addr, recv_size) != PL_OK)
	{
		cout << "ERROR: Failed read range of shared memory [0x";
		cout << hex << uppercase << setw(8) << setfill('0') << recv_addr << " - 0x";
		cout << hex << uppercase << setw(8) << setfill('0') << (recv_addr + recv_size - 1) << "]!\n";
		delete[] Board_memory;
		return 1;
	}

	ofstream Dump_file(recv_file_name, ios::binary | ios::out);
	if (!Dump_file)
	{
		cout << "ERROR: Failed open '" << recv_file_name << "' for writing!\n";
		delete[] Board_memory;
		return 1;
	}
	
	for (unsigned long i = 0; i < recv_size; i++)
		Dump_file.write((char *)(Board_memory + i), sizeof(PL_Word));

	Dump_file.close();
	delete[] Board_memory;

	return 0;
}

int Send_binary_file_to_board()
{
	if (!Quiet)
		cout << "Start loading binary file to shared memory...\n" << flush;

	ifstream Dump_file(send_file_name, ios::binary | ios::in);
	if (!Dump_file)
	{
		cout << "ERROR: Failed open '" << send_file_name << "' for reading!\n";
		return 1;
	}

	// Getting file size
	Dump_file.seekg(0, ios::end);
	unsigned long long Dump_file_size = (unsigned long long)Dump_file.tellg();
	Dump_file.seekg(0, ios::beg);

	unsigned long long Dump_size;

	if (!send_size) {
		Dump_size = Dump_file_size;
		send_size = (Dump_file_size + 3) >> 2;
	}
	else {
		Dump_size = send_size << 2;
	}


	if ((Dump_size > numeric_limits<DWORD>::max()) || (Dump_size > Dump_file_size))
	{
		cout << "ERROR: Specified file size is greater than real file size!\n";
		Dump_file.close();
		return 1;
	}

	PL_Word *Board_memory;
	try
	{
		Board_memory = new PL_Word[send_size];
	}
	catch (...)
	{
		cout << "ERROR: Insufficient memory!\n";
		Dump_file.close();
		return 1;
	}

	Board_memory[send_size - 1] = 0;

	for (unsigned long long i = 0; i < Dump_size; i++)
		Dump_file.read((char *)((unsigned char *)Board_memory + i), 1);

	Dump_file.close();

	if (PL_WriteMemBlock(Access, Board_memory, send_addr, send_size) != PL_OK)
	{
		cout << "ERROR: Failed write range of shared memory [0x";
		cout << hex << uppercase << setw(8) << setfill('0') << send_addr << " - 0x";
		cout << hex << uppercase << setw(8) << setfill('0') << (send_addr + send_size - 1) << "]!\n";
		delete[] Board_memory;
		return 1;
	}

	delete[] Board_memory;

	return 0;
}
